#version 150

#include "shaders/lib/camera.glsl"

#ifdef VERTEX_SHADER

in vec3 in_position;
in vec2 in_texcoord;

out vec3 vs_viewPosition;

uniform mat4 model;

void main() {
    vs_viewPosition = (view * (model * vec4(in_position, 1))).xyz;
    gl_Position = projection * vec4(vs_viewPosition, 1.0);
}

#endif

#ifdef FRAGMENT_SHADER

in vec3 vs_viewPosition;

out vec4 fragColor;

uniform sampler2D in_diffuse;
uniform sampler2D in_depth;
uniform sampler2D in_normal;

uniform vec3 lightPosition;
uniform vec3 lightDirection;
uniform vec3 lightColor;
uniform float lightRadius;
uniform float lightConeAngle;

void main() {
    ivec2 texelCoord = ivec2(gl_FragCoord.x - 0.5, gl_FragCoord.y - 0.5);
    vec3 viewRay = vec3(vs_viewPosition.xy / vs_viewPosition.z, 1.0);
    
    float nonlinearDepth = texelFetch(in_depth, texelCoord, 0).x * 2 - 1;
    const float epsilon = 1e-6;
    float linearDepth = -2.0 * nearClip / min(nonlinearDepth - 1.0, 1.0 - epsilon);

    vec3 viewPosition = viewRay * linearDepth;
    vec3 viewNormal = texelFetch(in_normal, texelCoord, 0).xyz;

    vec3 directionFromLight = normalize(viewPosition - lightPosition);

    vec3 diffuseTerm = texelFetch(in_diffuse, texelCoord, 0).rgb;
    float L = max(dot(-lightDirection, viewNormal), 0);

    float d = dot(lightDirection, viewPosition - lightPosition);
    float attenuation = (d / lightRadius);

    float minDot = cos(lightConeAngle * 0.5);
    float angularAttenuation = 0;

    if (dot(lightDirection, directionFromLight) < minDot) {
        attenuation = 1;
    }
    else {
        angularAttenuation = 1 - (minDot - dot(lightDirection, directionFromLight)) / (minDot - 1);
    }

    fragColor.rgb = diffuseTerm * lightColor * L * (1 - attenuation) * (1 - angularAttenuation);
    fragColor.a = 1.0;
}

#endif

