
vec3 HSVToRGB(vec3 hsv) {
    float chroma = hsv.b * hsv.g;
    float hPrime = hsv.r / 60.0f;
    float x = chroma * (1.0 - abs(mod(hPrime, 2.0) - 1.0));

    vec3 rgb = vec3(0, 0, 0);

    if (0.0 <= hPrime && hPrime < 1.0) {
        rgb = vec3(chroma, x, 0.0);
    }
    else if (1.0 <= hPrime && hPrime < 2.0) {
        rgb = vec3(x, chroma, 0.0);
    }
    else if (2.0 <= hPrime && hPrime < 3.0) {
        rgb = vec3(0.0, chroma, x);
    }
    else if (3.0 <= hPrime && hPrime < 4.0) {
        rgb = vec3(0.0, x, chroma);
    }
    else if (4.0 <= hPrime && hPrime < 5.0) {
        rgb = vec3(x, 0.0, chroma);
    }
    else if (5.0 <= hPrime && hPrime < 6.0) {
        rgb = vec3(chroma, 0.0, x);
    }

    float min = hsv.b - chroma;
    rgb = rgb + vec3(min, min, min);

    return rgb;
}

const vec3 white = vec3(1.0, 1.0, 1.0);
const vec3 black = vec3(0.012, 0.006, 0.02);
const vec3 red = vec3(1.0, 0.0, 0.0);
const vec3 yellow = vec3(1.0, 1.0, 0.0);
const vec3 blue = vec3(0.053, 0.343, 0.798);
const vec3 orange = vec3(1.0, 0.565, 0.0);
const vec3 green = vec3(0.0, 0.661, 0.357);
const vec3 purple = vec3(0.439, 0.212, 0.588);

vec3 RYBToRGB(vec3 ryb) {
    vec3 rgb = white +
        ryb[0] * (red - white) +
        ryb[1] * (yellow - white) +
        ryb[2] * (blue - white) +
        ryb[0] * ryb[1] * (white + orange - red - yellow) +
        ryb[0] * ryb[2] * (white + purple - red - blue) +
        ryb[1] * ryb[2] * (white + green - yellow - blue) +
        ryb[0] * ryb[1] * ryb[2] * (blue + black + red + yellow - purple - green - orange - white);

    return rgb;
}

