#version 150

#ifdef VERTEX_SHADER

in vec3 in_position;
in vec2 in_texcoord;

in int in_codepoint;
in vec3 in_glyphPosition;
in vec4 in_rotation;
in vec2 in_glyphDimensions;
in vec4 in_color;
in vec2 in_scale;
in vec2 in_anchor;

out vec2 texcoord;
out vec4 color;

uniform sampler1D texcoordsMap;

uniform mat4 model;
uniform mat4 viewProjection;

void main() {
    float xSquare = in_rotation.x * in_rotation.x;
    float ySquare = in_rotation.y * in_rotation.y;
    float zSquare = in_rotation.z * in_rotation.z;

    vec3 rot_col0 = vec3(1 - (2 * ySquare) - (2 * zSquare),
                         (2 * in_rotation.x * in_rotation.y) - (2 * in_rotation.w * in_rotation.z),
                         (2 * in_rotation.x * in_rotation.z) + (2 * in_rotation.w * in_rotation.y));

    vec3 rot_col1 = vec3((2 * in_rotation.x * in_rotation.y) + (2 * in_rotation.w * in_rotation.z),
                         1 - (2 * xSquare) - (2 * zSquare),
                         (2 * in_rotation.y * in_rotation.z) - (2 * in_rotation.w * in_rotation.x));

    vec3 rot_col2 = vec3((2 * in_rotation.x * in_rotation.z) - (2 * in_rotation.w * in_rotation.y),
                         (2 * in_rotation.y * in_rotation.z) + (2 * in_rotation.w * in_rotation.x),
                         1 - (2 * xSquare) - (2 * ySquare));

    vec3 modelSpaceAnchorPosition = in_glyphPosition + vec3(in_glyphDimensions * in_anchor, 0);

    mat3 rot_scale = mat3(in_scale.x * in_glyphDimensions.x * rot_col0,
                          in_scale.y * in_glyphDimensions.y * rot_col1,
                          rot_col2);
    vec3 modelPosition = rot_scale * (vec3(in_position.xz - in_anchor, 0));
    modelPosition += modelSpaceAnchorPosition;

    gl_Position = viewProjection * (model * vec4(modelPosition, 1));

    vec4 texcoords = texelFetch(texcoordsMap, in_codepoint, 0);
    vec2 texcoord0 = texcoords.xy;
    vec2 texcoord1 = texcoords.zw;
    texcoord = mix(texcoord0, texcoord1, in_texcoord);

    color = in_color;
}

#endif

#ifdef FRAGMENT_SHADER

in vec2 texcoord;
in vec4 color;

out vec4 fragColor;

uniform sampler2D atlas;

float median(float r, float g, float b) {
    return max(min(r, g), min(max(r, g), b));
}

void main() {
    vec4 s = texture(atlas, texcoord);

    float sigDist = median(s.r, s.g, s.b) - 0.47;
    float opacity = clamp(sigDist / fwidth(sigDist) + 0.5, 0.0, 1.0);

    fragColor = vec4(color.rgb, color.a * opacity);
}

#endif

