
#ifndef _ETC_BUFFERED_LIST_H
#define _ETC_BUFFERED_LIST_H

struct BufferedListNode {
    BufferedListNode *next;
    BufferedListNode *prev;

    uint32 dataSize;
    // @TODO: this should be uint8 data[0];
    void *data;
};

struct BufferedList {
    uint32 bufferSize;
    void *buffer;
    
    BufferedListNode *head;
    BufferedListNode *tail;
};

void AllocateBufferedList(BufferedList *bufferedList, uint32 bufferSize) {
    bufferedList->bufferSize = bufferSize;
    bufferedList->buffer = Plat.AllocateMem(bufferSize);
    
    bufferedList->head = NULL;
    bufferedList->tail = NULL;
}

void PopFront(BufferedList *bufferedList) {
    if (bufferedList->head != NULL) {
        if (bufferedList->head == bufferedList->tail) {
            bufferedList->tail = NULL;
        }
        
        bufferedList->head = bufferedList->head->next;
        bufferedList->head->prev = NULL;
    }
}

void *PushBackPtr(BufferedList *bufferedList, uint32 size) {
    uint32 sizeNeeded = size + sizeof(BufferedListNode);
    
    ASSERT(sizeNeeded <= bufferedList->bufferSize);

    if (bufferedList->head == NULL) {
        BufferedListNode *newHead = (BufferedListNode *)bufferedList->buffer;
        newHead->next = NULL;
        newHead->prev = NULL;
        newHead->dataSize = size;
        newHead->data = (uint8 *)newHead + sizeof(BufferedListNode);
        Plat.ClearMem(newHead->data, size);

        bufferedList->head = newHead;
        bufferedList->tail = newHead;
        
        return newHead->data;
    }

    uint32 insertPosition = (uint8 *)bufferedList->tail - (uint8 *)bufferedList->buffer + sizeof(BufferedListNode) + bufferedList->tail->dataSize;
    
    while (true) {
        uint32 headPosition = (uint8 *)bufferedList->head - (uint8 *)bufferedList->buffer;

        if (insertPosition > headPosition) {
            if (bufferedList->bufferSize - insertPosition >= sizeNeeded) {
                break;
            }
            insertPosition = 0;
        }
        else if (headPosition - insertPosition >= sizeNeeded) {
            break;
        }
        else {
            PopFront(bufferedList);
        }
    }

    BufferedListNode *newTail = (BufferedListNode *)((uint8 *)bufferedList->buffer + insertPosition);
    newTail->next = NULL;
    newTail->prev = bufferedList->tail;
    newTail->dataSize = size;
    newTail->data = (uint8 *)newTail + sizeof(BufferedListNode);
    Plat.ClearMem(newTail->data, size);

    bufferedList->tail->next = newTail;
    bufferedList->tail = newTail;
    
    if (bufferedList->head == NULL) {
        bufferedList->head = newTail;
    }

    return newTail->data;
}

#endif

