
ENUM()
enum CameraType : uint32 {
    CameraType_Perspective,
    CameraType_Orthographic,
};

struct CameraInfo {
    Transform worldTransform;
    mat4 view;
    mat4 projection;
    mat4 viewProjection;

    CameraType cameraType;

    // Orthographic
    real32 width;
    real32 height;

    // Perspective
    real32 fov; // vertical field of view
    real32 aspect;

    real32 nearClip;
    real32 farClip;
};

COMPONENT(Camera) {
    CameraType cameraType;

    // Orthographic
    real32 width;
    real32 height;

    // Perspective
    real32 fov; // vertical field of view
    real32 aspect;

    bool overrideAspect;

    real32 nearClip;
    real32 farClip;
    b32 useFiniteFarClip;

    // TODO: store all these fields in the system, they shouldn't be serialized!
    mat4 projection;
    mat4 view;
    mat4 viewProjection;
};

SYSTEM(Camera, Priority = 90) {
    DynamicArray<CameraComponent> cameras;

    EntityID mainCameraID;
};


mat4 GetViewProjection(CameraSystem *system, EntityID entityID);

CameraInfo GetCameraInfo(EntityDB *entityDB, EntityID id);
CameraInfo GetActiveCameraInfo();

void SetMainCamera(CameraSystem *system, EntityID id);

Ray MousePositionToWorldRay(CameraComponent camera, TransformComponent transform, vec2 mousePos);
Ray MousePositionToWorldRay(CameraInfo cameraInfo, vec2 mousePos);


#if ETC_DEBUG
// Editor Camera
enum EditorCameraState {
    EditorCameraState_None,
    EditorCameraState_DraggingRotation,
    EditorCameraState_SnappingRotation,
};

struct EditorCamera {
    EntityDB *entityDB;
    EntityID entityID;
    EntityID cameraIDToReturnTo;

    EditorCameraState state;

    real32 accelerationScale;

    vec3 linearVelocity;
    real32 yawVelocity;
    real32 pitchVelocity;

    quaternion originalDragRotation;
    vec2 originalMousePosNorm;
    vec2 originalMousePos;
    vec2 virtualMousePos;

    real32 snapRotationStartTime;
    quaternion snapRotationStart;
    quaternion snapRotationTarget;

    real32 speedMultiplier;
};

void InitEditorCamera(EditorCamera *camera, EntityDB *entityDB, vec3 position, quaternion rotation);
void ResetEditorCamera(EditorCamera *camera, vec3 position, quaternion rotation);
void ResetEditorCameraVelocity(EditorCamera *camera);

void UpdateEditorCamera(EditorCamera *camera);

void GetEditorCameraTransforms(EditorCamera *camera, mat4 *view, mat4 *projection);
Transform GetEditorCameraWorldTransform(EditorCamera camera);
CameraInfo GetEditorCameraInfo(EditorCamera camera);

#endif

