 
// @TODO: this might need to be moved to ecs.h/.cpp? Seems like it
typedef int32 EntityIDComparator(EntityDB *, EntityID, EntityID, void *);

void SortEntities_(EntityDB *entityDB, void *entityBuffer, uint32 count, uint32 size, uint32 offset, EntityIDComparator *comparator, void *data) {
    if (count < 2) {
        return;
    }
    EntityID pivot = (EntityID)*((uint8 *)entityBuffer + ((count / 2) * size) + offset);

    u32 i = 0;
    u32 j = count - 1;

    while (true) {
        EntityID a = (EntityID)*((uint8 *)entityBuffer + (i * size) + offset);
        EntityID b = (EntityID)*((uint8 *)entityBuffer + (j * size) + offset);
        while (comparator(entityDB, a, pivot, data) > 0) {
            i++;
            a = (EntityID)*((uint8 *)entityBuffer + (i * size) + offset);
        }
        while (comparator(entityDB, b, pivot, data) < 0) {
            j--;
            b = (EntityID)*((uint8 *)entityBuffer + (j * size) + offset);
        }
        if (i >= j) {
            break;
        }

        if (a != b) {
            void *temp = PushSize(Game.frameMem, size);
            Plat.CopyMem(temp, (uint8 *)entityBuffer + (i * size), size);

            Plat.MoveMem((uint8 *)entityBuffer + (i * size), (uint8 *)entityBuffer + (j * size), size);
            Plat.MoveMem((uint8 *)entityBuffer + (j * size), temp, size);
        }
        
        i++;
        j--;
    }

    SortEntities_(entityDB, entityBuffer, i, size, offset, comparator, data);
    SortEntities_(entityDB, ((uint8 *)entityBuffer + (i * size)), count - i, size, offset, comparator, data);
}
#define SortEntities(entityDB, entityBuffer, count, elemType, field, comparator, data) SortEntities_(entityDB, entityBuffer, count, sizeof(elemType), FIELD_OFFSET(elemType, field), comparator, data);

void SortEntityIDs(EntityDB *entityDB, void *entityIDs, uint32 count, EntityIDComparator *comparator, void *data) {
    SortEntities_(entityDB, entityIDs, count, sizeof(EntityID), 0, comparator, data);
}


// StaticMesh
BITFIELD(DebugStaticMesh, {
    DebugStaticMesh_DrawWireframe,
});

void AddStaticMeshComponent(EntityDB *entityDB, StaticMeshSystem *system, uint32 count, EntityID *entityIDs) {
    for (int i = 0; i < count; i++) {
        if (!Contains(&system->entityIDs, entityIDs[i])) {
            AddEntityToSystem(system, entityIDs[i]);

            StaticMeshComponent *mesh = PushBackPtr(&system->components);
            Plat.ClearMem(mesh, sizeof(StaticMeshComponent));
        }
    }
}

void RemoveStaticMeshComponent(EntityDB *entityDB, StaticMeshSystem *system, EntityID entityID) {
    uint32 index;
    if (Contains(&system->entityIDs, entityID, &index)) {
        RemoveAtIndexBySwap(&system->entityIDs, index);
        RemoveAtIndexBySwap(&system->components, index);
    }
}

b32 GetStaticMeshComponentValue(EntityDB *entityDB, StaticMeshSystem *system, EntityID entityID, StaticMeshComponent *component) {
    b32 result = false;

    uint32 index;
    if (Contains(&system->entityIDs, entityID, &index)) {
        *component = system->components[index];
        result = true;
    }

    return result;
}

void SetStaticMeshComponentValue(EntityDB *entityDB, StaticMeshSystem *system, EntityID entityID, StaticMeshComponent *component) {
    uint32 index;
    if (Contains(&system->entityIDs, entityID, &index)) {
        system->components[index] = *component;
    }
}

void TickStaticMeshSystem(EtcGameMemory *mem, EntityDB *entityDB, StaticMeshSystem *system) {
    TIMED_BLOCK();
    TransformSystem *transformSystem = GetSystem(entityDB, Transform);

    CommandList *commandList = &Game.graphics->gbufferPass;
    uint32 key = GenerateGBufferRenderKey(Game.sceneManager->activeSceneIndex, GBufferStage_Render);

    EnableDepthWrite(commandList, key);
    EnableDepthTest(commandList, key, DepthTest_LessEqual);

    // @NOTE: I think the projection matrix is inverting the winding
    // order of all triangles. This should be culling GL_BACK faces
    EnableFaceCulling(commandList, key, GL_FRONT);

    DisableBlending(commandList, key);
    //EnableBlending(commandList, key, GL_FUNC_ADD, GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    StaticMeshComponent *meshComponent;
    for (int i = 0; i < system->components.count; i++) {
        meshComponent = &system->components.data[i];

        if (meshComponent->hidden) {
            continue;
        }

        Mesh *mesh = GetMesh(Game.assets, meshComponent->mesh);
        ShaderProgram *shader = GetShaderProgram(Game.assets, meshComponent->shader);
        if (mesh == NULL || shader == NULL) {
            continue;
        } 

        SetShader(commandList, key, shader->programID);

        mat4 modelTransform = GetWorldTransformMatrix(transformSystem, system->entityIDs[i]);
        SetUniform(commandList, key, UniformLocation(shader, "model"), modelTransform);

        BindBuffer(commandList, key, GL_ARRAY_BUFFER, mesh->vboID);
        BindBuffer(commandList, key, GL_ELEMENT_ARRAY_BUFFER, mesh->iboID);

        if (mesh->positions != NULL) {
            EnableVertexAttribute(commandList, key,
                                  Position_Attrib, 3, GL_FLOAT, GL_FALSE, 0,
                                  POINTER_OFFSET(mesh->vertices, mesh->positions));
        }

        if (mesh->texcoords != NULL) {
            EnableVertexAttribute(commandList, key,
                                  TexCoord_Attrib, 2, GL_FLOAT, GL_FALSE, 0,
                                  POINTER_OFFSET(mesh->vertices, mesh->texcoords));
        }

        if (mesh->normals != NULL) {
            EnableVertexAttribute(commandList, key,
                                  Normal_Attrib, 3, GL_FLOAT, GL_FALSE, 0,
                                  POINTER_OFFSET(mesh->vertices, mesh->normals));
        }

        // TODO: handle cases with submeshCount == 0
        uint32 indexOffset = 0;
        for (int submesh = 0; submesh < mesh->submeshCount; submesh++) {
            Material material;

            if (submesh < meshComponent->overrideMaterials.count) {
                material = meshComponent->overrideMaterials[submesh];
            }
            else if (mesh->materials != NULL) {
                material = mesh->materials[submesh];
            }
            else {
                material = DebugMem->missingMaterial;
            }

            SetUniform(commandList, key, UniformLocation(shader, "diffuseColor"), material.diffuseColor);

            DrawIndexed(commandList, key, mesh->primitiveType, indexOffset, mesh->submeshLengths[submesh]);

            indexOffset += sizeof(uint32) * mesh->submeshLengths[submesh];
        }

        DisableVertexAttribute(commandList, key, Position_Attrib);
        DisableVertexAttribute(commandList, key, TexCoord_Attrib);
        DisableVertexAttribute(commandList, key, Normal_Attrib);
    }

    DisableFaceCulling(commandList, key);
}

uint32 RaycastStaticMeshSystem(EntityDB *entityDB, StaticMeshSystem *system, Ray ray, RaycastSceneResult *closestResult, ChunkedArray<RaycastSceneResult> *results) {
    uint32 intersectCount = 0;

    RaycastSceneResult closest = {};
    closest.t = INFINITY;

    for (int i = 0; i < system->entityIDs.count; i++) {
        EntityID id = system->entityIDs[i];
        mat4 localToWorld = GetWorldTransformMatrix(GetSystem(entityDB, Transform), id);
        mat4 worldToLocal = OrthogonalInverse(localToWorld);
        Ray localRay = TransformRay(worldToLocal, ray);

        StaticMeshComponent staticMesh = system->components[i];

        Mesh *mesh = GetMesh(Game.assets, staticMesh.mesh);
        if (mesh != NULL && mesh->positions != NULL) {

            real32 t;
            if (RaycastMesh(mesh->positions, mesh->indices, mesh->indexCount, localRay, &t)) {

                // Compute worldspace t parameter
                vec3 localPosition = PointAt(localRay, t);
                vec3 worldPosition = MultiplyPoint(localToWorld, localPosition);
                t = Distance(worldPosition, ray.origin);

                RaycastSceneResult raycastResult = {};
                raycastResult.entityID = id;
                raycastResult.componentType = ComponentType_StaticMesh;
                raycastResult.ray = ray;
                raycastResult.t = t;
                raycastResult.endPoint = worldPosition;

                if (t < closest.t) {
                    closest = raycastResult;
                }

                if (results != NULL) {
                    PushBack(results, raycastResult);
                }

                intersectCount++;
            }
        }
    }

    if (intersectCount > 0 && closestResult != NULL) {
        *closestResult = closest;
    }

    return intersectCount;
}


// Billboard
void AddBillboardComponent(EntityDB *entityDB, BillboardSystem *system, uint32 count, EntityID *entityIDs) {
    for (int i = 0; i < count; i++) {
        AddEntityToSystem(system, entityIDs[i]);
        BillboardComponent billboard = {};
        PushBack(&system->billboards, billboard);
    }
}
void RemoveBillboardComponent(EntityDB *entityDB, BillboardSystem *system, EntityID entityID) {
}

b32 GetBillboardComponentValue(EntityDB *entityDB, BillboardSystem *system, EntityID entityID, BillboardComponent *component) {
    b32 result = false;

    return false;
}
void SetBillboardComponentValue(EntityDB *entityDB, BillboardSystem *system, EntityID entityID, BillboardComponent *component) {
    for (int i = 0; i < system->entityIDs.count; i++) {
        if (system->entityIDs[i] == entityID) {
            system->billboards[i] = *component;
            break;
        }
    }
}

void TickBillboardSystem(EtcGameMemory *mem, EntityDB *entityDB, BillboardSystem *system) {
    TransformSystem *transformSystem = GetSystem(entityDB, Transform);
    for (int i = 0; i < system->billboards.count; i++) {
        EntityID id = system->entityIDs[i];
        BillboardComponent billboard = system->billboards[i];
        EntityID viewer = billboard.viewer;

        int32 billboardIndex = -1;
        if (!HashTableGet(&transformSystem->entityMap, id, &billboardIndex)) {
                continue;
        }

        int32 viewerIndex = -1;
        if (!HashTableGet(&transformSystem->entityMap, viewer, &viewerIndex)) {
            continue;
        }

        
        vec3 viewerPosition = GetWorldPosition(transformSystem, viewer);
        // @TODO: use world coordinates!!!!
        quaternion viewerRotation = transformSystem->rotations[viewerIndex];
        vec3 viewerForward = Rotate(viewerRotation, V3(0, 0, -1));

        SetLocalRotation(transformSystem, id, Inverse(viewerRotation));

        LOGF("lookingForward %f %f %f", viewerForward.x, viewerForward.y, viewerForward.z);
    }
}

// Skybox
void AddSkyboxComponent(EntityDB *entityDB, SkyboxSystem *system, uint32 count, EntityID *entityIDs) {

    for (int i = 0; i < count; i++) {
        PushBack(&system->entityIDs, entityIDs[i]);
        
        SkyboxComponent *skybox = PushBackPtr(&system->skyboxes);
        Plat.ClearMem(skybox, sizeof(SkyboxComponent));
    }
}

void RemoveSkyboxComponent(EntityDB *entityDB, SkyboxSystem *system, EntityID entityID) {
    uint32 index;
    if (Contains(&system->entityIDs, entityID, &index)) {
        
        uint32 endIndex = system->skyboxes.count - 1;

        RemoveAtIndexBySwap(&system->skyboxes, index);

        RemoveAtIndexBySwap(&system->entityIDs, index);
    }
}

b32 GetSkyboxComponentValue(EntityDB *entityDB, SkyboxSystem *system, EntityID entityID, SkyboxComponent *component) {
    b32 result = false;

    uint32 index;
    if (Contains(&system->entityIDs, entityID, &index)) {    
        *component = system->skyboxes[index];
        result = true;
    }

    return result;
}

void SetSkyboxComponentValue(EntityDB *entityDB, SkyboxSystem *system, EntityID entityID, SkyboxComponent *component) {
    uint32 index;
    if (Contains(&system->entityIDs, entityID, &index)) {
        system->skyboxes[index] = *component;
    }
}


void TickSkyboxSystem(EtcGameMemory *mem, EntityDB *entityDB, SkyboxSystem *system) {
    TIMED_BLOCK();

    if (system->skyboxes.count == 0) {
        return;
    }

    SkyboxComponent skybox = system->skyboxes[0];

    if (Cubemap *cubemap = GetCubemap(Game.assets, skybox.cubemap)) {
        uint32 key = GenerateShadingRenderKey(false, Game.sceneManager->activeSceneIndex, ShadingStage_Skybox);
        RenderSkybox(&Game.graphics->shadingPass, key, cubemap->textureID);
    }
}


// Traveller
b32 GetTravelerComponentValue(EntityDB *entityDB, TravelerSystem *system, EntityID entityID, TravelerComponent *component) {
    b32 result = false;

    u32 index;
    if (HashTableGet(&system->entityMap, entityID, &index)) {
        *component = system->travelers[index];
        result = true;
    }

    return result;
}

void SetTravelerComponentValue(EntityDB *entityDB, TravelerSystem *system, EntityID entityID, TravelerComponent *component) {
    for (int i = 0; i < system->entityIDs.count; i++) {
        if (system->entityIDs[i] == entityID) {
            system->travelers[i] = *component;
            break;
        }
    }
}

void AddTravelerComponent(EntityDB *entityDB, TravelerSystem *system, u32 count, EntityID *entityIDs) {
    if (system->entityMap.capacity == 0) {
        AllocateHashTable(&system->entityMap, 1024);
    }
    for (int i = 0; i < count; i++) {
        if (!HashTableContains(&system->entityMap, entityIDs[i])) {
            u32 index = system->entityIDs.count;
            HashTableInsert(&system->entityMap, entityIDs[i], index, true);

            AddEntityToSystem(system, entityIDs[i]);
            TravelerComponent traveler = {};
            PushBack(&system->travelers, traveler);

            TravelerData data = {};
            PushBack(&system->data, data);
        }
    }
}
void RemoveTravelerComponent(EntityDB *entityDB, TravelerSystem *system, EntityID entityID) {
    // deallocate all the paths
}

// @Note: pretty sure the standard use case is sigle entities moving to a unique point.
void GeneratePath(EntityDB *entityDB, EntityID entityID, vec3 *points, u32 pointCount) {
    TransformSystem *transformSystem = GetSystem(entityDB, Transform);
    TravelerSystem *travelerSystem = GetSystem(entityDB, Traveler);

    u32 travelerIndex;
    s32 transformIndex;
    if (!HashTableGet(&travelerSystem->entityMap, entityID, &travelerIndex) ||
        !HashTableGet(&transformSystem->entityMap, entityID, &transformIndex)) {
        return;
    }

    Path path = {};
    path.points = MakeDynamicArray<vec3>(pointCount + 1);
    // @TODE: dont hit the hash table again...
    vec3 position = GetWorldPosition(transformSystem, entityID);

    PushBack(&path.points, position);

    for (int i = 0; i < pointCount; i++) {
        // path.points[i] is the previously inserted point
        path.length += Distance(points[i], path.points[i]);
        PushBack(&path.points, points[i]);
    }

    TravelerData *data= &travelerSystem->data[travelerIndex];
    data->path = path;
}

// @Note: pretty sure the standard use case is sigle entities moving to a unique point.
void GenerateRotationPath(EntityDB *entityDB, EntityID entityID, quaternion *rotations, u32 rotationCount) {
    TransformSystem *transformSystem = GetSystem(entityDB, Transform);
    TravelerSystem *travelerSystem = GetSystem(entityDB, Traveler);

    u32 travelerIndex;
    s32 transformIndex;
    if (!HashTableGet(&travelerSystem->entityMap, entityID, &travelerIndex) ||
        !HashTableGet(&transformSystem->entityMap, entityID, &transformIndex)) {
        return;
    }

    RotationPath path = {};
    path.rotations = MakeDynamicArray<quaternion>(rotationCount + 1);
    // @TODE: dont hit the hash table again...
    quaternion rotation = GetWorldRotation(transformSystem, entityID); // WORLD ROTATION

    PushBack(&path.rotations, rotation);

    for (int i = 0; i < rotationCount; i++) {
        path.distance += Angle(rotations[i], path.rotations[i]);
        PushBack(&path.rotations, rotations[i]);
    }

    TravelerData *state = &travelerSystem->data[travelerIndex];
    state->rotationPath = path;
}

void EmbarkOnPathLerp(TravelerSystem *system, EntityID entityID, real32 duration) {
    u32 index;
    if (!HashTableGet(&system->entityMap, entityID, &index)) {
        return;
    }

    TravelerComponent traveler = system->travelers[index];
    if (traveler.method != TravelerMethod_Lerp) {
        return;
    }
    TravelerData *data = &system->data[index];

    if (data->path.length > 0.0f) {
        data->state = TravelerState_Traveling;
        data->segment = 0;
    }
    else {
        data->state = TravelerState_Arrived;
    }

    if (data->rotationPath.distance > 0.0f) {
        data->rotationState = TravelerState_Traveling;
        data->rotationSegment = 0;    
    }

    data->lerpData.segmentTimeElapsed = 0.0f;
    data->lerpData.duration = duration;
}

bool GetTravelerState(TravelerSystem *system, EntityID entityID, TravelerState *state) {
    u32 index;
    if (!HashTableGet(&system->entityMap, entityID, &index)) {
        return false;
    }

    TravelerData *data = &system->data[index];
    *state = data->state;

    return true;
}

bool SetTravelerState(TravelerSystem *system, EntityID entityID, TravelerState state) {
    u32 index;
    if (!HashTableGet(&system->entityMap, entityID, &index)) {
        return false;
    }

    TravelerData *data = &system->data[index];
    data->state = state;

    return true;
}

vec3 TravelerLerp(TravelerData *data) {
    ASSERT(data->path.length > 0);
    real32 segmentLength = Distance(data->path.points[data->segment], data->path.points[data->segment + 1]);

    real32 segmentPercentage = segmentLength / data->path.length;
    real32 segmentDuration = segmentPercentage * data->lerpData.duration;
    real32 t = data->lerpData.segmentTimeElapsed / segmentDuration;
    
    while (t > 1.0 && data->segment + 1 < data->path.points.count - 1) {
        data->segment++;
        // To carry over the timeElapsed we should recalculate the segmentDuration,
        // because we need to recalculate the percentage of the leftover timeElapsed from the new duration.
        // Lets say the leftover time is 0.1 of a segment that is 33 percent, 
        real32 leftOverTime = data->lerpData.segmentTimeElapsed - segmentDuration;
        segmentLength = Distance(data->path.points[data->segment], data->path.points[data->segment + 1]);
        segmentPercentage = segmentLength / data->path.length;
        segmentDuration = segmentPercentage * data->lerpData.duration;
        
        data->lerpData.segmentTimeElapsed = leftOverTime / segmentPercentage;
        t = data->lerpData.segmentTimeElapsed / segmentDuration;
    }
    // @NOTE: if t > 1 but we're on the last segement.
    // is this necessary
    if (t > 1.0f) {
        t = 1.0f;
    }
    
    //LOGF("%d %f %f %f", data->segment, segmentPercentage, segmentDuration, t);

    vec3 position = Lerp(data->path.points[data->segment], data->path.points[data->segment + 1], t);
    
    if (t == 1.0f) {
        data->segment++;
    }

    if (data->segment == data->path.points.count - 1) {
        data->state = TravelerState_Arrived;
    }

    data->lerpData.segmentTimeElapsed += Game.deltaTime;

    return position;
}

quaternion TravelerLerpRotation(TravelerData *data) {
    ASSERT(data->rotationPath.distance > 0);
    real32 segmentLength = Angle(data->rotationPath.rotations[data->rotationSegment],
                                 data->rotationPath.rotations[data->rotationSegment + 1]);

    real32 segmentPercentage = segmentLength / data->rotationPath.distance;
    real32 segmentDuration = segmentPercentage * data->lerpData.duration;
    real32 t = data->lerpData.segmentTimeElapsed / segmentDuration;
    
    while (t > 1.0 && data->rotationSegment + 1 < data->rotationPath.rotations.count - 1) {
        data->rotationSegment++;
        real32 leftOverTime = data->lerpData.segmentTimeElapsed - segmentDuration;
        segmentLength = Angle(data->rotationPath.rotations[data->rotationSegment],
                              data->rotationPath.rotations[data->rotationSegment + 1]);
        segmentPercentage = segmentLength / data->rotationPath.distance;
        segmentDuration = segmentPercentage * data->lerpData.duration;
        
        data->lerpData.segmentTimeElapsed = leftOverTime / segmentPercentage;
        t = data->lerpData.segmentTimeElapsed / segmentDuration;
    }
    // @NOTE: if t > 1 but we're on the last segement.
    // is this necessary
    if (t > 1.0f) {
        t = 1.0f;
    }
    
    LOGF("%d %f %f %f", data->rotationSegment, segmentPercentage, segmentDuration, t);

    quaternion rotation = Lerp(data->rotationPath.rotations[data->rotationSegment],
                               data->rotationPath.rotations[data->rotationSegment + 1], t);
    
    if (t == 1.0f) {
        data->rotationSegment++;;
    }

    if (data->rotationSegment == data->rotationPath.rotations.count - 1) {
        data->rotationState = TravelerState_Arrived;
    }

    data->lerpData.segmentTimeElapsed += Game.deltaTime;

    return rotation;
}


void TickTravelerSystem(EtcGameMemory *mem, EntityDB *entityDB, TravelerSystem *system) {
    //TIMED_BLOCK();
    
    TransformSystem *transformSystem = GetSystem(entityDB, Transform);
    for (int i = 0; i < system->travelers.count; i++) {
        EntityID id = system->entityIDs[i];
        TravelerComponent traveler = system->travelers[i];
        TravelerData *data = &system->data[i];

        int32 index = -1;
        if (!HashTableGet(&transformSystem->entityMap, id, &index)) {
            continue;
        }
        TravelerMethod method = traveler.method;
        
        if (data->state == TravelerState_Traveling) {
            vec3 position = {};
            switch (method) {
                case TravelerMethod_Lerp : {
                    position = TravelerLerp(data);
                } break;
            }

            SetLocalPosition(transformSystem, id, position); // @TODO: worldPosition
        }

        if (data->rotationState == TravelerState_Traveling) {
            quaternion rotation = {};
            switch (method) {
                case TravelerMethod_Lerp : {
                    rotation = TravelerLerpRotation(data);
                } break;
            }
        
            SetLocalRotation(transformSystem, id, rotation);
        }
    }
}


// Hittable
void AddHittableComponent(EntityDB *entityDB, HittableSystem *system, uint32 count, EntityID *entityIDs) {
    if (system->entityMap.capacity == 0) {
        AllocateHashTable(&system->entityMap, 1024);
    }

    for (int i = 0; i < count; i++) {
        AddEntityToSystem(system, entityIDs[i]);
        HashTableInsert(&system->entityMap, entityIDs[i], system->entityIDs.count - 1);
        
        HittableComponent *hittable = PushBackPtr(&system->hittables);
        Plat.ClearMem(hittable, sizeof(HittableComponent));
    }
}

void RemoveHittableComponent(EntityDB *entityDB, HittableSystem *system, EntityID entityID) {
    uint32 index;
    if (HashTableGet(&system->entityMap, entityID, &index)) {
        
        uint32 endIndex = system->hittables.count - 1;

        RemoveAtIndexBySwap(&system->hittables, index);

        HashTableInsert(&system->entityMap, system->entityIDs[endIndex], index);
        HashTableRemove(&system->entityMap, entityID);

        RemoveAtIndexBySwap(&system->entityIDs, index);
    }
}

b32 GetHittableComponentValue(EntityDB *entityDB, HittableSystem *system, EntityID entityID, HittableComponent *component) {
    b32 result = false;

    uint32 index;
    if (HashTableGet(&system->entityMap, entityID, &index)) {
        *component = system->hittables[index];
        result = true;
    }

    return result;
}

void SetHittableComponentValue(EntityDB *entityDB, HittableSystem *system, EntityID entityID, HittableComponent *component) {
    uint32 index;
    if (HashTableGet(&system->entityMap, entityID, &index)) {
        system->hittables[index] = *component;
    }
}

void CallHittableFunc(EntityDB *entityDB, HittableSystem *system, EntityID hitter, ComponentType hitterType, EntityID hittee, HittableResponseType type, vec3 hitFromDirection) {
    uint32 index;
    if (HashTableGet(&system->entityMap, hittee, &index)) {
        HittableComponent hittable = system->hittables[index];
        for (int i = 0; i < hittable.funcs.count; i++) {
            if (hittable.funcs[i].type == type) {
                CallFunc(hittable.funcs[i].func, entityDB, hitter, hitterType, hittee, hitFromDirection);        
            }
        }
    }
}

void TickHittableSystem(EtcGameMemory *mem, EntityDB *entityDB, HittableSystem *system) {
}


// Looker
void AddLookerComponent(EntityDB *entityDB, LookerSystem *system, uint32 count, EntityID *entityIDs) {
    if (system->entityMap.capacity == 0) {
        AllocateHashTable(&system->entityMap, 1024);
    }

    for (int i = 0; i < count; i++) {
        AddEntityToSystem(system, entityIDs[i]);
        HashTableInsert(&system->entityMap, entityIDs[i], system->entityIDs.count - 1);
        
        LookerComponent *looker = PushBackPtr(&system->lookers);
        Plat.ClearMem(looker, sizeof(LookerComponent));
    }
}

void RemoveLookerComponent(EntityDB *entityDB, LookerSystem *system, EntityID entityID) {
    uint32 index;
    if (HashTableGet(&system->entityMap, entityID, &index)) {
        
        uint32 endIndex = system->lookers.count - 1;

        RemoveAtIndexBySwap(&system->lookers, index);

        HashTableInsert(&system->entityMap, system->entityIDs[endIndex], index);
        HashTableRemove(&system->entityMap, entityID);

        RemoveAtIndexBySwap(&system->entityIDs, index);
    }
}

b32 GetLookerComponentValue(EntityDB *entityDB, LookerSystem *system, EntityID entityID, LookerComponent *component) {
    b32 result = false;

    uint32 index;
    if (HashTableGet(&system->entityMap, entityID, &index)) {
        *component = system->lookers[index];
        result = true;
    }

    return result;
}

void SetLookerComponentValue(EntityDB *entityDB, LookerSystem *system, EntityID entityID, LookerComponent *component) {
    uint32 index;
    if (HashTableGet(&system->entityMap, entityID, &index)) {
        system->lookers[index] = *component;
    }
}

void TickLookerSystem(EtcGameMemory *mem, EntityDB *entityDB, LookerSystem *system) {
    //TIMED_BLOCK();

    ShapeSystem *shapeSys = GetSystem(entityDB, Shape);
    TransformSystem *transSys = GetSystem(entityDB, Transform);

    for (int i = 0; i < system->lookers.count; i++) {
        LookerComponent *looker = &system->lookers[i];
        EntityID prevLookingAt = looker->lookingAt;
        looker->lookingAt = EntityID_Invalid;

        vec3 pos = GetWorldPosition(transSys, looker->eyes);
        quaternion rot = GetWorldRotation(transSys, looker->eyes);

        vec3 forward = Normalize(Rotate(rot, FORWARD));

        Ray ray = MakeRay(pos, forward);
        RaycastSceneResult sceneResult;

        if (RaycastShapeSystem(entityDB, shapeSys, ray, &sceneResult, NULL)) {
            looker->lookingAt = sceneResult.entityID;

            if (looker->lookingAt == prevLookingAt) {
                looker->timeLookingAt += Game.deltaTime;

                // @BUG: this doesnt work with the mouse because lookInput will be ~(0.5, 0.5) will be the
                // value when the mouse hasnt been moved.
                EtcInputDevice *lookDevice = GetMostRecentDevice(Game.inputActions, InputAction_FreeLookXY);
                if (lookDevice == NULL) {
                    continue;
                }
                vec2 lookInput = InputActionAnalogue2D(Game.inputActions, InputAction_FreeLookXY, true);

                bool moved = false;
                switch (lookDevice->type) {
                    case Etc_Mouse: {
                        moved = lookInput.x == 0.5f && lookInput.y == 0.5f;
                    } break;
                    case Etc_Gamepad: {
                        moved = lookInput.x == 0.0f && lookInput.y == 0.0f;
                    } break;
                    default: {
                    } break;
                }
                
                if (!moved) {
                    looker->timeStaringAt += Game.deltaTime;
                }
                else {
                    looker->timeStaringAt = 0;
                }
            }
            else {
                looker->timeStaringAt = 0;
                looker->timeLookingAt = 0;
            }
        }
    }
}

