
COMPONENT(StaticMesh) {
    MeshHandle mesh;
    ShaderProgramHandle shader;

    DynamicArray<Material> overrideMaterials;
    b32 hidden;
};

SYSTEM(StaticMesh, Priority = 100) {
    DynamicArray<StaticMeshComponent> components;
};

COMPONENT(Billboard) {
    EntityID viewer;
};

SYSTEM(Billboard) {
    DynamicArray<BillboardComponent> billboards;
};

COMPONENT(Skybox) {
    CubemapHandle cubemap;
};

SYSTEM(Skybox) {
    DynamicArray<EntityID> id;
    DynamicArray<SkyboxComponent> skyboxes;
};

STRUCT()
struct Path {
    DynamicArray<vec3> points;
    real32 length;
};

STRUCT()
struct RotationPath {
    DynamicArray<quaternion> rotations;
    real32 distance; // the total distance that will be rotated
};

// sometimes we want to serialize a path, sometimes we want to generate one...
ENUM()
enum TravelerMethod : uint32 {
    TravelerMethod_Lerp,
};

COMPONENT(Traveler) {
    // @TODO: pointer to movement function
    TravelerMethod method;
    // do we want different methods for rotating?
};

enum TravelerState {
    TravelerState_Unembarked,
    TravelerState_Traveling,
    TravelerState_Arrived,
};

// @Note: is this the access pattern?
struct TravelerData {
    Path path;
    u32 segment;

    RotationPath rotationPath;
    u32 rotationSegment;
    
    TravelerState state;
    TravelerState rotationState;
    // @TODO: distance along
    // @Maybe: facing point? problem then is multiple things setting the rotation...
    union {
        struct {
            real32 segmentTimeElapsed;
            real32 duration;
        } lerpData;
    };
};

SYSTEM(Traveler) {
    HashTable<EntityID, u32> entityMap;

    // @Maybe: organize components by their movement function
    DynamicArray<TravelerComponent> travelers;
    // segregate rotation and position
    DynamicArray<TravelerData> data;
};


// @TODO: this can be generalized to a StimulusComponent like in LookingGlass games
// a flame hits an oilspill, calls that oil's stimulus function
// The function can have a bitfield, where one of the types is a bullet, so we can call the same FlameOn function
// with multiple types as long as we make sure the component has a list of all the types that can make it on fire
DEFINE_FUNCTIONTYPE(HittableFunc, void, EntityDB *entityDB, EntityID hitter, ComponentType hitterType, EntityID hittee, vec3 hitFromDirection);

ENUM()
enum HittableResponseType : uint32 {
    HittableResponseType_Bullet,
    HittableResponseType_Water,
 };

STRUCT()
struct HittableTypeFuncPair {
    HittableResponseType type;
    HittableFuncHandle func;
};

COMPONENT(Hittable) {
    DynamicArray<HittableTypeFuncPair> funcs;
};

SYSTEM(Hittable) {
    HashTable<EntityID, uint32> entityMap;

    DynamicArray<HittableComponent> hittables;
};


COMPONENT(Looker) {
    EntityID eyes;

    real32 fov; 

    EntityID lookingAt;
    real32 timeLookingAt;
    // @NOTE: staring means that there's no FreeLook input
    real32 timeStaringAt;
};

SYSTEM(Looker) {
    HashTable<EntityID, uint32> entityMap;

    DynamicArray<LookerComponent> lookers;
};

BITFIELD(DebugLooker, {
        DebugLooker_Draw,
});
