
#ifndef _ETC_CORE_OPENGL_H
#define _ETC_CORE_OPENGL_H

#ifdef _WIN32
#define GLEW_STATIC
#include <gl/glew.h>
#include <gl/wglew.h>

#elif __APPLE__
#import <OpenGL/gl3.h>

#elif __linux__
#include <GL/glew.h>
#include <GL/glx.h>

#endif

void CreateRenderTargetOpenGL(RenderTarget *renderTarget, uint32 width, uint32 height, uint32 colorTargetCount, ColorFormat *colorFormats, uint8 depthBits);
void DeleteRenderTargetOpenGL(RenderTarget *renderTarget);

void ShareDepthBuffer(RenderTarget *source, RenderTarget *dest);

void CreateSamplerOpenGL(Sampler *sampler, TextureFilterMode filterMode, TextureWrapMode wrapMode);
void CreateSamplerOpenGL(Sampler *sampler, TextureFilterMode filterMode, TextureWrapMode wrapMode, vec4 borderColor);
void DeleteSamplerOpenGL(Sampler *sampler);

void CreateTexture1DOpenGL(Texture *texture);
void CreateTexture2DOpenGL(Texture *texture);
void DeleteTextureOpenGL(Texture *texture);

void UpdateTexture1DOpenGL(Texture *texture);
void UpdateTexture2DOpenGL(Texture *texture);

void CreateCubemapOpenGL(Cubemap *cubemap);
void DeleteCubemapOpenGL(Cubemap *cubemap);

void CreateMeshOpenGL(Mesh *mesh);
void DeleteMeshOpenGL(Mesh *mesh);

void CreateShaderProgramOpenGL(ShaderProgram *program);
void DeleteShaderProgramOpenGL(ShaderProgram *program);

#if !ETC_DEBUG
#define glCheckError()

#else

inline b32 glCheckError_(const char *file, uint32 line) {
    GLenum _glError = glGetError();
    if (_glError != GL_NO_ERROR) {
        //LOGF("OpenGL error (%s:%d): 0x%x (%d)\n", file, line, _glError, _glError);
        return true;
    }

    return false;
}

#define glCheckError() glCheckError_(__FILE__, __LINE__);

#endif

#endif

