
#ifndef _ETC_INPUT_H
#define _ETC_INPUT_H

ENUM()
enum EtcInputDeviceType : uint32 {
    Etc_Keyboard,
    Etc_Mouse,
    Etc_Gamepad,
};

struct EtcInputDevice {
    EtcInputDeviceType type;
    int32 id;

    b32 connected;
    b32 blocked;

    // -1 means nothing has written to the Device this frame
    int32 lastEventIndex;

    int32 discreteCount;
    bool *pressed;
    bool *hold;
    bool *released;

    // These values are only relevant for the discrete events
    uint32 *framesSinceChange;
    real32 *secondsSinceChange;


    int32 analogueCount;
    real32 *prevAnalogue;
    real32 *analogue;


    int32 outputCount;
    real32 *output;
};

ENUM()
enum EtcKeyboardDiscrete : uint32 {
    Etc_A,
    Etc_B,
    Etc_C,
    Etc_D,
    Etc_E,
    Etc_F,
    Etc_G,
    Etc_H,
    Etc_I,
    Etc_J,
    Etc_K,
    Etc_L,
    Etc_M,
    Etc_N,
    Etc_O,
    Etc_P,
    Etc_Q,
    Etc_R,
    Etc_S,
    Etc_T,
    Etc_U,
    Etc_V,
    Etc_W,
    Etc_X,
    Etc_Y,
    Etc_Z,
    Etc_0,
    Etc_1,
    Etc_2,
    Etc_3,
    Etc_4,
    Etc_5,
    Etc_6,
    Etc_7,
    Etc_8,
    Etc_9,
    Etc_Equal,
    Etc_Minus,
    Etc_RightBracket,
    Etc_LeftBracket,
    Etc_Quote,
    Etc_Semicolon,
    Etc_Backslash,
    Etc_Comma,
    Etc_Slash,
    Etc_Period,
    Etc_Return,
    Etc_Tab,
    Etc_Space,
    Etc_Backspace,
    Etc_Escape,
    Etc_Tick,
    Etc_Win,
    Etc_Shift,
    Etc_CapsLock,
    Etc_Alt,
    Etc_Control,
    Etc_RightWin,
    Etc_RightShift,
    Etc_RightAlt,
    Etc_RightControl,
    Etc_Function,
    Etc_F1,
    Etc_F2,
    Etc_F3,
    Etc_F4,
    Etc_F5,
    Etc_F6,
    Etc_F7,
    Etc_F8,
    Etc_F9,
    Etc_F10,
    Etc_F11,
    Etc_F12,
    Etc_F13,
    Etc_F14,
    Etc_F15,
    Etc_F16,
    Etc_F17,
    Etc_F18,
    Etc_F19,
    Etc_F20,
    Etc_F21,
    Etc_F22,
    Etc_F23,
    Etc_F24,
    Etc_Help,
    Etc_Home,
    Etc_Insert,
    Etc_PageUp,
    Etc_ForwardDelete,
    Etc_End,
    Etc_PageDown,
    Etc_LeftArrow,
    Etc_RightArrow,
    Etc_DownArrow,
    Etc_UpArrow,

    Etc_KeyboardDiscreteCount,
};

ENUM()
enum EtcMouseDiscrete : uint32 {
    Etc_MouseLeft,
    Etc_MouseRight,
    Etc_MouseMiddle,

    Etc_MouseDiscreteCount,
};

ENUM()
enum EtcMouseAnalogue : uint32 {
    Etc_MousePositionX,
    Etc_MousePositionY,
    Etc_MousePositionXNorm,
    Etc_MousePositionYNorm,

    Etc_MousePositionXOffset,
    Etc_MousePositionYOffset,
        
    Etc_ScrollDirection,

    Etc_MouseAnalogueCount,
};

#define Etc_MousePositionXY Etc_MousePositionX
#define Etc_MousePositionXYNorm Etc_MousePositionXNorm

ENUM()
enum EtcGamepadDiscrete : uint32 {
    // Face buttons use Xbox 360 layout
    Etc_FaceA,
    Etc_FaceB,
    Etc_FaceX,
    Etc_FaceY,
    Etc_DirLeft,
    Etc_DirRight,
    Etc_DirUp,
    Etc_DirDown,
    Etc_LeftShoulder,
    Etc_RightShoulder,
    Etc_LeftStick,
    Etc_RightStick,
    Etc_Start,
    Etc_Select,

    Etc_GamepadDiscreteCount,
};

ENUM()
enum EtcGamepadAnalogue : uint32 {
    Etc_LeftX,
    Etc_LeftY,
    Etc_RightX,
    Etc_RightY,
    Etc_LeftTrigger,
    Etc_RightTrigger,

    Etc_GamepadAnalogueCount,
};

#define Etc_LeftXY Etc_LeftX
#define Etc_RightXY Etc_RightX


enum EtcGamepadOutput {
    Etc_LowFrequencyRumble,
    Etc_HighFrequencyRumble,

    Etc_GamepadOutputCount,
};


#define DEADZONE_TRIGGER_MIN 0.1f
#define DEADZONE_TRIGGER_MAX 0.9f

#define DEADZONE_STICK_MIN 0.1f
#define DEADZONE_STICK_MAX 0.9f

inline real32 InputDeadzone(real32 value, real32 deadzoneMin, real32 deadzoneMax) {
    real32 result = value;

    if (value <= deadzoneMin) {
        result = 0.0f;
    }

    if (value >= deadzoneMax) {
        result = 1.0f;
    }

    return result;
}

inline vec2 InputDeadzone(vec2 value, real32 deadzoneMin, real32 deadzoneMax) {

    vec2 result = value;

    real32 length = Length(value);

    if (length <= deadzoneMin) {
        result = V2(0, 0);
    }
    else {
        //real32 t = Clamp01(InvLerp(deadzoneMin, deadzoneMax, length));
        r32 t = (length - deadzoneMin) / (1 - deadzoneMin);
        result = t * Normalize(value);
    }

    return result;
}


void InputDeviceSetDiscreteState(EtcInputDevice *device, int index, b32 value);
void InputDeviceSetAnalogueState(EtcInputDevice *device, int index, real32 value);
void InputDeviceFlush(EtcInputDevice *device, real32 deltaTime);

b32 InputPressed(EtcInputDevice *device, int index, b32 force = false);
b32 InputHold(EtcInputDevice *device, int index, b32 force = false);
b32 InputReleased(EtcInputDevice *device, int index, b32 force = false);
real32 InputSecondsSinceChanged(EtcInputDevice *device, int index, b32 force = false);

real32 InputPrevAnalogue(EtcInputDevice *device, int index, b32 force = false);
vec2 InputPrevAnalogue2D(EtcInputDevice *device, int index, b32 force = false);
real32 InputAnalogue(EtcInputDevice *device, int index, b32 force = false);
vec2 InputAnalogue2D(EtcInputDevice *device, int index, b32 force = false);

void InputSetOutput(EtcInputDevice *device, int index, real32 value);


// Input Event Queue
enum EtcInputEventType {
    InputEvent_Discrete,
    InputEvent_Analogue,
};

struct EtcInputEvent {
    EtcInputDevice *device;

    EtcInputEventType type;
    int32 index;

    b32 discreteValue;
    real32 analogueValue;
};

struct EtcInputEventQueue {
    int32 capacity;
    int32 count;
    EtcInputEvent *queue;
};

inline void EtcInputEventQueuePush(EtcInputEventQueue *queue, EtcInputEvent event) {
    if (queue->count < queue->capacity) {
        queue->queue[queue->count] = event;
        ++queue->count;
    }
}

inline void EtcInputEventQueuePushParams(EtcInputEventQueue *queue, EtcInputDevice *device,
                                         EtcInputEventType type, int32 index, b32 discreteValue, real32 analogueValue) {
    EtcInputEvent event;
    event.device = device;
    event.type = type;
    event.index = index;
    event.discreteValue = discreteValue;
    event.analogueValue = analogueValue;

    EtcInputEventQueuePush(queue, event);
}

void ProcessInputEventQueue(EtcInputEventQueue *queue);


// Mouse Look Management
// @TODO: this is maybe also the start of stealing input in general (for UI and such)
b32 RequestMouseLook();
void ProcessMouseLookRequests();


// Input Actions
ENUM()
enum InputActionBindingType : uint32 {
    InputActionBindingType_Discrete,
    InputActionBindingType_Analogue,
};

#define ACTION(...)

enum InputActionType : uint32;

ACTION(MoveX, Analogue);
ACTION(MoveY, Analogue);
ACTION(FreeLookX, Analogue);
ACTION(FreeLookY, Analogue);

ACTION(LeftHandUse);
ACTION(RightHandUse);
ACTION(LeftHandUseAnalogue, Analogue);
ACTION(RightHandUseAnalogue, Analogue);
ACTION(Action1);
ACTION(Action2);
ACTION(Action3);
ACTION(Action4);

// Editor Actions
ACTION(Pause);
ACTION(StepFrame);

ACTION(EditorPrimaryClick);
ACTION(EditorSecondaryClick);
ACTION(EditorTertiaryClick);
ACTION(EditorScroll);

ACTION(EditorModifier0);
ACTION(EditorModifier1);
ACTION(EditorModifier2);

ACTION(EditorMultiSelect);

ACTION(EditorDeselect);

ACTION(EditorUndo);
ACTION(EditorRedo);

ACTION(EditorMoveX, Analogue);
ACTION(EditorMoveY, Analogue);
ACTION(EditorMoveZ, Analogue);
ACTION(EditorMoveZScroll, Analogue);

ACTION(EditorSelectWorldTransformMode);
ACTION(EditorSelectLocalTransformMode);
ACTION(EditorToggleTransformMode);

ACTION(EditorSelectAxisX);
ACTION(EditorSelectAxisY);
ACTION(EditorSelectAxisZ);

ACTION(EditorToggleSnapPosition);
ACTION(EditorToggleSnapRotation);
ACTION(EditorToggleSnapScale);

ACTION(EditorFreeMove);
ACTION(EditorFreeScale);


ACTION(EditorSnapPosX);
ACTION(EditorSnapNegX);
ACTION(EditorSnapPosY);
ACTION(EditorSnapNegY);
ACTION(EditorSnapPosZ);
ACTION(EditorSnapNegZ);

ACTION(EditorChangeCameraAcceleration);

ACTION(EditorRotateCamera);
ACTION(EditorRotateYaw, Analogue);
ACTION(EditorRotatePitch, Analogue);

ACTION(EditorBoost);
ACTION(EditorAntiBoost);

ACTION(EditorUsePositionHandles);
ACTION(EditorUseRotationHandles);
ACTION(EditorUseScaleHandles);

ACTION(EditorSave);

ACTION(EditorTogglePlayerCamera);

ACTION(EditorMoveCursorRight);
ACTION(EditorMoveCursorLeft);
ACTION(EditorMoveCursorUp);
ACTION(EditorMoveCursorDown);
ACTION(EditorMoveBeginningOfLine);
ACTION(EditorMoveEndOfLine);
ACTION(EditorMarkRegion);

ACTION(EditorSearch);
ACTION(EditorSelectPrevWidget);
ACTION(EditorSelectNextWidget);

ACTION(EditorNextEntity);
ACTION(EditorPrevEntity);

ACTION(ConsoleToggle);
ACTION(LogToggle);
ACTION(StatsToggle);
ACTION(InspectorToggle);
ACTION(SystemDebugPreferencesToggle);

ACTION(RenderModeCycle);

ACTION(ReloadCurrentScene);

ACTION(ToggleSceneEditor);
ACTION(ToggleScriptEditor);
ACTION(ToggleGlyphAnimationEditor);

ACTION(ToggleDebugCamera);
ACTION(EditorCameraJumpToEntity);


#define InputAction_MoveXY InputAction_MoveX
#define InputAction_FreeLookXY InputAction_FreeLookX

#include "input.generated.h"


ENUM()
enum InputModifierOp : uint32 {
    InputModifierOp_And,
    InputModifierOp_Or,
    InputModifierOp_Nand,
};

STRUCT()
struct InputActionBinding {
    EtcInputDeviceType deviceType;
    InputActionBindingType inputType;

    // @NOTE: if discrete0 and discrete1 are equal, then we treat it as if it were 0-1 instead
    // of -1 to 1 analogue range
    uint32 discrete0;
    uint32 discrete1;
    uint32 analogue;
    real32 analogueScale;

    uint32 modifierCount;
    uint32 modifiers[4];
    InputModifierOp modifierOps[4];
};

struct InputActionPrevDeviceInfo {
    EtcInputDevice *device;
    EtcInputDevice *modifierDevice;
    int32 bindingIndex;
};

// @TODO: probably want to store information about how long actions have been in their current state

struct InputActions {
    InputActionBindingType bindingTypes[ENUM_VALUECOUNT(InputActionType)];
    DynamicArray<InputActionBinding> bindings[ENUM_VALUECOUNT(InputActionType)];

    b32 pressed[ENUM_VALUECOUNT(InputActionType)];
    b32 hold[ENUM_VALUECOUNT(InputActionType)];
    b32 released[ENUM_VALUECOUNT(InputActionType)];

    real32 prevAnalogue[ENUM_VALUECOUNT(InputActionType)];
    real32 analogue[ENUM_VALUECOUNT(InputActionType)];

    b32 deviceBlocked[ENUM_VALUECOUNT(InputActionType)];
    b32 actionBlocked[ENUM_VALUECOUNT(InputActionType)];
    EtcInputDevice *device[ENUM_VALUECOUNT(InputActionType)];
    InputActionPrevDeviceInfo prevDeviceInfo[ENUM_VALUECOUNT(InputActionType)];
};

void SerializeInputActions(const char *path, InputActions *actions);
b32 DeserializeInputActions(const char *path, InputActions *actions);


void ProcessInputActions(InputActions *actions, EtcInputDevice *keyboard, EtcInputDevice *mouse, EtcInputDevice *gamepads, uint32 gamepadCount);

EtcInputDevice *GetMostRecentDevice(InputActions *actions, InputActionType type);


b32 InputActionPressed(InputActions *actions, InputActionType type, b32 force = false);
b32 InputActionHold(InputActions *actions, InputActionType type, b32 force = false);
b32 InputActionReleased(InputActions *actions, InputActionType type, b32 force = false);
real32 InputActionAnalogue(InputActions *actions, InputActionType type, b32 applyDeadzone = false, b32 force = false);
real32 InputActionPrevAnalogue(InputActions *actions, InputActionType type, b32 applyDeadzone = false, b32 force = false);

b32 InputActionAnaloguePressed(InputActions *actions, InputActionType type, real32 value, b32 applyDeadzone = false, b32 force = false);
b32 InputActionAnalogueHold(InputActions *actions, InputActionType type, real32 value, b32 applyDeadzone = false, b32 force = false);
b32 InputActionAnalogueReleased(InputActions *actions, InputActionType type, real32 value, b32 applyDeadzone = false, b32 force = false);

vec2 InputActionAnalogue2D(InputActions *actions, InputActionType type, b32 applyDeadzone = false, b32 force = false);
vec2 InputActionPrevAnalogue2D(InputActions *actions, InputActionType type, b32 applyDeadzone = false, b32 force = false);

b32 AnyInputAction(InputActions *inputActions, InputActionType *type = NULL);
b32 AnyInputActionPressed(InputActions *inputActions, InputActionType *type = NULL);

b32 InputActionBound(InputActions *actions, InputActionType type);
b32 BindInputAction(InputActions *actions, InputActionType type, b32 writeStateWhenBound = true, InputActionType *exclusives = NULL, uint32 exclusiveCount = 0);
#endif

