
void AddLightComponent(EntityDB *entityDB, LightSystem *system, uint32 count, EntityID *entityIDs) {
    for (int i = 0; i < count; i++) {
        AddEntityToSystem(system, entityIDs[i]);

        LightComponent component = {};
        PushBack(&system->components, component);
    }
}

void RemoveLightComponent(EntityDB *entityDB, LightSystem *system, EntityID entityID) {
    uint32 index;
    if (Contains(&system->entityIDs, entityID, &index)) {
        RemoveAtIndexBySwap(&system->entityIDs, index);
        RemoveAtIndexBySwap(&system->components, index);
    }
}

b32 GetLightComponentValue(EntityDB *entityDB, LightSystem *system, EntityID entityID, LightComponent *value) {
    b32 result = false;

    uint32 index;
    if (Contains(&system->entityIDs, entityID, &index)) {
        *value = system->components[index];
        result = true;
    }

    return result;
}

void SetLightComponentValue(EntityDB *entityDB, LightSystem *system, EntityID entityID, LightComponent *value) {
    uint32 index;
    if (Contains(&system->entityIDs, entityID, &index)) {
        system->components[index] = *value;
    }
}

uint32 RaycastLightSystem(EntityDB *entityDB, LightSystem *system, Ray ray, RaycastSceneResult *closestResult, ChunkedArray<RaycastSceneResult> *results) {

    TransformSystem *transSys = GetSystem(entityDB, Transform);

    uint32 intersectCount = 0;

    RaycastSceneResult closest = {};
    closest.t = INFINITY;

    real32 lightRadius = 0.2f;

    return intersectCount;
}

void DebugDrawLightSystem(EntityDB *entityDB, LightSystem *system) {
    TransformSystem *transSys = GetSystem(entityDB, Transform);
}

void TickLightSystem(EtcGameMemory *mem, EntityDB *entityDB, LightSystem *system) {

    Mesh *quad = GetMesh(&mem->assets, DebugMem->quad);
    if (quad == NULL) {
        return;
    }

    Mesh *sphere = GetMesh(&mem->assets, DebugMem->sphere);
    if (sphere == NULL) {
        return;
    }

    Mesh *cone = GetMesh(&mem->assets, DebugMem->spotlightCone);
    if (cone == NULL) {
        return;
    }

    CommandList *shadingPass = &mem->graphics.shadingPass;
    CameraInfo cameraInfo = GetActiveCameraInfo();


    // Main rendering setup

    // Setup Directional Lights
    {
        uint32 setupKey = GenerateLightRenderKey(Game.sceneManager->activeSceneIndex, LightType_Directional, false);
        uint32 teardownKey = GenerateLightRenderKey(Game.sceneManager->activeSceneIndex, LightType_Directional, true);

        // Setup
        BindBuffer(shadingPass, setupKey, GL_ARRAY_BUFFER, quad->vboID);
        BindBuffer(shadingPass, setupKey, GL_ELEMENT_ARRAY_BUFFER, quad->iboID);

        EnableVertexAttribute(shadingPass, setupKey, Position_Attrib, 3, GL_FLOAT, GL_FALSE, 0, POINTER_OFFSET(quad->vertices, quad->positions));

        EnableVertexAttribute(shadingPass, setupKey, TexCoord_Attrib, 2, GL_FLOAT, GL_FALSE, 0, POINTER_OFFSET(quad->vertices, quad->texcoords));

        DisableDepthWrite(shadingPass, setupKey);
        DisableDepthTest(shadingPass, setupKey);

        // Teardown
        DisableVertexAttribute(shadingPass, teardownKey, Position_Attrib);
        DisableVertexAttribute(shadingPass, teardownKey, TexCoord_Attrib);
    }

    // Setup Point Lights
    {
        uint32 setupKey = GenerateLightRenderKey(Game.sceneManager->activeSceneIndex, LightType_Point, false);
        uint32 teardownKey = GenerateLightRenderKey(Game.sceneManager->activeSceneIndex, LightType_Point, true);

        // Setup
        BindBuffer(shadingPass, setupKey, GL_ARRAY_BUFFER, sphere->vboID);
        BindBuffer(shadingPass, setupKey, GL_ELEMENT_ARRAY_BUFFER, sphere->iboID);

        EnableVertexAttribute(shadingPass, setupKey, Position_Attrib, 3, GL_FLOAT, GL_FALSE, 0, POINTER_OFFSET(sphere->vertices, sphere->positions));

        EnableVertexAttribute(shadingPass, setupKey, TexCoord_Attrib, 2, GL_FLOAT, GL_FALSE, 0, POINTER_OFFSET(sphere->vertices, sphere->texcoords));

        DisableDepthWrite(shadingPass, setupKey);
        DisableDepthTest(shadingPass, setupKey);

        EnableFaceCulling(shadingPass, setupKey, GL_FRONT);

        // Teardown
        DisableVertexAttribute(shadingPass, teardownKey, Position_Attrib);
        DisableVertexAttribute(shadingPass, teardownKey, TexCoord_Attrib);

        DisableFaceCulling(shadingPass, teardownKey);
    }

    // Setup Spotlights
    {
        uint32 setupKey = GenerateLightRenderKey(Game.sceneManager->activeSceneIndex, LightType_Spotlight, false);
        uint32 teardownKey = GenerateLightRenderKey(Game.sceneManager->activeSceneIndex, LightType_Spotlight, true);

        // Setup
        BindBuffer(shadingPass, setupKey, GL_ARRAY_BUFFER, cone->vboID);
        BindBuffer(shadingPass, setupKey, GL_ELEMENT_ARRAY_BUFFER, cone->iboID);

        EnableVertexAttribute(shadingPass, setupKey, Position_Attrib, 3, GL_FLOAT, GL_FALSE, 0, POINTER_OFFSET(cone->vertices, cone->positions));

        EnableVertexAttribute(shadingPass, setupKey, TexCoord_Attrib, 2, GL_FLOAT, GL_FALSE, 0, POINTER_OFFSET(cone->vertices, cone->texcoords));

        DisableDepthWrite(shadingPass, setupKey);
        DisableDepthTest(shadingPass, setupKey);

        EnableFaceCulling(shadingPass, setupKey, GL_FRONT);

        // Teardown
        DisableVertexAttribute(shadingPass, teardownKey, Position_Attrib);
        DisableVertexAttribute(shadingPass, teardownKey, TexCoord_Attrib);

        DisableFaceCulling(shadingPass, teardownKey);
    }
    

    // LightType setup
    uint32 keys[ENUM_VALUECOUNT(LightType)] = {
        GenerateLightRenderKey(Game.sceneManager->activeSceneIndex, LightType_Directional, false),
        GenerateLightRenderKey(Game.sceneManager->activeSceneIndex, LightType_Point, false),
        GenerateLightRenderKey(Game.sceneManager->activeSceneIndex, LightType_Spotlight, false),
    };

    ShaderProgram *programs[ENUM_VALUECOUNT(LightType)] = {
        GetShaderProgram(&mem->assets, mem->graphics.deferredDirectionalLightShader),
        GetShaderProgram(&mem->assets, mem->graphics.deferredPointLightShader),
        GetShaderProgram(&mem->assets, mem->graphics.deferredSpotlightShader),
    };

    RenderTarget *gbufferTarget = &Game.graphics->gbufferTargets[Game.sceneManager->activeSceneIndex];

    for (int i = 0; i < ENUM_VALUECOUNT(LightType); i++) {
        ShaderProgram *program = programs[i];
        if (program == NULL) {
            continue;
        }

        uint32 key = keys[i];

        SetShader(shadingPass, key, program->programID);

        SetUniform(shadingPass, key, UniformLocation(program, "in_diffuse"), 0, gbufferTarget->colorTargets[0], GL_TEXTURE_2D);
        SetSampler(shadingPass, key, 0, mem->graphics.linearClampEdge);

        SetUniform(shadingPass, key, UniformLocation(program, "in_depth"), 1, gbufferTarget->depthTarget, GL_TEXTURE_2D);
        SetSampler(shadingPass, key, 1, mem->graphics.nearestClampEdge);

        SetUniform(shadingPass, key, UniformLocation(program, "in_normal"), 2, gbufferTarget->colorTargets[1], GL_TEXTURE_2D);
        SetSampler(shadingPass, key, 2, mem->graphics.nearestClampEdge);
    }


    // Rendering
    for (int i = 0; i < system->components.count; i++) {
        EntityID entityID = system->entityIDs[i];
        LightComponent light = system->components[i];

        ShaderProgram *program = programs[light.type];
        if (program ==  NULL) {
            continue;
        }

        uint32 key = keys[light.type];
        Transform transform = GetWorldTransform(GetSystem(entityDB, Transform), entityID);

        switch (light.type) {
            case LightType_Directional: {
                vec3 lightDirectionWorld = Rotate(transform.rotation, V3(1, 0, 0));
                vec3 lightDirectionView = (cameraInfo.view * V4(lightDirectionWorld, 0.0f)).xyz;

                SetUniform(shadingPass, key, UniformLocation(program, "lightDirection"), lightDirectionView);
                SetUniform(shadingPass, key, UniformLocation(program, "lightColor"), light.color);

                DrawIndexed(shadingPass, key, quad->primitiveType, 0, quad->indexCount);
            } break;

            case LightType_Point: {
                vec3 lightPositionView = MultiplyPoint(cameraInfo.view, transform.position);
                
                SetUniform(shadingPass, key, UniformLocation(program, "lightPosition"), lightPositionView);
                SetUniform(shadingPass, key, UniformLocation(program, "lightColor"), light.color);
                SetUniform(shadingPass, key, UniformLocation(program, "lightRadius"), light.radius);

                DrawIndexed(shadingPass, key, sphere->primitiveType, 0, sphere->indexCount);
            } break;

            case LightType_Spotlight: {

                real32 coneAngleRad = DegToRad(light.coneAngle);
                Transform coneTransform = SpotlightConeTransform(light.radius, coneAngleRad, transform);
                mat4 model = ToMat4(coneTransform);

                vec3 lightPositionView = MultiplyPoint(cameraInfo.view, transform.position);
                vec3 lightDirectionWorld = Rotate(transform.rotation, V3(1, 0, 0));
                vec3 lightDirectionView = (cameraInfo.view * V4(lightDirectionWorld, 0.0f)).xyz;

                SetUniform(shadingPass, key, UniformLocation(program, "model"), model);
                SetUniform(shadingPass, key, UniformLocation(program, "lightPosition"), lightPositionView);
                SetUniform(shadingPass, key, UniformLocation(program, "lightDirection"), lightDirectionView);
                SetUniform(shadingPass, key, UniformLocation(program, "lightColor"), light.color);
                SetUniform(shadingPass, key, UniformLocation(program, "lightRadius"), light.radius);
                SetUniform(shadingPass, key, UniformLocation(program, "lightConeAngle"), coneAngleRad);

                DrawIndexed(shadingPass, key, cone->primitiveType, 0, cone->indexCount);
            } break;
        }
    }
}

