
ENUM()
enum LightType : uint32 {
    LightType_Directional,
    LightType_Point,
    LightType_Spotlight,
};

COMPONENT(Light) {
    LightType type;
    vec3 color;
    real32 radius;
    real32 coneAngle;
};

SYSTEM(Light, Priority = 100) {
    DynamicArray<LightComponent> components;
};

inline uint32 GenerateLightRenderKey(uint32 sceneIndex, LightType lightType, b32 isLightTeardown) {
    uint32 tempBitsLeft;
    uint32 key = GenerateShadingRenderKey(false, sceneIndex, ShadingStage_LightAccumulation, &tempBitsLeft);

    AddRenderKeyValue(&key, &tempBitsLeft, lightType, 4);
    AddRenderKeyValue(&key, &tempBitsLeft, (bool)isLightTeardown, 1);

    return key;
}

inline Transform SpotlightConeTransform(real32 radius, real32 coneAngle, Transform transform) {

    Transform coneTransform = {};

    vec3 coneDirection = Rotate(transform.rotation, V3(1, 0, 0));
    coneTransform.position = transform.position + radius * coneDirection;

    coneTransform.rotation = transform.rotation;

    // Two negations are necessary to preserve winding order of cone
    // triangles. This is effectively just doing a 180 deg rotation
    // about Y
    real32 horizScale = radius * tanf(0.5f * coneAngle);
    coneTransform.scale = V3(-radius, horizScale, -horizScale);

    return coneTransform;
}


BITFIELD(DebugLight, {
        DebugLight_Draw,
    });
