
#ifndef _ETC_MATH_CATMULLROM_H
#define _ETC_MATH_CATMULLROM_H

inline void CatmullRomKnotsCentripetal(vec3 p0, vec3 p1, vec3 p2, vec3 p3, real32 *t0, real32 *t1, real32 *t2, real32 *t3) {
    *t0 = 0.0f;
    *t1 = sqrtf(Distance(p0, p1)) + *t0;
    *t2 = sqrtf(Distance(p1, p2)) + *t1;
    *t3 = sqrtf(Distance(p2, p3)) + *t2;
}

inline void CatmullRomKnots(vec3 p0, vec3 p1, vec3 p2, vec3 p3, real32 alpha, real32 *t0, real32 *t1, real32 *t2, real32 *t3) {
    *t0 = 0.0f;
    *t1 = powf(Distance(p0, p1), alpha) + *t0;
    *t2 = powf(Distance(p1, p2), alpha) + *t1;
    *t3 = powf(Distance(p2, p3), alpha) + *t2;
}

inline vec3 CatmullRom(vec3 p0, vec3 p1, vec3 p2, vec3 p3, real32 t0, real32 t1, real32 t2, real32 t3, real32 t) {
    ASSERT(t1 <= t);
    ASSERT(t <= t2);

    vec3 A1 = (t1 - t) / (t1 - t0) * p0 + (t - t0) / (t1 - t0) * p1;
    vec3 A2 = (t2 - t) / (t2 - t1) * p1 + (t - t1) / (t2 - t1) * p2;
    vec3 A3 = (t3 - t) / (t3 - t2) * p2 + (t - t2) / (t3 - t2) * p3;

    vec3 B1 = (t2 - t) / (t2 - t0) * A1 + (t - t0) / (t2 - t0) * A2;
    vec3 B2 = (t3 - t) / (t3 - t1) * A2 + (t - t1) / (t3 - t1) * A3;

    vec3 C = (t2 - t) / (t2 - t1) * B1 + (t - t1) / (t2 - t1) * B2;

    return C;
}

#endif

