
#ifndef ETC_COLOR_H
#define ETC_COLOR_H

typedef vec4 Color;

const vec4 WHITE = V4(1.0f, 1.0f, 1.0f, 1.0f);
const vec4 BLACK = V4(0.0f, 0.0f, 0.0f, 1.0f);
const vec4 CLEAR = V4(0.0f, 0.0f, 0.0f, 0.0f);
const vec4 MAGENTA = V4(1.0f, 0.0f, 1.0f, 1.0f);
const vec4 CYAN = V4(0.0f, 1.0f, 1.0f, 1.0f);
const vec4 YELLOW = V4(1.0f, 1.0f, 0.0f, 1.0f);
const vec4 RED = V4(1.0f, 0.0f, 0.0f, 1.0f);
const vec4 GREEN = V4(0.0f, 1.0f, 0.0f, 1.0f);
const vec4 BLUE = V4(0.0f, 0.0f, 1.0f, 1.0f);
const vec4 DARK_GREY = V4(0.25f, 0.25f, 0.25f, 1.0f);
const vec4 GREY = V4(0.5f, 0.5f, 0.5f, 1.0f);
const vec4 LIGHT_GREY = V4(0.75f, 0.75f, 0.75f, 1.0f);
const vec4 BROWN = V4(0.5f, 0.25f, 0.0f, 1.0f);
const vec4 PURPLE = V4(0.5f, 0.0f, 0.5f, 1.0f);

const vec4 PASTEL_RED = V4(0.91f, 0.30f, 0.24f, 1.0f);
const vec4 PASTEL_YELLOW = V4(0.95f, 0.77f, 0.06f, 1.0f);
const vec4 PASTEL_ORANGE = V4(0.90f, 0.49f, 0.13f, 1.0f);
const vec4 PASTEL_GREEN = V4(0.25f, 0.80f, 0.44f, 1.0f);
const vec4 PASTEL_BLUE = V4(0.20f, 0.60f, 0.86f, 1.0f);
const vec4 PASTEL_PURPLE = V4(0.61f, 0.35f, 0.71f, 1.0f);

const RGBA8 WHITE8 = {255, 255, 255, 255};
const RGBA8 BLACK8 = {0, 0, 0, 255};
const RGBA8 CLEAR8 = {0, 0, 0, 0};
const RGBA8 MAGENTA8 = {255, 0, 255, 255};

#define HexColorf(hexString) V4(((hexString >> 16) & 0xFF) / 255.0f,    \
                                ((hexString >> 8) & 0xFF) / 255.0f,     \
                                (hexString & 0xFF) / 255.0f,            \
                                1.0f)

#define UInt8Colorf(r, g, b) V4((real32)r / 255.0f, \
                                (real32)g / 255.0f, \
                                (real32)b / 255.0f, \
                                1.0f)


// Conversion factors for RYB -> RGB
const vec3 FROM_RYB_WHITE = V3(1.0f, 1.0f, 1.0f);
const vec3 FROM_RYB_BLACK = V3(0.012f, 0.006f, 0.02f);
const vec3 FROM_RYB_RED = V3(1.0f, 0.0f, 0.0f);
const vec3 FROM_RYB_YELLOW = V3(1.0f, 1.0f, 0.0f);
const vec3 FROM_RYB_BLUE = V3(0.053f, 0.343f, 0.798f);
const vec3 FROM_RYB_ORANGE = V3(1.0f, 0.565f, 0.0f);
const vec3 FROM_RYB_GREEN = V3(0.0f, 0.661f, 0.357f);
const vec3 FROM_RYB_PURPLE = V3(0.439f, 0.212f, 0.588f);


vec4 WithAlpha(vec3 color, real32 alpha) {
    vec4 result;
    result.rgb = color;
    result.a = alpha;

    return result;
}

vec4 WithAlpha(vec4 color, real32 alpha) {
    vec4 result;
    color.rgb = color.rgb;
    color.a = alpha;

    return color;
}

vec3 RYBToRGB(vec3 ryb) {
    // This is just expanded trilinear interpolation
    vec3 result = FROM_RYB_WHITE +
        ryb[0] * (FROM_RYB_RED - FROM_RYB_WHITE) +
        ryb[1] * (FROM_RYB_YELLOW - FROM_RYB_WHITE) +
        ryb[2] * (FROM_RYB_BLUE - FROM_RYB_WHITE) +
        ryb[0] * ryb[1] * (FROM_RYB_WHITE + FROM_RYB_ORANGE - FROM_RYB_RED - FROM_RYB_YELLOW) +
        ryb[0] * ryb[2] * (FROM_RYB_WHITE + FROM_RYB_PURPLE - FROM_RYB_RED - FROM_RYB_BLUE) +
        ryb[1] * ryb[2] * (FROM_RYB_WHITE + FROM_RYB_GREEN - FROM_RYB_YELLOW - FROM_RYB_BLUE) +
        ryb[0] * ryb[1] * ryb[2] * (FROM_RYB_BLUE + FROM_RYB_BLACK + FROM_RYB_RED + FROM_RYB_YELLOW -
                                    FROM_RYB_PURPLE - FROM_RYB_GREEN - FROM_RYB_ORANGE - FROM_RYB_WHITE);

    return result;
}

mat3 RGBToRYB_Jacobian(vec3 ryb0) {
    real32 d0d0 =
        (FROM_RYB_RED - FROM_RYB_WHITE)[0] +
        ryb0[1] * (FROM_RYB_WHITE + FROM_RYB_ORANGE - FROM_RYB_RED - FROM_RYB_YELLOW)[0] +
        ryb0[2] * (FROM_RYB_WHITE + FROM_RYB_PURPLE - FROM_RYB_RED - FROM_RYB_BLUE)[0] +
        ryb0[1] * ryb0[2] * (FROM_RYB_BLUE + FROM_RYB_BLACK + FROM_RYB_RED + FROM_RYB_YELLOW - FROM_RYB_PURPLE - FROM_RYB_GREEN - FROM_RYB_ORANGE - FROM_RYB_WHITE)[0];

    real32 d0d1 = 
        (FROM_RYB_YELLOW - FROM_RYB_WHITE)[0] +
        ryb0[0] * (FROM_RYB_WHITE + FROM_RYB_ORANGE - FROM_RYB_RED - FROM_RYB_YELLOW)[0] +
        ryb0[2] * (FROM_RYB_WHITE + FROM_RYB_GREEN - FROM_RYB_YELLOW - FROM_RYB_BLUE)[0] +
        ryb0[0] * ryb0[2] * (FROM_RYB_BLUE + FROM_RYB_BLACK + FROM_RYB_RED + FROM_RYB_YELLOW - FROM_RYB_PURPLE - FROM_RYB_GREEN - FROM_RYB_ORANGE - FROM_RYB_WHITE)[0];

    real32 d0d2 = 
        (FROM_RYB_BLUE - FROM_RYB_WHITE)[0] +
        ryb0[0] * (FROM_RYB_WHITE + FROM_RYB_PURPLE - FROM_RYB_RED - FROM_RYB_BLUE)[0] +
        ryb0[1] * (FROM_RYB_WHITE + FROM_RYB_GREEN - FROM_RYB_YELLOW - FROM_RYB_BLUE)[0] +
        ryb0[0] * ryb0[1] * (FROM_RYB_BLUE + FROM_RYB_BLACK + FROM_RYB_RED + FROM_RYB_YELLOW - FROM_RYB_PURPLE - FROM_RYB_GREEN - FROM_RYB_ORANGE - FROM_RYB_WHITE)[0];

    real32 d1d0 =
        (FROM_RYB_RED - FROM_RYB_WHITE)[1] +
        ryb0[1] * (FROM_RYB_WHITE + FROM_RYB_ORANGE - FROM_RYB_RED - FROM_RYB_YELLOW)[1] +
        ryb0[2] * (FROM_RYB_WHITE + FROM_RYB_PURPLE - FROM_RYB_RED - FROM_RYB_BLUE)[1] +
        ryb0[1] * ryb0[2] * (FROM_RYB_BLUE + FROM_RYB_BLACK + FROM_RYB_RED + FROM_RYB_YELLOW - FROM_RYB_PURPLE - FROM_RYB_GREEN - FROM_RYB_ORANGE - FROM_RYB_WHITE)[1];

    real32 d1d1 = 
        (FROM_RYB_YELLOW - FROM_RYB_WHITE)[1] +
        ryb0[0] * (FROM_RYB_WHITE + FROM_RYB_ORANGE - FROM_RYB_RED - FROM_RYB_YELLOW)[1] +
        ryb0[2] * (FROM_RYB_WHITE + FROM_RYB_GREEN - FROM_RYB_YELLOW - FROM_RYB_BLUE)[1] +
        ryb0[0] * ryb0[2] * (FROM_RYB_BLUE + FROM_RYB_BLACK + FROM_RYB_RED + FROM_RYB_YELLOW - FROM_RYB_PURPLE - FROM_RYB_GREEN - FROM_RYB_ORANGE - FROM_RYB_WHITE)[1];

    real32 d1d2 = 
        (FROM_RYB_BLUE - FROM_RYB_WHITE)[1] +
        ryb0[0] * (FROM_RYB_WHITE + FROM_RYB_PURPLE - FROM_RYB_RED - FROM_RYB_BLUE)[1] +
        ryb0[1] * (FROM_RYB_WHITE + FROM_RYB_GREEN - FROM_RYB_YELLOW - FROM_RYB_BLUE)[1] +
        ryb0[0] * ryb0[1] * (FROM_RYB_BLUE + FROM_RYB_BLACK + FROM_RYB_RED + FROM_RYB_YELLOW - FROM_RYB_PURPLE - FROM_RYB_GREEN - FROM_RYB_ORANGE - FROM_RYB_WHITE)[1];

    real32 d2d0 =
        (FROM_RYB_RED - FROM_RYB_WHITE)[2] +
        ryb0[1] * (FROM_RYB_WHITE + FROM_RYB_ORANGE - FROM_RYB_RED - FROM_RYB_YELLOW)[2] +
        ryb0[2] * (FROM_RYB_WHITE + FROM_RYB_PURPLE - FROM_RYB_RED - FROM_RYB_BLUE)[2] +
        ryb0[1] * ryb0[2] * (FROM_RYB_BLUE + FROM_RYB_BLACK + FROM_RYB_RED + FROM_RYB_YELLOW - FROM_RYB_PURPLE - FROM_RYB_GREEN - FROM_RYB_ORANGE - FROM_RYB_WHITE)[2];

    real32 d2d1 = 
        (FROM_RYB_YELLOW - FROM_RYB_WHITE)[2] +
        ryb0[0] * (FROM_RYB_WHITE + FROM_RYB_ORANGE - FROM_RYB_RED - FROM_RYB_YELLOW)[2] +
        ryb0[2] * (FROM_RYB_WHITE + FROM_RYB_GREEN - FROM_RYB_YELLOW - FROM_RYB_BLUE)[2] +
        ryb0[0] * ryb0[2] * (FROM_RYB_BLUE + FROM_RYB_BLACK + FROM_RYB_RED + FROM_RYB_YELLOW - FROM_RYB_PURPLE - FROM_RYB_GREEN - FROM_RYB_ORANGE - FROM_RYB_WHITE)[2];

    real32 d2d2 = 
        (FROM_RYB_BLUE - FROM_RYB_WHITE)[2] +
        ryb0[0] * (FROM_RYB_WHITE + FROM_RYB_PURPLE - FROM_RYB_RED - FROM_RYB_BLUE)[2] +
        ryb0[1] * (FROM_RYB_WHITE + FROM_RYB_GREEN - FROM_RYB_YELLOW - FROM_RYB_BLUE)[2] +
        ryb0[0] * ryb0[1] * (FROM_RYB_BLUE + FROM_RYB_BLACK + FROM_RYB_RED + FROM_RYB_YELLOW - FROM_RYB_PURPLE - FROM_RYB_GREEN - FROM_RYB_ORANGE - FROM_RYB_WHITE)[2];

    
    mat3 result = {d0d0, d1d0, d2d0,
                   d0d1, d1d1, d2d1,
                   d0d2, d1d2, d2d2};
    return result;
}

vec3 RGBToRYB(vec3 rgb) {
    // The forward process (ryb -> rgb) is straightforward:
    // f(ryb) = rgb

    // To solve inverse, find roots of f(ryb) - rgb = 0 with multivariate Newton-Raphson
    // ryb1 <= ryb0 - J^-1(ryb0) * (f(ryb0) - rgb)

    const real32 threshold = 0.1f;
    const int iterations = 6;
    const int numGuesses = 1;

    vec3 result = V3(0, 0, 0);
    real32 closestDistance = INFINITY;

    for (int g = 0; g < numGuesses; g++) {
        real32 guess = g * (1.0 - numGuesses);

        vec3 ryb0 = V3(g, g, g);
        vec3 ryb1;

        for (int i = 0; i < iterations; i++) {
            mat3 jacobian = RGBToRYB_Jacobian(ryb0);
            mat3 jacobianInverse = {};
            Inverse(jacobian, &jacobianInverse);

            ryb1 = ryb0 - jacobianInverse * (RYBToRGB(ryb0) - rgb);

            ryb0 = ryb1;
        }

        // Normalization process (from Zhang paper)
        real32 minComp = Min(Min(ryb0[0], ryb0[1]), ryb0[2]);
        if (minComp < 0.0f) {
            ryb0 = ryb0 - V3(minComp, minComp, minComp);
        }

        real32 maxComp = Max(Max(ryb0[0], ryb0[1]), ryb0[2]);
        if (maxComp > 1.0f) {
            ryb0 = ryb0 / maxComp;
        }

        vec3 convertedBack = RYBToRGB(ryb0);
        real32 distance = Distance(convertedBack, rgb);

        if (distance < closestDistance) {
            result = ryb0;
            closestDistance = distance;
        }
    }

    return result;
}


void FloatColorToUInt8(uint32 width, uint32 height, uint32 components, real32 *input, uint8 *output) {

    for (int i = 0; i < components * width * height; i++) {
        output[i] = (uint8)Roundf(input[i] * 255);
    }
}


vec3 RGBToHSV(vec3 rgb) {
    int32 minComponent = -1;
    int32 maxComponent = -1;

    if (rgb[0] >= rgb[1] && rgb[0] >= rgb[2]) {
        maxComponent = 0;
    }
    else if (rgb[1] >= rgb[0] && rgb[1] >= rgb[2]) {
        maxComponent = 1;
    }
    else if (rgb[2] >= rgb[0] && rgb[2] >= rgb[1]) {
        maxComponent = 2;
    }

    if (rgb[0] <= rgb[1] && rgb[0] <= rgb[2]) {
        minComponent = 0;
    }
    else if (rgb[1] <= rgb[0] && rgb[1] <= rgb[2]) {
        minComponent = 1;
    }
    else if (rgb[2] <= rgb[0] && rgb[2] <= rgb[1]) {
        minComponent = 2;
    }

    ASSERT(maxComponent != -1 && minComponent != -1);
    
    real32 chroma = rgb[maxComponent] - rgb[minComponent];
    real32 hue = 0.0f;

    if (chroma != 0.0f) {
        if (maxComponent == 0) {
            hue = 60.0f * fmodf(6.0f + (rgb.g - rgb.b) / chroma, 6.0f);
        }
        else if (maxComponent == 1) {
            hue = 60.0f * ((rgb.b - rgb.r) / chroma + 2.0f);
        }
        else if (maxComponent == 2) {
            hue = 60.0f * ((rgb.r - rgb.g) / chroma + 4.0f);
        }
    }

    real32 value = rgb[maxComponent];
    real32 saturation = SafeRatio(chroma, value, 0.0f);

    vec3 hsv = V3(hue, saturation, value);
    return hsv;
}

vec3 HSVToRGB(vec3 hsv) {
    real32 chroma = hsv.v * hsv.s;
    real32 hPrime = hsv.h / 60.0f;
    real32 x = chroma * (1.0f - Absf(fmodf(hPrime, 2.0f) - 1.0f));

    vec3 rgb = {};

    if (0.0f <= hPrime && hPrime < 1.0f) {
        rgb = V3(chroma, x, 0.0f);
    }
    else if (1.0f <= hPrime && hPrime < 2.0f) {
        rgb = V3(x, chroma, 0.0f);
    }
    else if (2.0f <= hPrime && hPrime < 3.0f) {
        rgb = V3(0.0f, chroma, x);
    }
    else if (3.0f <= hPrime && hPrime < 4.0f) {
        rgb = V3(0.0f, x, chroma);
    }
    else if (4.0f <= hPrime && hPrime < 5.0f) {
        rgb = V3(x, 0.0f, chroma);
    }
    else if (5.0f <= hPrime && hPrime < 6.0f) {
        rgb = V3(chroma, 0.0f, x);
    }

    real32 min = hsv.v - chroma;
    rgb = rgb + V3(min, min, min);

    return rgb;
}

vec4 RGBToHSV(vec4 rgba) {
    return WithAlpha(RGBToHSV(rgba.rgb), rgba.a);
}

vec4 HSVToRGB(vec4 hsva) {
    return WithAlpha(HSVToRGB(hsva.rgb), hsva.a);
}

#endif

