
#ifndef _ETC_MATH_RANDOM_H
#define _ETC_MATH_RANDOM_H

struct LCGState {
    int32 seed;
    int32 state;
};

// @TODO:: need to reseed on game library recompiles
static LCGState defaultLCGState;
const int32 aLCG = 1103515245;
const int32 cLCG = 12345;

#define ETC_RAND_MAX ((1 << 15) - 1)

inline void SeedRand(int32 seed) {
    defaultLCGState.seed = seed;
    defaultLCGState.state = seed;
}

inline void SeedRand(LCGState *state, int32 seed) {
    state->seed = seed;
    state->state = seed;
}


inline int32 Randi() {
    defaultLCGState.state = aLCG * defaultLCGState.state + cLCG;
    
    return (defaultLCGState.state >> 16) & 0x7FFF;
}

inline int32 Randi(LCGState *state) {
    state->state = aLCG * state->state + cLCG;
    
    return (state->state >> 16) & 0x7FFF;
}

inline int32 Randi(int32 upperLimit) {
    int32 n = Randi() % upperLimit;
    
    return n;
}

inline int32 Randi(LCGState *state, int32 upperLimit) {
    int32 n = Randi(state) % upperLimit;
    
    return n;
}

inline int32 RandRangei(LCGState *state, int32 lowerLimit = 0, int32 upperLimit = 1) {
    ASSERT(lowerLimit <= upperLimit);

    int32 result;

    if (lowerLimit == upperLimit) {
        result = lowerLimit;
    }
    else {
        int32 range = upperLimit - lowerLimit;
        result = lowerLimit + (Randi(state) % range);
    }

    return result;
}

inline int32 RandRangei(int32 lowerLimit = 0, int32 upperLimit = 1) {
    ASSERT(lowerLimit <= upperLimit);

    int32 result;

    if (lowerLimit == upperLimit) {
        result = lowerLimit;
    }
    else {
        int32 range = upperLimit - lowerLimit;
        result = lowerLimit + (Randi() % range);
    }

    return result;
}

inline real32 Randf(real32 upperLimit = 1) {
    real32 result = ((real32)Randi() / (real32)(ETC_RAND_MAX)) * upperLimit;
    return result;
}

inline real32 Randf(LCGState *state, real32 upperLimit = 1) {
    real32 result = ((real32)Randi(state) / (real32)(ETC_RAND_MAX)) * upperLimit;
    return result;
}

inline real32 Randf(real32 lowerLimit, real32 upperLimit) {
    ASSERT(lowerLimit <= upperLimit);

    real32 result = lowerLimit + Randf(upperLimit - lowerLimit);
    return result;
}

inline real32 Randf(int32 lowerLimit, int32 upperLimit) {
    ASSERT(lowerLimit <= upperLimit);

    real32 result = (real32)lowerLimit + Randf((real32)upperLimit - (real32)lowerLimit);
    return result;
}

inline real32 Randf(LCGState *state, real32 lowerLimit, real32 upperLimit) {
    ASSERT(lowerLimit <= upperLimit);

    real32 result = lowerLimit + Randf(state, upperLimit - lowerLimit);
    return result;
}

#endif

