
STRUCT()
struct Material {
    Color diffuseColor;
    TextureHandle diffuseMap;
};

inline b32 operator==(Material a, Material b) {
    b32 result = a.diffuseColor == b.diffuseColor;
    result &= a.diffuseMap == b.diffuseMap;

    return result;
}

ASSET({"etc_mesh"}, Serializable = false)
struct Mesh {
    PrimitiveType primitiveType;

    uint32 vertexCount;
    uint32 vertexBufferSize;
    void *vertices;

    vec3 *positions;
    vec2 *texcoords;
    vec3 *normals;
    vec4 *colors;

    vec4i *boneIndices;
    vec4 *boneWeights;

    uint32 indexCount;
    uint32 indexBufferSize;
    uint32 *indices;

    uint32 submeshCapacity;
    uint32 submeshCount;
    uint32 *submeshLengths;

    // Materials is either NULL or there is one material per submesh
    Material *materials;

    // runtime data
    uint32 vboID;
    uint32 iboID;
};

void AllocateMesh(Mesh *mesh, uint32 vertexCount, uint32 vertexBufferSize, uint32 indexCount, uint32 submeshCount, b32 allocateMaterials = false);
void ReallocateMesh(Mesh *mesh, uint32 vertexCount, uint32 vertexBufferSize, uint32 indexCount, uint32 submeshCount, b32 allocateMaterials = false);

void DeallocateMesh(Mesh *mesh);

// @TODO: this should go through a unified asset writing system
void MeshWriteFile(const char *path, Mesh *mesh);


// Adjacency Map
ENUM()
enum MeshFeatureType : uint32 { 
    FeatureType_Point,
    FeatureType_Edge,
};

STRUCT()
struct AdjacentTriangle {
    int32 centralIndex;
    uint32 index;

    MeshFeatureType featureType;
    uint32 centralFeatureIndex;
    uint32 featureIndex;
};

STRUCT()
struct AdjacentTriangleSet {
    uint32 start;
    // @NOTE: end is one past-the end of set
    uint32 end;

    int32 centralIndex;
    int32 adjacentIndexAtEdge[3];
};

STRUCT()
struct MeshAdjacencyMap {
    // Maps central triangle index -> (offset, count) in triangles list
    DynamicArray<AdjacentTriangleSet> map;

    DynamicArray<AdjacentTriangle> triangles;
};


void DeallocateMeshAdjacencyMap(MeshAdjacencyMap *adjacency);
void ClearMeshAdjacencyMap(MeshAdjacencyMap *adjacency);

// Assumes indices are packed together in groups of 3 and that the/ next group is `indicesStride`
// bytes from the beginning of the current group. A stride of 0 means tightly packed
void GenerateAdjacencyMap(vec3 *positions, uint32 *indices, uint32 indicesStride, uint32 indexStart, uint32 indexCount, MeshAdjacencyMap *adjacency);
void GenerateAdjacencyMap(Mesh *mesh, MeshAdjacencyMap *adjacency);


