
#ifndef _ETC_PLATFORM_H
#define _ETC_PLATFORM_H

#if __linux__
#include <pthread.h>
#endif

#if __APPLE__
#endif

struct FileHandle;
struct TaskScheduler;
template <typename T> struct DynamicArray;


typedef void PlatformDebugPrintFunc(const char *fmt, ...);

typedef b32 PlatformOpenFileForReadFunc(const char *path, FileHandle *handle);
typedef b32 PlatformOpenFileForWriteFunc(const char *path, FileHandle *handle, uint64 initialDataLength);
typedef b32 PlatformResizeFileFunc(FileHandle *handle, uint64 fileSize, uint64 dataLength);
typedef void PlatformCloseFileFunc(FileHandle *handle);

typedef int PlatformSystemFunc(const char *command);

typedef void PlatformMakeDirectoryFunc(const char *path);
typedef void PlatformRemoveFileFunc(const char *path);

typedef char *PlatformFindFileInDirectoryFunc(const char *directory, const char *name, uint32 extensionCount, const char **extensions);
typedef uint64 PlatformFileLastModifiedTimeFunc(const char *path);
typedef DynamicArray<const char *> PlatformGetAllFilesWithExtensionsFunc(const char *directory, uint32 extensionCount, const char **extensions);

typedef void *PlatformAllocateMemFunc(uint32 size);
typedef void *PlatformResizeMemFunc(void *memory, uint32 size);
typedef void *PlatformCopyMemFunc(void *destination, const void *source, uint32 size);
typedef void *PlatformMoveMemFunc(void *destination, const void *source, uint32 size);
typedef void PlatformDeallocateMemFunc(void *memory);
typedef void PlatformClearMemFunc(void *mem, uint32 size);

typedef void PlatformMoveMouseFunc(int32 screenX, int32 screenY);
typedef void PlatformSetMouseWrapFunc(bool wrapMouse);
typedef void PlatformShowMouseFunc();
typedef void PlatformHideMouseFunc();
typedef void PlatformResizeWindow(int32 w, int32 h);

typedef void PlatformSemaSignalFunc(void *semaPtr);
typedef b32 PlatformSemaWaitFunc(void *semaPtr, b32 block);

struct ThreadLocalStorageKey {
#if _WIN32
    uint32 key;
#elif __linux__
    pthread_key_t key;
#elif __APPLE__
#endif
};

typedef ThreadLocalStorageKey PlatformCreateTLSKeyFunc();
typedef void *PlatformGetTLSFunc(ThreadLocalStorageKey key);
typedef void PlatformSetTLSFunc(ThreadLocalStorageKey key, void *data);

typedef real64 PlatformGetTime();

struct PlatformAPI {
    PlatformDebugPrintFunc *DebugPrint;

    PlatformOpenFileForReadFunc *OpenFileForRead;
    PlatformOpenFileForWriteFunc *OpenFileForWrite;
    PlatformResizeFileFunc *ResizeFile;
    PlatformCloseFileFunc *CloseFile;

    PlatformSystemFunc *System;

    PlatformMakeDirectoryFunc *MakeDirectory;
    PlatformRemoveFileFunc *RemoveFile;

    PlatformFindFileInDirectoryFunc *FindFileInDirectory;
    PlatformFileLastModifiedTimeFunc *FileLastModifiedTime;
    PlatformGetAllFilesWithExtensionsFunc *GetAllFilesWithExtensions;

    PlatformAllocateMemFunc *AllocateMem;
    PlatformResizeMemFunc *ResizeMem;
    PlatformCopyMemFunc *CopyMem;
    PlatformMoveMemFunc *MoveMem;
    PlatformDeallocateMemFunc *DeallocateMem;
    PlatformClearMemFunc *ClearMem;

    PlatformMoveMouseFunc *MoveMouse;
    PlatformShowMouseFunc *ShowMouse;
    PlatformHideMouseFunc *HideMouse;
    PlatformSetMouseWrapFunc *SetMouseWrap;

    PlatformResizeWindow *ResizeWindow;

    PlatformSemaSignalFunc *SemaSignal;
    PlatformSemaWaitFunc *SemaWait;

    PlatformCreateTLSKeyFunc *CreateTLSKey;
    PlatformGetTLSFunc *GetTLS;
    PlatformSetTLSFunc *SetTLS;

    PlatformGetTime *GetTime;

    ThreadLocalStorageKey taskSchedulerTLSKey;

    TaskScheduler *mainThreadScheduler;
    TaskScheduler *scheduler;

    void *platformMemory;
};

extern "C" PlatformAPI Plat;

typedef void RegisterPlatformAPIFunc(PlatformAPI *api);
extern "C" DLLEXPORT void RegisterPlatformAPI(PlatformAPI *api);

#endif

