
#define _CRT_SECURE_NO_WARNINGS
#include <windows.h>

#include <stdlib.h>
#include <io.h>
#include <Objbase.h>
	
#include <mmdeviceapi.h>
#include <audioclient.h>

#define GLEW_STATIC
#include <gl/glew.h>
#include <gl/wglew.h>

#include "etcetera.h"
#include "etc_opengl.h"

#include "io/utf8.cpp"


#define ETC_AUDIO_BUFFER_COUNT 4
struct WinAudioOutput {
    IMMDevice *device;
    IAudioClient *audioClient;
    IAudioRenderClient *renderClient;
    HANDLE renderEvent;
    HANDLE audioSema;

    WAVEFORMATEX waveFormat;

    uint32 bufferSampleCount;
    uint32 samplesRendered;
};

struct EtcPlatformMemory {
    EtcGameMemory *gameMemory;

    HMODULE gameLibraryHandle;
    FILETIME gameLibraryLastWriteTime;
    char gameLibraryPath[MAX_PATH];
    char gameLibraryPathTemp[MAX_PATH];

    HWND window;
    HMODULE XInputHandle;

    TRACKMOUSEEVENT trackMouseEvent;
    bool mouseTracked;
    vec2 prevMousePosition;
    vec2 mouseVelocity;

    WinAudioOutput audio;

    TaskScheduler mainThreadScheduler;
    TaskScheduler scheduler;
    HANDLE schedulerSema;

    LARGE_INTEGER PerfCountFrequency;

#if ETC_DEBUG
    EtcDebugMemory *debugMemory;

    HANDLE assetDirectoryReadChangesHandle;
    uint8 assetDirectoryChanges[Kilobytes(512)];

    OVERLAPPED assetDirectoryOverlapped;
#endif
};

static EtcPlatformMemory *GlobalPlatformMemory = NULL;

#include "task.cpp"
#include "platform/windows_input.cpp"
#include "platform/windows_api.cpp"

PlatformAPI Plat = WindowsPlatformAPI();
GameAPI Game = {};

#if ETC_DEBUG
EtcDebugMemory *DebugMem = NULL;
#endif

static b32 etcPlatformRunning = true;

FILETIME GetLastWriteTime(char *path) {
    FILETIME result;

    WIN32_FILE_ATTRIBUTE_DATA Data;
    if(GetFileAttributesEx(path, GetFileExInfoStandard, &Data))
    {
        result = Data.ftLastWriteTime;
    }

    return result;
}

#if !ETC_RELEASE
void LoadGameLibrary(EtcPlatformMemory *platform) {
    // Unload previous handle
    if (platform->gameLibraryHandle != NULL) {
        FreeLibrary(platform->gameLibraryHandle);
    }

    platform->gameLibraryHandle = LoadLibrary(platform->gameLibraryPath);

    RegisterPlatformAPIFunc *registerAPI = (RegisterPlatformAPIFunc *)GetProcAddress(platform->gameLibraryHandle, "RegisterPlatformAPI");
    ASSERT(registerAPI != NULL);
    registerAPI(&Plat);

#if ETC_DEBUG
    RegisterDebugMemoryFunc *registerDebugMemory = (RegisterDebugMemoryFunc *)GetProcAddress(platform->gameLibraryHandle, "RegisterDebugMemory");
    ASSERT(registerDebugMemory);
    registerDebugMemory(DebugMem);
#endif

    ReloadGameAPIFunc *reloadGameAPI = (ReloadGameAPIFunc *)GetProcAddress(platform->gameLibraryHandle, "ReloadGameAPI");
    ASSERT(reloadGameAPI != NULL);
    Game = reloadGameAPI(platform->gameMemory);
}

b32 ReloadGameLibrary(EtcPlatformMemory *platform) {
    b32 reloaded = false;

    FILETIME lastWriteTime = GetLastWriteTime(platform->gameLibraryPath);
    if (CompareFileTime(&platform->gameLibraryLastWriteTime, &lastWriteTime) < 0) {
        if (WaitForSingleObject(platform->audio.audioSema, 0) == WAIT_OBJECT_0) {
            LoadGameLibrary(platform);
            platform->gameLibraryLastWriteTime = GetLastWriteTime(platform->gameLibraryPath);
            ReleaseSemaphore(platform->audio.audioSema, 1, NULL);

            reloaded = true;
        }
    }

    return reloaded;
}
#endif

LRESULT CALLBACK MainWindowCallback(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam) {

    LRESULT result = 0;
    
    switch (msg) {
        case WM_SIZE: {
        } break;

        case WM_DESTROY: {
        } break;

        case WM_QUIT: {
            etcPlatformRunning = false;
        } break;

        case WM_CLOSE: {
            etcPlatformRunning = false;
        } break;

        case WM_ACTIVATEAPP: {
        } break;

        case WM_ACTIVATE: {
        } break;

        default: {
            result = DefWindowProc(hwnd, msg, wParam, lParam);
        } break;
    }

    return result;
}


void InitOpenGL(EtcPlatformMemory *platform) {
    HDC deviceContext = GetDC(platform->window);

    PIXELFORMATDESCRIPTOR desiredPixelFormat = {};
    desiredPixelFormat.nSize = sizeof(desiredPixelFormat);
    desiredPixelFormat.nVersion = 1;
    desiredPixelFormat.iPixelType = PFD_TYPE_RGBA;
    desiredPixelFormat.dwFlags = PFD_SUPPORT_OPENGL|PFD_DRAW_TO_WINDOW|PFD_DOUBLEBUFFER;
    desiredPixelFormat.cColorBits = 32;
    desiredPixelFormat.cAlphaBits = 8;
    desiredPixelFormat.cDepthBits = 24;
    desiredPixelFormat.iLayerType = PFD_MAIN_PLANE;

    int suggestedPixelFormatIndex = ChoosePixelFormat(deviceContext, &desiredPixelFormat);
    PIXELFORMATDESCRIPTOR suggestedPixelFormat;
    DescribePixelFormat(deviceContext, suggestedPixelFormatIndex,
                        sizeof(suggestedPixelFormat), &suggestedPixelFormat);
    SetPixelFormat(deviceContext, suggestedPixelFormatIndex, &suggestedPixelFormat);
    
    HGLRC renderContext = wglCreateContext(deviceContext);
    if (wglMakeCurrent(deviceContext, renderContext)) {

        {
            OpenGLInfo *glInfo = &platform->gameMemory->graphics.opengl;
            glInfo->vendor = glGetString(GL_VENDOR);
            glCheckError();
            glInfo->renderer = glGetString(GL_RENDERER);
            glCheckError();
            glInfo->version = glGetString(GL_VERSION);
            glCheckError();
            glInfo->shadingLanguageVersion = glGetString(GL_SHADING_LANGUAGE_VERSION);
            glCheckError();
            glInfo->extensions = glGetString(GL_EXTENSIONS);
            glCheckError();

            Plat.DebugPrint("%s", glInfo->vendor);
            Plat.DebugPrint("%s", glInfo->renderer);
            Plat.DebugPrint("%s", glInfo->version);
            Plat.DebugPrint("%s", glInfo->shadingLanguageVersion);
            Plat.DebugPrint("%s", glInfo->extensions);
        }

        GLenum err = glewInit();
        if (err == GLEW_OK) {
            {
                ASSERT(WGLEW_ARB_create_context);

                int attribs[] = {
                    WGL_CONTEXT_MAJOR_VERSION_ARB, 4,
                    WGL_CONTEXT_MINOR_VERSION_ARB, 3,
                    WGL_CONTEXT_FLAGS_ARB, WGL_CONTEXT_FORWARD_COMPATIBLE_BIT_ARB,
                    WGL_CONTEXT_PROFILE_MASK_ARB, WGL_CONTEXT_CORE_PROFILE_BIT_ARB,
                    0
                };
                
                HGLRC sharedContext = 0;
                HGLRC modernContext = wglCreateContextAttribsARB(deviceContext, sharedContext, attribs);

                ASSERT(modernContext != NULL);
                wglMakeCurrent(deviceContext, modernContext);
                wglDeleteContext(renderContext);
                renderContext = modernContext;
            }

            if (WGLEW_EXT_swap_control) {
                wglSwapIntervalEXT(1);
            }
        }
        else {
            Plat.DebugPrint((char *)glewGetErrorString(err));
        }
    }
}

void WinRenderOpenGL(EtcPlatformMemory *platform) {

    HDC deviceContext = GetDC(platform->window);

    SwapBuffers(deviceContext);

    ReleaseDC(platform->window, deviceContext);
}

void InitWASAPI(EtcPlatformMemory *platform) {
    IMMDeviceEnumerator *pEnumerator = NULL;
    IMMDevice *pDevice = NULL;
    IAudioClient *pAudioClient = NULL;
    IAudioRenderClient *pRenderClient = NULL;
    HANDLE renderEvent;

    uint32 actualBufferSampleCount;

    REFERENCE_TIME bufferDuration = 1000000;

    HRESULT result;
    result = CoCreateInstance(__uuidof(MMDeviceEnumerator), NULL, CLSCTX_ALL, __uuidof(IMMDeviceEnumerator), (void**)&pEnumerator);
    if (SUCCEEDED(result)) {

        // @TODO: make sure to close this device later!
        result = pEnumerator->GetDefaultAudioEndpoint(eRender, eConsole, &pDevice);
        if (SUCCEEDED(result)) {
            result = pDevice->Activate(__uuidof(IAudioClient), CLSCTX_ALL, NULL, (void**)&pAudioClient);
            if (SUCCEEDED(result)) {

                WAVEFORMATEX *waveFormat = &platform->audio.waveFormat;
                waveFormat->wFormatTag = WAVE_FORMAT_IEEE_FLOAT;
                waveFormat->nChannels = AUDIO_NUM_CHANNELS;
                waveFormat->nSamplesPerSec = AUDIO_SAMPLERATE;
                waveFormat->wBitsPerSample = AUDIO_BITS_PER_SAMPLE;
                waveFormat->nBlockAlign = (waveFormat->nChannels * waveFormat->wBitsPerSample) / 8;
                waveFormat->nAvgBytesPerSec = waveFormat->nSamplesPerSec * waveFormat->nBlockAlign;
                waveFormat->cbSize = 0;

                WAVEFORMATEXTENSIBLE *closestWaveFormat;
                result = pAudioClient->IsFormatSupported(AUDCLNT_SHAREMODE_SHARED, waveFormat, (WAVEFORMATEX **)&closestWaveFormat);

                if (result == S_FALSE) {
                    if (closestWaveFormat->Format.nSamplesPerSec != waveFormat->nSamplesPerSec) {
                        // @TODO: need to resample, there appears to be an API...
                        // https://msdn.microsoft.com/en-us/library/windows/desktop/ff819070(v=vs.85).aspx
                        // https://sourceforge.net/p/playpcmwin/wiki/HowToUseResamplerMFT/
                    }

                    waveFormat = (WAVEFORMATEX *)closestWaveFormat;
                }
                    
                result = pAudioClient->Initialize(AUDCLNT_SHAREMODE_SHARED,
                                                  AUDCLNT_STREAMFLAGS_EVENTCALLBACK,
                                                  bufferDuration, 0, waveFormat, NULL);
                if (SUCCEEDED(result)) {

                    result = pAudioClient->GetBufferSize(&actualBufferSampleCount);
                    if (SUCCEEDED(result)) {

                        result = pAudioClient->GetService(__uuidof(IAudioRenderClient), (void**)&pRenderClient);
                        if (SUCCEEDED(result)) {

                            renderEvent = CreateEvent(NULL, FALSE, FALSE, NULL);
                            result = pAudioClient->SetEventHandle(renderEvent);
                            if (SUCCEEDED(result)) {
                                // @NOTE: success!
                            }
                        }
                    }
                }
            }
        }
    }

    platform->audio.device = pDevice;
    platform->audio.audioClient = pAudioClient;
    platform->audio.renderClient = pRenderClient;
    platform->audio.renderEvent = renderEvent;
    platform->audio.bufferSampleCount = actualBufferSampleCount;

    platform->audio.audioSema = CreateSemaphoreEx(NULL, 1, 1, NULL, 0, SEMAPHORE_ALL_ACCESS);
}

void WASAPIThreadProc(void *data) {
    CoInitialize(NULL);

    EtcPlatformMemory *platform = (EtcPlatformMemory *)data;
    WinAudioOutput *audio = &platform->audio;

    audio->audioClient->Start();

    uint32 framesMissed = 0;

    while (platform->gameMemory->running) {
        WaitForSingleObject(audio->renderEvent, INFINITE);

        uint32 numFramesPadding;
        if (SUCCEEDED(audio->audioClient->GetCurrentPadding(&numFramesPadding))) {
            uint32 numFramesAvailable = audio->bufferSampleCount - numFramesPadding;

            if (numFramesAvailable == 0) {
                continue;
            }

            real32 *buffer;
            if (SUCCEEDED(audio->renderClient->GetBuffer(numFramesAvailable, (BYTE **)&buffer))) {

                if (WaitForSingleObject(audio->audioSema, 0) == WAIT_OBJECT_0) {

                    while (framesMissed > 0) {
                        uint32 framesToRender = Min(numFramesAvailable, framesMissed);
                        Game.WriteSoundSamples(platform->gameMemory, framesToRender, buffer);
                        framesMissed -= framesToRender;
                    }
                    
                    Game.WriteSoundSamples(platform->gameMemory, numFramesAvailable, buffer);
                    ReleaseSemaphore(audio->audioSema, 1, NULL);
                }
                else {
                    Plat.ClearMem(buffer, sizeof(real32) * 2 * numFramesAvailable);
                    framesMissed += numFramesAvailable;
                }

                audio->renderClient->ReleaseBuffer(numFramesAvailable, 0);
            }
        }
    }

    CoUninitialize();
}

void StartWASAPIThread(EtcPlatformMemory *platform) {
    DWORD threadID;
    HANDLE threadHandle = CreateThread(0, 0, (LPTHREAD_START_ROUTINE)WASAPIThreadProc, platform, 0, &threadID);
    CloseHandle(threadHandle);
}

int CALLBACK WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow) {
    CoInitialize(NULL);

    // Initialization and stuff
    EtcPlatformMemory platform = {};
    GlobalPlatformMemory = &platform;
    
    platform.gameMemory = (EtcGameMemory *)Plat.AllocateMem(sizeof(EtcGameMemory));
    Plat.ClearMem(platform.gameMemory, sizeof(EtcGameMemory));

    AllocateTaskScheduler(&platform.mainThreadScheduler, 1024);

    // @GACK: look at this disgusting max count!
    AllocateTaskScheduler(&platform.scheduler, 4096);
    platform.schedulerSema = CreateSemaphoreEx(NULL, 0, 1 << 31 - 1, NULL, 0, SEMAPHORE_ALL_ACCESS);
    ASSERT(platform.schedulerSema);

    platform.scheduler.semaPtr = &platform.schedulerSema;

    Plat.mainThreadScheduler = &platform.mainThreadScheduler;
    Plat.scheduler = &platform.scheduler;


    // Determine the locations of the game code
    char executablePath[MAX_PATH];

    GetModuleFileName(NULL, executablePath, sizeof(executablePath));
    char drive[MAX_PATH];
    char dir[MAX_PATH];
    _splitpath(executablePath, drive, dir, NULL, NULL);

    char newWorkingDirectory[MAX_PATH];
    wsprintf(newWorkingDirectory, "%s%s..\\", drive, dir);
    SetCurrentDirectory(newWorkingDirectory);


    wsprintf(platform.gameLibraryPath, "%s%s%s", drive, dir, "etcetera.dll");
    wsprintf(platform.gameLibraryPathTemp, "%s%s%s", drive, dir, "etcetera_temp.dll");

    WNDCLASS windowClass = {};
    windowClass.lpszClassName = "EtcWindowClass";

    windowClass.style = CS_HREDRAW | CS_VREDRAW;
    windowClass.lpfnWndProc = MainWindowCallback;
    windowClass.hInstance = hInstance;
    windowClass.hbrBackground = (HBRUSH)GetStockObject(BLACK_BRUSH);
    if (RegisterClass(&windowClass)) {

        // Start the window as zero-size, then resize once preferences
        // have been read by the game
        RECT windowRect;
        windowRect.left = 0;
        windowRect.right = 0;
        windowRect.top = 0;
        windowRect.bottom = 0;

        DWORD dwExStyle = WS_EX_APPWINDOW | WS_EX_WINDOWEDGE;
        DWORD dwStyle = WS_OVERLAPPEDWINDOW & ~WS_SIZEBOX & ~WS_MAXIMIZEBOX;

        AdjustWindowRectEx(&windowRect, dwStyle, false, dwExStyle);
        
        HWND window = CreateWindowEx(dwExStyle,
                                     windowClass.lpszClassName,
                                     "Breakout",
                                     WS_CLIPSIBLINGS | WS_CLIPCHILDREN | dwStyle,
                                     CW_USEDEFAULT,
                                     CW_USEDEFAULT,
                                     (windowRect.right - windowRect.left),
                                     (windowRect.bottom - windowRect.top),
                                     NULL,
                                     NULL,
                                     hInstance,
                                     NULL);

        platform.window = window;

        if (window != NULL) {

            LARGE_INTEGER systemTime;
            
            QueryPerformanceFrequency(&platform.PerfCountFrequency);

            InitOpenGL(&platform);
            
            InitWASAPI(&platform);

            LoadGameLibrary(&platform);
            platform.gameLibraryLastWriteTime = GetLastWriteTime(platform.gameLibraryPath);

            // Initialize Game Library now that the window has loaded
            Game.InitEtcGameMemory(platform.gameMemory);

#if ETC_DEBUG
            GetDebugMemoryFunc *getDebugMemory = (GetDebugMemoryFunc *)GetProcAddress(platform.gameLibraryHandle, "GetDebugMemory");
            ASSERT(getDebugMemory);
            DebugMem = getDebugMemory();
#endif

            InitInputDevices(platform.gameMemory, &platform.XInputHandle);

            platform.trackMouseEvent = {};
            platform.trackMouseEvent.cbSize = sizeof(TRACKMOUSEEVENT);
            platform.trackMouseEvent.dwFlags = TME_LEAVE;
            platform.trackMouseEvent.hwndTrack = platform.window;

            bool trackedSuccess = TrackMouseEvent(&platform.trackMouseEvent);
            ASSERT(trackedSuccess);
            platform.mouseTracked = true;

#if ETC_DEBUG
            // Setup asset hotloading
            const char *assetSearchPath = platform.gameMemory->assets.searchPath;

            platform.assetDirectoryReadChangesHandle = CreateFile(assetSearchPath,
                                                                  GENERIC_READ,
                                                                  FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE,
                                                                  NULL,
                                                                  OPEN_EXISTING,
                                                                  FILE_FLAG_BACKUP_SEMANTICS | FILE_FLAG_OVERLAPPED,
                                                                  NULL);

            ReadDirectoryChangesW(platform.assetDirectoryReadChangesHandle,
                                  platform.assetDirectoryChanges,
                                  sizeof(platform.assetDirectoryChanges),
                                  TRUE,
                                  FILE_NOTIFY_CHANGE_LAST_WRITE,
                                  NULL,
                                  &platform.assetDirectoryOverlapped,
                                  NULL);
#endif


            // Launch worker threads
#define NUM_THREADS 4
            for (int i = 0; i < NUM_THREADS; i++) {
                DWORD threadID;
                HANDLE threadHandle = CreateThread(0, 0, (LPTHREAD_START_ROUTINE)PlatformTaskSchedulerThreadProc, &platform.scheduler, 0, &threadID);
                CloseHandle(threadHandle);
            }

            AllocateTLSData(&platform.gameMemory->mainThreadTaskSchedulerTLSData);

            
            // Resize window based off of loaded user preferences
            RECT windowRect;
            windowRect.left = 0;
            windowRect.right = platform.gameMemory->userPreferences.screenWidth;
            windowRect.top = 0;
            windowRect.bottom = platform.gameMemory->userPreferences.screenHeight;

            AdjustWindowRectEx(&windowRect, dwStyle, false, dwExStyle);
            int adjustedWidth = windowRect.right - windowRect.left;
            int adjustedHeight = windowRect.bottom - windowRect.top;

            int screenWidth = GetSystemMetrics(SM_CXSCREEN);
            int screenHeight = GetSystemMetrics(SM_CYSCREEN);

            SetWindowPos(window, HWND_NOTOPMOST,
                         (screenWidth - adjustedWidth) / 2,
                         (screenHeight - adjustedHeight) / 2,
                         adjustedWidth, adjustedHeight,
                         SWP_NOZORDER);

            ShowWindow(window, SW_SHOW);
            SetForegroundWindow(window);


            // Force an arrow cursor (still spins for a bit, tho!)
            HCURSOR cursor = LoadCursor(NULL, IDC_ARROW);
            SetCursor(cursor);


            QueryPerformanceCounter(&systemTime);

            platform.gameMemory->running = true;
            StartWASAPIThread(&platform);

            // GameLoop
            while (platform.gameMemory->running) {
#if !ETC_RELEASE
                ReloadGameLibrary(&platform);
#endif

#if ETC_DEBUG
                // Asset hotloading notifications
                DWORD bytesReturned;
                if (GetOverlappedResult(platform.assetDirectoryReadChangesHandle,
                                        &platform.assetDirectoryOverlapped,
                                        &bytesReturned,
                                        false)) {

                    FILE_NOTIFY_INFORMATION *info = (FILE_NOTIFY_INFORMATION *)platform.assetDirectoryChanges;
                    char utf8FilePath[MAX_PATH];

                    while (true) {
                        ucs16_toutf8((uint16 *)info->FileName, info->FileNameLength / sizeof(WCHAR),
                                     (uint8 *)utf8FilePath, sizeof(utf8FilePath));

                        const char *path = PathConcat(platform.gameMemory->assets.searchPath, utf8FilePath);
                        RegisterModifiedAssetPath(&platform.gameMemory->assets, path);

                        if (info->NextEntryOffset == 0) {
                            break;
                        }

                        info = (FILE_NOTIFY_INFORMATION *)((uint8 *)info + info->NextEntryOffset);
                    }

                    ReadDirectoryChangesW(platform.assetDirectoryReadChangesHandle,
                                          platform.assetDirectoryChanges,
                                          sizeof(platform.assetDirectoryChanges),
                                          TRUE,
                                          FILE_NOTIFY_CHANGE_LAST_WRITE,
                                          NULL,
                                          &platform.assetDirectoryOverlapped,
                                          NULL);
                }
#endif

                LARGE_INTEGER prevSystemTime = systemTime;
                QueryPerformanceCounter(&systemTime);
                real64 deltaTime = (real64)(systemTime.QuadPart - prevSystemTime.QuadPart) / (real64)platform.PerfCountFrequency.QuadPart;

                platform.gameMemory->platformDeltaTime = Min(0.1, deltaTime);

                // Run some main thread queued events
                CompleteNextTasks(Plat.mainThreadScheduler, 4);

                // Process Input
                WinPushKeyboardMouseEvents(&platform, platform.gameMemory, &platform.gameMemory->inputQueue, &platform.gameMemory->keyboard, &platform.gameMemory->mouse);

                for (int i = 0; i < platform.gameMemory->gamepadCount; i++) {
                    WinPushGamepadEvents(&platform.gameMemory->inputQueue, &platform.gameMemory->gamepads[i]);
                    WinSetInputDeviceOutputState(&platform.gameMemory->gamepads[i]);
                }

                platform.gameMemory->running = etcPlatformRunning;
                Game.UpdateAndRender(platform.gameMemory);

                WinRenderOpenGL(&platform);
            }
        }
    }
    

    // Free Memory
    platform.audio.audioClient->Stop();

    Game.DeinitEtcGameMemory(platform.gameMemory);
    Plat.DeallocateMem(platform.gameMemory);

    CoUninitialize();
    
    return 0;
}

