
#include <xinput.h>
#include <WindowsX.h>

// [0]: https://msdn.microsoft.com/en-us/library/windows/desktop/dd375731(v=vs.85).aspx
static int32 WinVirtualKeyMap[] = {
    -1,
    -1, // VK_LBUTTON
    -1, // VK_RBUTTON
    -1, // VK_CANCEL
    -1, // VK_MBUTTON
    -1, // VK_XBUTTON1
    -1, // VK_XBUTTON2

    -1, // Undefined
    Etc_Backspace, // VK_BACK
    Etc_Tab, // VK_TAB

    // Reserved
    -1, -1,
    -1, // VK_CLEAR

    Etc_Return, // VK_RETURN

    // Undefined
    -1, -1, 

    Etc_Shift, // VK_SHIFT
    Etc_Control, // VK_CONTROL
    Etc_Alt, // VK_MENU

    -1, // VK_PAUSE
    -1, // VK_CAPITAL
    -1, // VK_KANA, VK_HANGUEL VK_HANGUL
    -1, // Undefined
    -1, // VK_JUNJA
    -1, // VK_FINAL
    -1, // VK_HANJA, VK_KANJI

    -1, // Undefined
    Etc_Escape, // VK_ESCAPE

    -1, // VK_CONVERT
    -1, //VK_NONCONVERT
    -1, //VK_ACCEPT
    -1, //VK_MODECHANGE

    Etc_Space, // VK_SPACE
    -1, // VK_PRIOR
    -1, // VK_NEXT
    Etc_End, // VK_END
    Etc_Home, // VK_HOME
    Etc_LeftArrow, // VK_LEFT
    Etc_UpArrow, // VK_UP
    Etc_RightArrow, // VK_RIGHT
    Etc_DownArrow, // VK_DOWN
    -1, // VK_SELECT
    -1, // VK_PRINT
    -1, // VK_EXECUTE
    -1, // VK_SNAPSHOT
    Etc_Insert, // VK_INSERT
    Etc_ForwardDelete, // VK_DELETE
    Etc_Help, // VK_HELP
    Etc_0,
    Etc_1,
    Etc_2,
    Etc_3,
    Etc_4,
    Etc_5,
    Etc_6,
    Etc_7,
    Etc_8,
    Etc_9,

    // Undefined
    -1, -1, -1, -1, -1, -1, -1,

    Etc_A,
    Etc_B,
    Etc_C,
    Etc_D,
    Etc_E,
    Etc_F,
    Etc_G,
    Etc_H,
    Etc_I,
    Etc_J,
    Etc_K,
    Etc_L,
    Etc_M,
    Etc_N,
    Etc_O,
    Etc_P,
    Etc_Q,
    Etc_R,
    Etc_S,
    Etc_T,
    Etc_U,
    Etc_V,
    Etc_W,
    Etc_X,
    Etc_Y,
    Etc_Z,
    Etc_Win, // VK_LWIN
    Etc_RightWin, // VK_RWIN
    -1, // VK_APPS
    -1, // Reserved
    -1, // VK_SLEEP
    -1, // VK_NUMPAD0
    -1, // VK_NUMPAD1
    -1, // VK_NUMPAD2
    -1, // VK_NUMPAD3
    -1, // VK_NUMPAD4
    -1, // VK_NUMPAD5
    -1, // VK_NUMPAD6
    -1, // VK_NUMPAD7
    -1, // VK_NUMPAD8
    -1, // VK_NUMPAD9
    -1, // VK_MULTIPLY
    -1, // VK_ADD
    -1, // VK_SEPARATOR
    -1, // VK_SUBTRACT
    -1, // VK_DECIMAL
    -1, // VK_DIVIDE
    Etc_F1,
    Etc_F2,
    Etc_F3,
    Etc_F4,
    Etc_F5,
    Etc_F6,
    Etc_F7,
    Etc_F8,
    Etc_F9,
    Etc_F10,
    Etc_F11,
    Etc_F12,
    Etc_F13,
    Etc_F14,
    Etc_F15,
    Etc_F16,
    Etc_F17,
    Etc_F18,
    Etc_F19,
    Etc_F20,
    Etc_F21,
    Etc_F22,
    Etc_F23,
    Etc_F24,

    // Unassigned
    -1, -1, -1, -1, -1, -1, -1, -1,

    -1, // VK_NUMLOCK
    -1, // VK_SCROLL

    // OEM specific
    -1, -1, -1, -1, -1,

    // Unassigned
    -1, -1, -1, -1, -1, -1, -1, -1, -1,

    Etc_Shift, // VK_LSHIFT
    Etc_RightShift, // VK_RSHIFT
    Etc_Control, // VK_LCONTROL
    Etc_RightControl, // VK_RCONTROL

    Etc_Alt, // VK_LMENU
    Etc_RightAlt, // VK_RMENU
    
    -1, // VK_BROWSER_BACK
    -1, // VK_BROWSER_FORWARD
    -1, // VK_BROWSER_REFRESH
    -1, // VK_BROWSER_STOP
    -1, // VK_BROWSER_SEARCH
    -1, // VK_BROWSER_FAVORITES
    -1, // VK_BROWSER_HOME
    -1, // VK_VOLUME_MUTE
    -1, // VK_VOLUME_DOWN
    -1, // VK_VOLUME_UP
    -1, // VK_MEDIA_NEXT_TRACK
    -1, // VK_MEDIA_PREV_TRACK
    -1, // VK_MEDIA_STOP
    -1, // VK_MEDIA_PLAY_PAUSE
    -1, // VK_LAUNCH_MAIL
    -1, // VK_LAUNCH_MEDIA_SELECT
    -1, // VK_LAUNCH_APP1
    -1, // VK_LAUNCH_APP2

    // Reserved
    -1, -1,

    Etc_Semicolon, // VK_OEM_1
    Etc_Equal, // VK_OEM_PLUS
    Etc_Comma, // VK_OEM_COMMA
    Etc_Minus, // VK_OEM_MINUS
    Etc_Period, // VK_OEM_PERIOD
    Etc_Slash, // VK_OEM_2
    Etc_Tick, // VK_OEM_3

    // Reserved
    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 1, -1, -1, -1, -1, -1, -1, -1, -1,
    
    // Unassigned
    -1, -1, -1,

    Etc_LeftBracket, // VK_OEM_4
    Etc_Backslash, // VK_OEM_5
    Etc_RightBracket, // VK_OEM_6
    Etc_Quote, // VK_OEM_7

    // VK_OEM_8...
};


typedef DWORD WINAPI xinput_get_state(DWORD dwUserIndex, XINPUT_STATE *state);
DWORD WINAPI XInputGetStateStub(DWORD dwUserIndex, XINPUT_STATE *state) {
    return ERROR_DEVICE_NOT_CONNECTED;
}
xinput_get_state *XInputGetState_ = XInputGetStateStub;
#define XInputGetState XInputGetState_

typedef DWORD WINAPI xinput_set_state(DWORD dwUserIndex, XINPUT_VIBRATION *vibration);
DWORD WINAPI XInputSetStateStub(DWORD dwUserIndex, XINPUT_VIBRATION *state) {
    return ERROR_DEVICE_NOT_CONNECTED;
}
xinput_set_state *XInputSetState_ = XInputSetStateStub;
#define XInputSetState XInputSetState_


void InitInputDevices(EtcGameMemory *gameMemory, HMODULE *XInputHandle) {

    // Initialize XInput
    *XInputHandle = LoadLibrary("xinput1_4.dll");
    if (*XInputHandle == NULL) {
        *XInputHandle = LoadLibrary("xinput9_1_0.dll");
    }
    if (*XInputHandle == NULL) {
        *XInputHandle = LoadLibrary("xinput1_3.dll");
    }

    if (*XInputHandle != NULL) {
        XInputGetState_ = (xinput_get_state *)GetProcAddress(*XInputHandle, "XInputGetState");
        XInputSetState_ = (xinput_set_state *)GetProcAddress(*XInputHandle, "XInputSetState");
    }


    // Allocate EtcInputDevices
    EtcInputDevice *keyboard = &gameMemory->keyboard;
    EtcInputDevice *mouse = &gameMemory->mouse;

    gameMemory->gamepadCount = 4;
    EtcInputDevice *gamepads = gameMemory->gamepads;
    
    {
        keyboard->type = Etc_Keyboard;
        keyboard->connected = true;

        keyboard->lastEventIndex = -1;

        keyboard->discreteCount = Etc_KeyboardDiscreteCount;
        keyboard->analogueCount = 0;


        keyboard->pressed = PushArray(&gameMemory->permanentArena, bool, Etc_KeyboardDiscreteCount);
        keyboard->hold = PushArray(&gameMemory->permanentArena, bool, Etc_KeyboardDiscreteCount);
        keyboard->released = PushArray(&gameMemory->permanentArena, bool, Etc_KeyboardDiscreteCount);

        keyboard->framesSinceChange = PushArray(&gameMemory->permanentArena, u32, Etc_KeyboardDiscreteCount);
        keyboard->secondsSinceChange = PushArray(&gameMemory->permanentArena, r32, Etc_KeyboardDiscreteCount);

        keyboard->prevAnalogue = PushArray(&gameMemory->permanentArena, r32, keyboard->analogueCount);
        keyboard->analogue = PushArray(&gameMemory->permanentArena, r32, keyboard->analogueCount);
    }

    {
        mouse->type = Etc_Mouse;
        mouse->connected = true;

        mouse->lastEventIndex = -1;

        mouse->discreteCount = Etc_MouseDiscreteCount;
        mouse->analogueCount = Etc_MouseAnalogueCount;

        mouse->pressed = PushArray(&gameMemory->permanentArena, bool, Etc_MouseDiscreteCount);
        mouse->hold = PushArray(&gameMemory->permanentArena, bool, Etc_MouseDiscreteCount);
        mouse->released = PushArray(&gameMemory->permanentArena, bool, Etc_MouseDiscreteCount);

        mouse->framesSinceChange = PushArray(&gameMemory->permanentArena, u32, Etc_MouseDiscreteCount);
        mouse->secondsSinceChange = PushArray(&gameMemory->permanentArena, r32, Etc_MouseDiscreteCount);

        mouse->prevAnalogue = PushArray(&gameMemory->permanentArena, r32, mouse->analogueCount);
        mouse->analogue = PushArray(&gameMemory->permanentArena, r32, mouse->analogueCount);

        
    }

    for (int i = 0; i < gameMemory->gamepadCount; i++) {
        gamepads[i].type = Etc_Gamepad;
        gamepads[i].id = i;

        gamepads[i].connected = false;
        gamepads[i].lastEventIndex = -1;

        gamepads[i].discreteCount = Etc_GamepadDiscreteCount;
        gamepads[i].analogueCount = Etc_GamepadAnalogueCount;
        gamepads[i].outputCount = Etc_GamepadOutputCount;

        gamepads[i].pressed = PushArray(&gameMemory->permanentArena, bool, Etc_GamepadDiscreteCount);
        gamepads[i].hold = PushArray(&gameMemory->permanentArena, bool, Etc_GamepadDiscreteCount);
        gamepads[i].released = PushArray(&gameMemory->permanentArena, bool, Etc_GamepadDiscreteCount);

        gamepads[i].framesSinceChange = PushArray(&gameMemory->permanentArena, u32, Etc_GamepadDiscreteCount);
        gamepads[i].secondsSinceChange = PushArray(&gameMemory->permanentArena, r32, Etc_GamepadDiscreteCount);

        gamepads[i].prevAnalogue = PushArray(&gameMemory->permanentArena, r32, Etc_GamepadAnalogueCount);
        gamepads[i].analogue = PushArray(&gameMemory->permanentArena, r32, Etc_GamepadAnalogueCount);

        gamepads[i].output = PushArray(&gameMemory->permanentArena, r32, Etc_GamepadOutputCount);
    }
}

// [0]: https://stackoverflow.com/questions/15966642/how-do-you-tell-lshift-apart-from-rshift-in-wm-keydown-events
WPARAM MapLeftRightKeys(WPARAM vk, LPARAM lParam) {
    WPARAM newVK = vk;
    UINT scancode = (lParam & 0x00ff0000) >> 16;
    int extended  = (lParam & 0x01000000) != 0;

    switch (vk) {
        case VK_SHIFT: {
            newVK = MapVirtualKey(scancode, MAPVK_VSC_TO_VK_EX);
        } break;
        case VK_CONTROL: {
            newVK = extended ? VK_RCONTROL : VK_LCONTROL;
        } break;
        case VK_MENU: {
            newVK = extended ? VK_RMENU : VK_LMENU;
        } break;
        default: {
            // Not a key we map from generic to left/right specialized, sojust return it.
            newVK = vk;
        } break;
    }

    return newVK;
}

void WinPushKeyboardMouseEvents(EtcPlatformMemory *platform, EtcGameMemory *mem, EtcInputEventQueue *queue, EtcInputDevice *keyboard, EtcInputDevice *mouse) {
    MSG msg;
    if (!platform->mouseTracked) {
        POINT point = {};
        GetCursorPos(&point);

        ScreenToClient(platform->window, &point);

        real32 posX = (real32)point.x;
        real32 posY = mem->graphics.height - point.y;

        real32 posXNorm = posX / mem->graphics.width;
        real32 posYNorm = posY / mem->graphics.height;
        
        {
            EtcInputEvent event;
            event.device = mouse;
            event.type = InputEvent_Analogue;
            event.index = Etc_MousePositionX;
            event.analogueValue = posX;

            EtcInputEventQueuePush(queue, event);
        }

        {
            EtcInputEvent event;
            event.device = mouse;
            event.type = InputEvent_Analogue;
            event.index = Etc_MousePositionY;
            event.analogueValue = posY;

            EtcInputEventQueuePush(queue, event);
        }

        {
            EtcInputEvent event;
            event.device = mouse;
            event.type = InputEvent_Analogue;
            event.index = Etc_MousePositionXNorm;
            event.analogueValue = posXNorm;

            EtcInputEventQueuePush(queue, event);
        }

        {
            EtcInputEvent event;
            event.device = mouse;
            event.type = InputEvent_Analogue;
            event.index = Etc_MousePositionYNorm;
            event.analogueValue = posYNorm;

            EtcInputEventQueuePush(queue, event);
        }
    }

    while (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE)) {
        switch (msg.message) {
            case WM_KEYDOWN: {
                if (msg.wParam < ARRAY_LENGTH(int32, WinVirtualKeyMap)) {
                    EtcInputEvent event;
                    event.device = keyboard;
                    event.type = InputEvent_Discrete;
                    event.index = WinVirtualKeyMap[MapLeftRightKeys(msg.wParam, msg.lParam)];
                    event.discreteValue = true;

                    EtcInputEventQueuePush(queue, event);
                }

                // Translate the message so WM_CHAR events are generated
                TranslateMessage(&msg);
            } break;

            case WM_KEYUP: {
                if (msg.wParam < ARRAY_LENGTH(int32, WinVirtualKeyMap)) {
                    EtcInputEvent event;
                    event.device = keyboard;
                    event.type = InputEvent_Discrete;
                    event.index = WinVirtualKeyMap[MapLeftRightKeys(msg.wParam, msg.lParam)];
                    event.discreteValue = false;

                    EtcInputEventQueuePush(queue, event);
                }
            } break;

            case WM_SYSKEYDOWN: {
                if (msg.wParam < ARRAY_LENGTH(int32, WinVirtualKeyMap)) {
                    EtcInputEvent event;
                    event.device = keyboard;
                    event.type = InputEvent_Discrete;
                    event.index = WinVirtualKeyMap[msg.wParam];
                    event.discreteValue = true;

                    EtcInputEventQueuePush(queue, event);
                }
            } break;

            case WM_SYSKEYUP: {
                if (msg.wParam < ARRAY_LENGTH(int32, WinVirtualKeyMap)) {
                    EtcInputEvent event;
                    event.device = keyboard;
                    event.type = InputEvent_Discrete;
                    event.index = WinVirtualKeyMap[msg.wParam];
                    event.discreteValue = false;

                    EtcInputEventQueuePush(queue, event);
                }
            } break;

            case WM_CHAR: {
                if (msg.wParam > 0 && msg.wParam < 0x10000) {
                    uint16 character = (uint16)msg.wParam;

                    char str[2];
                    str[0] = character;
                    str[1] = 0;

                    AddInputCharacter(Game.uiManager, character);
                    DebugAddInputCharacters(DebugMem, str);
                    DebugAddScriptInputCharacters(DebugMem, str);
                }
            } break;

            case WM_MOUSELEAVE: {
                platform->mouseTracked = false;
            } break;

            case WM_MOUSEMOVE: {
                if (!platform->mouseTracked) {
                    TrackMouseEvent(&platform->trackMouseEvent);
                    platform->mouseTracked = true;
                }

                real32 posX = (real32)GET_X_LPARAM(msg.lParam); 
                real32 posY = mem->graphics.height - (real32)GET_Y_LPARAM(msg.lParam);

                real32 posXNorm = posX / mem->graphics.width;
                real32 posYNorm = posY / mem->graphics.height;

                {
                    EtcInputEvent event;
                    event.device = mouse;
                    event.type = InputEvent_Analogue;
                    event.index = Etc_MousePositionX;
                    event.analogueValue = posX;

                    EtcInputEventQueuePush(queue, event);
                }

                {
                    EtcInputEvent event;
                    event.device = mouse;
                    event.type = InputEvent_Analogue;
                    event.index = Etc_MousePositionY;
                    event.analogueValue = posY;

                    EtcInputEventQueuePush(queue, event);
                }

                {
                    EtcInputEvent event;
                    event.device = mouse;
                    event.type = InputEvent_Analogue;
                    event.index = Etc_MousePositionXNorm;
                    event.analogueValue = posXNorm;

                    EtcInputEventQueuePush(queue, event);
                }

                {
                    EtcInputEvent event;
                    event.device = mouse;
                    event.type = InputEvent_Analogue;
                    event.index = Etc_MousePositionYNorm;
                    event.analogueValue = posYNorm;

                    EtcInputEventQueuePush(queue, event);
                }
                
            } break;

            case WM_MOUSEWHEEL: {
                EtcInputEvent event;
                event.device = mouse;
                event.type = InputEvent_Analogue;
                event.index = Etc_ScrollDirection;
                event.analogueValue = GET_WHEEL_DELTA_WPARAM(msg.wParam) / 120.0f;

                EtcInputEventQueuePush(queue, event);
            } break;

            case WM_LBUTTONDOWN: {
                SetCapture(platform->window);

                EtcInputEvent event;
                event.device = mouse;
                event.type = InputEvent_Discrete;
                event.index = Etc_MouseLeft;
                event.discreteValue = true;

                EtcInputEventQueuePush(queue, event);
            } break;

            case WM_RBUTTONDOWN: {
                SetCapture(platform->window);

                EtcInputEvent event;
                event.device = mouse;
                event.type = InputEvent_Discrete;
                event.index = Etc_MouseRight;
                event.discreteValue = true;

                EtcInputEventQueuePush(queue, event);
            } break;

            case WM_MBUTTONDOWN: {
                SetCapture(platform->window);

                EtcInputEvent event;
                event.device = mouse;
                event.type = InputEvent_Discrete;
                event.index = Etc_MouseMiddle;
                event.discreteValue = true;

                EtcInputEventQueuePush(queue, event);
            } break;

            case WM_LBUTTONUP: {
                ReleaseCapture();

                EtcInputEvent event;
                event.device = mouse;
                event.type = InputEvent_Discrete;
                event.index = Etc_MouseLeft;
                event.discreteValue = false;

                EtcInputEventQueuePush(queue, event);
            } break;

            case WM_RBUTTONUP: {
                ReleaseCapture();

                EtcInputEvent event;
                event.device = mouse;
                event.type = InputEvent_Discrete;
                event.index = Etc_MouseRight;
                event.discreteValue = false;

                EtcInputEventQueuePush(queue, event);
            } break;

            case WM_MBUTTONUP: {
                ReleaseCapture();

                EtcInputEvent event;
                event.device = mouse;
                event.type = InputEvent_Discrete;
                event.index = Etc_MouseMiddle;
                event.discreteValue = false;

                EtcInputEventQueuePush(queue, event);
            } break;


            default: {
                TranslateMessage(&msg);
                DispatchMessage(&msg);
            } break;
        }
    }
}

void WinPushGamepadEvents(EtcInputEventQueue *queue, EtcInputDevice *device) {
    XINPUT_STATE state;
    DWORD error = XInputGetState(device->id, &state);

    if (error == ERROR_DEVICE_NOT_CONNECTED) {
        // Clear the device if it just disconnected
        if (device->connected) {
            memset(device->pressed, 0, device->discreteCount * sizeof(b32));
            memset(device->hold, 0, device->discreteCount * sizeof(b32));
            memset(device->released, 0, device->discreteCount * sizeof(b32));
            memset(device->framesSinceChange, 0, device->discreteCount * sizeof(uint32));
            memset(device->secondsSinceChange, 0, device->discreteCount * sizeof(real32));

            memset(device->analogue, 0, device->analogueCount * sizeof(real32));
            memset(device->output, 0, device->outputCount * sizeof(real32));
        }

        device->connected = false;
        return;
    }
    else if (error != ERROR_SUCCESS) {
        return;
    }

    device->connected = true;

    XINPUT_GAMEPAD gamepad = state.Gamepad;

    EtcInputEventQueuePush(queue, {device, InputEvent_Discrete, Etc_FaceA, gamepad.wButtons & XINPUT_GAMEPAD_A});
    EtcInputEventQueuePush(queue, {device, InputEvent_Discrete, Etc_FaceB, gamepad.wButtons & XINPUT_GAMEPAD_B});
    EtcInputEventQueuePush(queue, {device, InputEvent_Discrete, Etc_FaceX, gamepad.wButtons & XINPUT_GAMEPAD_X});
    EtcInputEventQueuePush(queue, {device, InputEvent_Discrete, Etc_FaceY, gamepad.wButtons & XINPUT_GAMEPAD_Y});


    EtcInputEventQueuePush(queue, {device, InputEvent_Discrete, Etc_DirLeft, gamepad.wButtons & XINPUT_GAMEPAD_DPAD_LEFT});
    EtcInputEventQueuePush(queue, {device, InputEvent_Discrete, Etc_DirRight, gamepad.wButtons & XINPUT_GAMEPAD_DPAD_RIGHT});
    EtcInputEventQueuePush(queue, {device, InputEvent_Discrete, Etc_DirUp, gamepad.wButtons & XINPUT_GAMEPAD_DPAD_UP});
    EtcInputEventQueuePush(queue, {device, InputEvent_Discrete, Etc_DirDown, gamepad.wButtons & XINPUT_GAMEPAD_DPAD_DOWN});


    EtcInputEventQueuePush(queue, {device, InputEvent_Discrete, Etc_LeftShoulder, gamepad.wButtons & XINPUT_GAMEPAD_LEFT_SHOULDER});
    EtcInputEventQueuePush(queue, {device, InputEvent_Discrete, Etc_RightShoulder, gamepad.wButtons & XINPUT_GAMEPAD_RIGHT_SHOULDER});

    EtcInputEventQueuePush(queue, {device, InputEvent_Discrete, Etc_LeftStick, gamepad.wButtons & XINPUT_GAMEPAD_LEFT_THUMB});
    EtcInputEventQueuePush(queue, {device, InputEvent_Discrete, Etc_RightStick, gamepad.wButtons & XINPUT_GAMEPAD_RIGHT_THUMB});

    EtcInputEventQueuePush(queue, {device, InputEvent_Discrete, Etc_Start, gamepad.wButtons & XINPUT_GAMEPAD_START});
    EtcInputEventQueuePush(queue, {device, InputEvent_Discrete, Etc_Select, gamepad.wButtons & XINPUT_GAMEPAD_BACK});


    vec2 left = V2(LinearRemap(-32768, 32767, gamepad.sThumbLX, -1.0f, 1.0f),
                   LinearRemap(-32768, 32767, gamepad.sThumbLY, -1.0f, 1.0f));

    vec2 right = V2(LinearRemap(-32768, 32767, gamepad.sThumbRX, -1.0f, 1.0f),
                    LinearRemap(-32768, 32767, gamepad.sThumbRY, -1.0f, 1.0f));

    left = InputDeadzone(left, DEADZONE_STICK_MIN, DEADZONE_STICK_MAX);
    right = InputDeadzone(right, DEADZONE_STICK_MIN, DEADZONE_STICK_MAX);

    EtcInputEventQueuePush(queue, {device, InputEvent_Analogue, Etc_LeftX, 0, left.x});
    EtcInputEventQueuePush(queue, {device, InputEvent_Analogue, Etc_LeftY, 0, left.y});
    EtcInputEventQueuePush(queue, {device, InputEvent_Analogue, Etc_RightX, 0, right.x});
    EtcInputEventQueuePush(queue, {device, InputEvent_Analogue, Etc_RightY, 0, right.y});

    r32 leftTrigger = InputDeadzone(gamepad.bLeftTrigger / 255.0f, DEADZONE_TRIGGER_MIN, DEADZONE_TRIGGER_MAX);
    r32 rightTrigger = InputDeadzone(gamepad.bRightTrigger / 255.0f, DEADZONE_TRIGGER_MIN, DEADZONE_TRIGGER_MAX);

    EtcInputEventQueuePush(queue, {device, InputEvent_Analogue, Etc_LeftTrigger, 0, leftTrigger});
    EtcInputEventQueuePush(queue, {device, InputEvent_Analogue, Etc_RightTrigger, 0, rightTrigger});
}

void WinSetInputDeviceOutputState(EtcInputDevice *device) {
    XINPUT_VIBRATION vibration;
    vibration.wLeftMotorSpeed = (uint16)(device->output[Etc_LowFrequencyRumble] * 65535);
    vibration.wRightMotorSpeed = (uint16)(device->output[Etc_HighFrequencyRumble] * 65535);
    XInputSetState(device->id, &vibration);
}

