
void SerializeSchemaDataValue(EtcTxtWriteState *writer, SchemaTypeDescription type, void *value, const char *fieldName) {

    if (!type.canSerialize) {
        return;
    }

    int32 valuesToSerialize = 1;

    if (type.arrayLength > 0) {
        WritePushArrayValue(writer, fieldName);

        valuesToSerialize = type.arrayLength;
    }

    for (int i = 0; i < valuesToSerialize; i++) {
        void *offsetValue = (void *)((uint8 *)value + type.sizeInBytes * i);

        switch (type.dataType) {
            case SchemaDataType_Invalid: {
                ASSERT(false);
            } break;

            case SchemaDataType_void: {
                ASSERT(false);
            } break;

            case SchemaDataType_bool32: {
                WriteBoolean32Value(writer, *(b32 *)offsetValue, fieldName);
            } break;
            case SchemaDataType_bool: {
                WriteBooleanValue(writer, *(bool *)offsetValue, fieldName);
            } break;
            case SchemaDataType_char: {
                WriteStringValue(writer, (char *)offsetValue, 1, fieldName);
            } break;
            case SchemaDataType_int8: {
                WriteInt8Value(writer, *(int8 *)offsetValue, fieldName);
            } break;
            case SchemaDataType_int16: {
                WriteInt16Value(writer, *(int16 *)offsetValue, fieldName);
            } break;
            case SchemaDataType_int32: {
                WriteInt32Value(writer, *(int32 *)offsetValue, fieldName);
            } break;
            case SchemaDataType_int64: {
                WriteInt64Value(writer, *(int64 *)offsetValue, fieldName);
            } break;
            case SchemaDataType_uint8: {
                WriteUInt8Value(writer, *(uint8 *)offsetValue, fieldName);
            } break;
            case SchemaDataType_uint16: {
                WriteUInt16Value(writer, *(uint16 *)offsetValue, fieldName);
            } break;
            case SchemaDataType_uint32: {
                WriteUInt32Value(writer, *(uint32 *)offsetValue, fieldName);
            } break;
            case SchemaDataType_uint64: {
                WriteUInt64Value(writer, *(uint64 *)offsetValue, fieldName);
            } break;
            case SchemaDataType_real32: {
                WriteFloatValue(writer, *(real32 *)offsetValue, fieldName);
            } break;
            case SchemaDataType_vec2: {
                WriteVec2Value(writer, *(vec2 *)offsetValue, fieldName);
            } break;
            case SchemaDataType_vec3: {
                WriteVec3Value(writer, *(vec3 *)offsetValue, fieldName);
            } break;
            case SchemaDataType_vec4: {
                WriteVec4Value(writer, *(vec4 *)offsetValue, fieldName);
            } break;
            case SchemaDataType_Color: {
                WriteVec4Value(writer, *(Color *)offsetValue, fieldName);
            } break;
            case SchemaDataType_vec2i: {
                WriteVec2iValue(writer, *(vec2i *)offsetValue, fieldName);
            } break;
            case SchemaDataType_vec3i: {
                WriteVec3iValue(writer, *(vec3i *)offsetValue, fieldName);
            } break;
            case SchemaDataType_vec4i: {
                WriteVec4iValue(writer, *(vec4i *)offsetValue, fieldName);
            } break;
            case SchemaDataType_mat3: {
                WriteMat3Value(writer, *(mat3 *)offsetValue, fieldName);
            } break;
            case SchemaDataType_mat4: {
                WriteMat4Value(writer, *(mat4 *)offsetValue, fieldName);
            } break;
            case SchemaDataType_quaternion: {
                WriteQuaternionValue(writer, *(quaternion *)offsetValue, fieldName);
            } break;

            case SchemaDataType_Schema: {
                Schema *schema = *(Schema **)offsetValue;
                if (schema != NULL) {
                    WriteStringValue(writer, schema->name, fieldName);
                }
                else {
                    WriteStringValue(writer, NULL, fieldName);
                }
            } break;

            case SchemaDataType_SchemaData: {
                SchemaData schemaData = *(SchemaData *)offsetValue;

                WritePushObjectValue(writer, fieldName);

                if (schemaData.schemaName != NULL) {
                    Schema *schema = Game.GetSchema(schemaData.schemaName);
                    if (schema != NULL) {
                        WriteStringValue(writer, schema->name, "schema");
                        SerializeValue(writer, schema, schemaData.data, "data");
                    }
                    else {
                        WriteStringValue(writer, NULL, "schema");
                    }
                }
                else {
                    WriteStringValue(writer, NULL, "schema");
                }

                WritePopObjectValue(writer);
            } break;

            case SchemaDataType_EntityID: {
                WriteUInt64Value(writer, *(uint64 *)offsetValue, fieldName);
            } break;

            case SchemaDataType_DynamicArray: {
                WritePushArrayValue(writer, fieldName);

                DynamicArray_Untyped *array = (DynamicArray_Untyped *)offsetValue;

                SchemaTypeDescription elementType = type.templateTypes[0];
                uint32 elementSize = elementType.sizeInBytes;

                for (int j = 0; j < array->count; j++) {
                    void *element = (void *)((uint8 *)array->data + j * elementSize);
                    SerializeSchemaDataValue(writer, elementType, element, NULL);
                }

                WritePopArrayValue(writer);
            } break;

#if 0
            case SchemaDataType_ChunkedArray: {
                WritePushArrayValue(writer, fieldName);

                ChunkedArray_Untyped *array = (ChunkedArray_Untyped *)offsetValue;

                SchemaTypeDescription elementType = type.templateTypes[0];
                uint32 elementSize = elementType.sizeInBytes;

                for (int j = 0; j < array->count; j++) {
                    void *element = (void *)((uint8 *)array->chunks[j / array->chunkSize] + (elementSize * (j % array->chunkSize));
                    SerializeSchemaDataValue(writer, elementType, element, NULL);
                }

                WritePopArrayValue(writer);
            } break;
#endif

            case SchemaDataType_ChunkedArray: {
                // @TODO: ...
               ASSERT(false);
            } break;

            case SchemaDataType_HashTable: {
                // @TODO: ...
                ASSERT(false);
            } break;

            case SchemaDataType_String: {
                if (type.maxStringLength > 0) {
                    uint32 length = StringLength((char *)offsetValue, type.maxStringLength);
                    if (length > 0) {
                        WriteStringValue(writer, (char *)offsetValue, type.maxStringLength, fieldName);
                    }
                    else {
                        WriteStringValue(writer, NULL, fieldName);
                    }
                }
                else if (*(char **)offsetValue != NULL) {
                    WriteStringValue(writer, *(char **)offsetValue, fieldName);
                }
                else {
                    WriteStringValue(writer, NULL, fieldName);
                }
            } break;

            case SchemaDataType_Struct:
            case SchemaDataType_Enum:
            case SchemaDataType_Bitfield:
            case SchemaDataType_Bitarray:
            case SchemaDataType_TaggedUnion:
            case SchemaDataType_PersistentFunctionHandle:
            case SchemaDataType_PersistentFunctionDataPair:
            case SchemaDataType_AssetHandle: {
                Schema *schema = Game.GetSchema(type.typeName);
                SerializeValue(writer, schema, offsetValue, fieldName);
            } break;
        }
    }

    if (type.arrayLength > 0) {
        WritePopArrayValue(writer);
    }
}

void SerializeValue(EtcTxtWriteState *writer, Schema *schema, void *value, const char *fieldName) {
    ASSERT(schema != NULL);

    if (!schema->serializable) {
        return;
    }

    switch (schema->type) {
        case SchemaType_Struct: {
            WritePushObjectValue(writer, fieldName);

            for (int i = 0; i < schema->_struct.fields.count; i++) {
                SchemaField field = schema->_struct.fields[i];
                if (!field.serializable) {
                    continue;
                }

                void *offsetValue = (void *)((uint8 *)value + field.byteOffset);
                SerializeSchemaDataValue(writer, field.type, offsetValue, field.name);
            }

            WritePopObjectValue(writer);
        } break;

        case SchemaType_Enum: {
            uint32 intValue = *(uint32 *)value;
            if (intValue < schema->_enum.count) {
                WriteEnumValue(writer, schema->_enum.stringValues[intValue], fieldName);
            }
        } break;

        case SchemaType_Bitfield: {
            uint32 intValue = *(uint32 *)value;
            WritePushArrayValue(writer, fieldName);

            uint32 *bits = (uint32 *)((u8 *)value + sizeof(uint32));

            for (int i = 0; i < schema->_bitfield.count; i++) {
                if (intValue & (1 << i)) {
                    WriteEnumValue(writer, schema->_bitfield.stringValues[i]);
                }
            }
            
            WritePopArrayValue(writer);
        } break;

            
        case SchemaType_Bitarray: {
            WritePushArrayValue(writer, fieldName);

            for (int i = 0; i < schema->_bitarray.count; i++) {
                if (TestBit_((uint32 *)value, i)) {
                    WriteEnumValue(writer, schema->_bitarray.stringValues[i]);
                }
            }
            
            WritePopArrayValue(writer);
        } break;

        case SchemaType_TaggedUnion: {
            uint32 typeTag = ((TaggedUnion_Untyped *)value)->type;
            ASSERT(typeTag < schema->_tagged_union.fields.count);

            void *offsetValue = TaggedUnionDataPointer(value);

            SchemaField field = schema->_tagged_union.fields[typeTag];
            if (field.serializable) {
                WritePushObjectValue(writer, fieldName);

                SerializeSchemaDataValue(writer, field.type, offsetValue, field.name);

                WritePopObjectValue(writer);
            }
        } break;

        case SchemaType_PersistentFunctionHandle: {
            WritePushObjectValue(writer, fieldName);
            
            FuncHandle funcHandle = *(FuncHandle *)value;
            const char *functionName = NULL;
            if (funcHandle.tableIndex != PersistentFunctionIndex_Invalid) {
                functionName = GetPersistentFunctionName(&Game.persistentFunctionTable, funcHandle.tableIndex, funcHandle.funcType);

                if (functionName != NULL && (uint32)funcHandle.funcType < ENUM_VALUECOUNT(PersistentFuncType)) {
                    WriteEnumValue(writer, ENUM_STRINGVALUE(PersistentFuncType, funcHandle.funcType), "funcType");
                    WriteStringValue(writer, functionName, "functionName");
                }
                else {
                    WriteEnumValue(writer, ENUM_STRINGVALUE(PersistentFuncType, PersistentFuncType_Invalid), "funcType");
                    WriteStringValue(writer, NULL, "functionName");
                }
            }

            else {
                WriteEnumValue(writer, ENUM_STRINGVALUE(PersistentFuncType, PersistentFuncType_Invalid), "funcType");
                WriteStringValue(writer, NULL, "functionName");
            }

            WritePopObjectValue(writer);
        } break;

        case SchemaType_PersistentFunctionDataPair: {
            WritePushObjectValue(writer, fieldName);
            
            for (int i = 0; i < 2; i++) {
                SchemaField field = schema->_func_data_pair.fields[i];
                if (!field.serializable) {
                    continue;
                }

                void *offsetValue = (void *)((uint8 *)value + field.byteOffset);
                SerializeSchemaDataValue(writer, field.type, offsetValue, field.name);
            }

            WritePopObjectValue(writer);
        } break;

        case SchemaType_AssetHandle: {
            AssetHandle handle = *(AssetHandle *)value;

            if (!AssetHandleIsProcedural(handle)) {
                WritePushObjectValue(writer, fieldName);

                if ((uint32)handle.type < ENUM_VALUECOUNT(AssetType)) {
                    WriteEnumValue(writer, ENUM_STRINGVALUE(AssetType, handle.type), "type");

                    Asset *asset = handle.asset;
                    if (asset != NULL &&
                        asset->type == handle.type) {
                        WriteStringValue(writer, asset->name, "name");
                    }
                    else {
                        WriteStringValue(writer, NULL, "name");
                    }
                }
                else {
                    WriteEnumValue(writer, ENUM_STRINGVALUE(AssetType, AssetType_Invalid), "type");
                    WriteStringValue(writer, NULL, "name");
                }
                
                WritePopObjectValue(writer);
            }
        } break;
    }
}

b32 DeserializeSchemaDataValue(EtcTxtParseState *parser, SchemaTypeDescription type, void *value) {

    if (!type.canSerialize) {
        return false;
    }

    EtcTxtParseCursor startCursor = GetCursor(parser);
        
    b32 success = true;

    int32 maxValuesToDeserialize = 1;
    b32 isArray = false;

    if (type.arrayLength > 0) {
        if (ParsePushArrayValue(parser)) {
            isArray = true;
            maxValuesToDeserialize = type.arrayLength;
        }
        else {
            success = false;
        }
    }

    uint32 valuesDeserialized = 0;

    while (!ParseHasError(parser) && success) {

        if (isArray) {
            if (ParsePopArrayValue(parser)) {
                break;
            }

            // If we can't deserialize more, skip values until the end
            // of the array is hit
            if (valuesDeserialized >= maxValuesToDeserialize) {
                if (!ParseSkipValue(parser)) {
                    success = false;
                }

                continue;
            }
        }
        else if (valuesDeserialized >= maxValuesToDeserialize) {
            break;
        }

        void *offsetValue = (void *)((uint8 *)value + type.sizeInBytes * valuesDeserialized);

        b32 elementSuccess = false;

        switch (type.dataType) {
            case SchemaDataType_Invalid: {
                ASSERT(false);
            } break;

            case SchemaDataType_void: {
                ASSERT(false);
            } break;

            case SchemaDataType_bool32: {
                elementSuccess = ParseBoolean32Value(parser, (b32 *)offsetValue);
            } break;
            case SchemaDataType_bool: {
                elementSuccess = ParseBooleanValue(parser, (bool *)offsetValue);
            } break;
            case SchemaDataType_char: {
                elementSuccess = ParseStringValue(parser, (char *)offsetValue, 1);
            } break;
            case SchemaDataType_int8: {
                elementSuccess = ParseInt8Value(parser, (int8 *)offsetValue);
            } break;
            case SchemaDataType_int16: {
                elementSuccess = ParseInt16Value(parser, (int16 *)offsetValue);
            } break;
            case SchemaDataType_int32: {
                elementSuccess = ParseInt32Value(parser, (int32 *)offsetValue);
            } break;
            case SchemaDataType_int64: {
                elementSuccess = ParseInt64Value(parser, (int64 *)offsetValue);
            } break;
            case SchemaDataType_uint8: {
                elementSuccess = ParseUInt8Value(parser, (uint8 *)offsetValue);
            } break;
            case SchemaDataType_uint16: {
                elementSuccess = ParseUInt16Value(parser, (uint16 *)offsetValue);
            } break;
            case SchemaDataType_uint32: {
                elementSuccess = ParseUInt32Value(parser, (uint32 *)offsetValue);
            } break;
            case SchemaDataType_uint64: {
                elementSuccess = ParseUInt64Value(parser, (uint64 *)offsetValue);
            } break;
            case SchemaDataType_real32: {
                elementSuccess = ParseFloatValue(parser, (real32 *)offsetValue);
            } break;
            case SchemaDataType_vec2: {
                elementSuccess = ParseVec2Value(parser, (vec2 *)offsetValue);
            } break;
            case SchemaDataType_vec3: {
                elementSuccess = ParseVec3Value(parser, (vec3 *)offsetValue);
            } break;
            case SchemaDataType_vec4: {
                elementSuccess = ParseVec4Value(parser, (vec4 *)offsetValue);
            } break;
            case SchemaDataType_Color: {
                elementSuccess = ParseVec4Value(parser, (vec4 *)offsetValue);
            } break;
            case SchemaDataType_vec2i: {
                elementSuccess = ParseVec2iValue(parser, (vec2i *)offsetValue);
            } break;
            case SchemaDataType_vec3i: {
                elementSuccess = ParseVec3iValue(parser, (vec3i *)offsetValue);
            } break;
            case SchemaDataType_vec4i: {
                elementSuccess = ParseVec4iValue(parser, (vec4i *)offsetValue);
            } break;
            case SchemaDataType_mat3: {
                elementSuccess = ParseMat3Value(parser, (mat3 *)offsetValue);
            } break;
            case SchemaDataType_mat4: {
                elementSuccess = ParseMat4Value(parser, (mat4 *)offsetValue);
            } break;
            case SchemaDataType_quaternion: {
                elementSuccess = ParseQuaternionValue(parser, (quaternion *)offsetValue);
            } break;

            case SchemaDataType_Schema: {
                const char *typeName;
                if (ParseStringValue(parser, &typeName)) {
                    if (typeName != NULL) {
                        *(Schema **)offsetValue = Game.GetSchema(typeName);
                    }
                    else {
                        *(Schema **)offsetValue = NULL;
                    }

                    elementSuccess = true;
                }
            } break;

            case SchemaDataType_SchemaData: {
                SchemaData schemaData = {};

                const char *identifier;
                if (ParsePushObjectValue(parser) && ParseIdentifier(parser, &identifier)) {
                    const char *schemaName;
                    if (ParseStringValue(parser, &schemaName)) {
                        if (schemaName != NULL) {
                            schemaData.schemaName = (char *)StringCopy(schemaName);

                            Schema *schema = Game.GetSchema(schemaData.schemaName);
                            if (schemaData.schemaName != NULL && schema != NULL) {
                                
                                schemaData.data = Plat.AllocateMem(schema->sizeInBytes);
                                Plat.ClearMem(schemaData.data, schema->sizeInBytes);

                                if (ParseIdentifier(parser, &identifier)) {
                                    DeserializeValue(parser, schema, schemaData.data);    
                                }
                            }
                        }
                    }

                    *(SchemaData *)offsetValue = schemaData;

                    elementSuccess = ParsePopObjectValue(parser);
                }
                
            } break;

            case SchemaDataType_EntityID: {
                elementSuccess = ParseUInt64Value(parser, (uint64 *)offsetValue);
            } break;

            case SchemaDataType_DynamicArray: {
                SchemaTypeDescription elementType = type.templateTypes[0];
                    
                uint32 arrayValueCount;
                if (ParseCountArrayValues(parser, &arrayValueCount)) {

                    if (ParsePushArrayValue(parser)) {
                        DynamicArray_Untyped *dynamicArray = (DynamicArray_Untyped *)value;
                        dynamicArray->count = 0;
                        dynamicArray->capacity = arrayValueCount;

                        if (arrayValueCount > 0) {
                            dynamicArray->data = Plat.AllocateMem(elementType.sizeInBytes * dynamicArray->capacity);
                            // @NOTE: clear data so that if we're deserializing to a larger size than what's
                            // previously been serialized, then we dont write out garbage
                            Plat.ClearMem(dynamicArray->data, elementType.sizeInBytes * dynamicArray->capacity); 

                            for (int i = 0; i < arrayValueCount; i++) {
                                void *elementValue = (void *)((uint8 *)dynamicArray->data + i * elementType.sizeInBytes);
                                if (DeserializeSchemaDataValue(parser, elementType, elementValue)) {
                                    dynamicArray->count++;
                                }
                                else {
                                    break;
                                }
                            }

                            if (dynamicArray->count == arrayValueCount) {
                                elementSuccess = ParsePopArrayValue(parser);
                            }

                            if (!elementSuccess) {
                                // @LEAK: this will potentially leak allocated by elements
                                Plat.DeallocateMem(dynamicArray->data);

                                dynamicArray->count = 0;
                                dynamicArray->capacity = 0;
                                dynamicArray->data = NULL;
                            }
                        }
                        else {
                            elementSuccess = ParsePopArrayValue(parser);
                        }
                    }
                }
            } break;


            case SchemaDataType_ChunkedArray: {
                // @TODO: ...
                ASSERT(false);
            } break;
                
            case SchemaDataType_HashTable: {
                // @TODO: ...
                ASSERT(false);
            } break;

            case SchemaDataType_String: {
                if (type.maxStringLength > 0) {
                    elementSuccess = ParseStringValue(parser, (char *)offsetValue, type.maxStringLength);
                }
                else {
                    // @NOTE: the extra copy is necessary because the
                    // parser keeps track of string it's allocated and
                    // deallocates them on file close
                    const char *stringValue;
                    if (ParseStringValue(parser, &stringValue)) {
                        if (stringValue != NULL) {
                            *(const char **)offsetValue = StringCopy(stringValue);
                        }
                        else {
                            *(char **)offsetValue = NULL;
                        }
                        elementSuccess = true;
                    }
                }
            } break;

            case SchemaDataType_Struct:
            case SchemaDataType_Enum:
            case SchemaDataType_Bitfield:
            case SchemaDataType_Bitarray:
            case SchemaDataType_TaggedUnion:
            case SchemaDataType_PersistentFunctionHandle:
            case SchemaDataType_PersistentFunctionDataPair:
            case SchemaDataType_AssetHandle: {
                Schema *schema = Game.GetSchema(type.typeName);
                elementSuccess = DeserializeValue(parser, schema, offsetValue);
            } break;
        }

        valuesDeserialized++;
        success &= elementSuccess;
    }

    if (!success) {
        SetCursor(parser, startCursor);
    }

    return success;
}

// @NOTE: assumes `value` is cleared already
b32 DeserializeValue(EtcTxtParseState *parser, Schema *schema, void *value) {

    if (!schema->serializable) {
        return false;
    }

    EtcTxtParseCursor startCursor = GetCursor(parser);
    
    b32 success = false;

    switch (schema->type) {
        case SchemaType_Struct: {
            if (ParsePushObjectValue(parser)) {

                while (!ParseHasError(parser)) {

                    const char *fieldName;
                    if (ParseIdentifier(parser, &fieldName)) {

                        uint32 fieldIndex;
                        if (FindString(&schema->_struct.fields[0].name, schema->_struct.fields.count, fieldName, &fieldIndex, sizeof(SchemaField))) {
                            SchemaField field = schema->_struct.fields[fieldIndex];

                            void *offsetValue = (void *)((uint8 *)value + field.byteOffset);
                            if (field.serializable &&
                                DeserializeSchemaDataValue(parser, field.type, offsetValue)) {
                            }
                            else if (!ParseSkipValue(parser)) {
                                break;
                            }
                        }
                        else if (!ParseSkipValue(parser)) {
                            break;
                        }
                    }
                    else {
                        if (ParsePopObjectValue(parser)) {
                            success = true;
                        }

                        break;
                    }
                }
            }
        } break;

        case SchemaType_Enum: {
            if (ParseEnumValue_(parser, schema->_enum.stringValues, schema->_enum.count, (uint32 *)value)) {
                success = true;
            }
        } break;

        case SchemaType_Bitfield: {
            if (ParsePushArrayValue(parser)) {
                while (!ParseHasError(parser)) {
                    if (ParsePopArrayValue(parser)) {
                        success = true;
                        break;
                    }

                    uint32 shiftValue;
                    if (ParseEnumValue_(parser, schema->_bitfield.stringValues, schema->_bitfield.count, &shiftValue)) {
                        *(uint32 *)value = *(uint32 *)value | (1 << shiftValue);
                    }
                    else if (!ParseSkipValue(parser)) {
                        break;
                    }
                }
            }
        } break;

            
        case SchemaType_Bitarray: {
            if (ParsePushArrayValue(parser)) {
                while (!ParseHasError(parser)) {
                    if (ParsePopArrayValue(parser)) {
                        success = true;
                        break;
                    }

                    uint32 index;
                    if (ParseEnumValue_(parser, schema->_bitarray.stringValues, schema->_bitarray.count, &index)) {
                        SetBit_((uint32 *)value, index);
                    }
                    else if (!ParseSkipValue(parser)) {
                        break;
                    }
                }
            }
        } break;

            
        case SchemaType_TaggedUnion: {
            if (ParsePushObjectValue(parser)) {
                TaggedUnion_Untyped *unionPtr = (TaggedUnion_Untyped *)value;

                const char *identifier;
                if (ParseIdentifier(parser, &identifier)) {
                    uint32 index;
                    if (FindString(&schema->_tagged_union.fields[0].name, schema->_tagged_union.fields.count, identifier, &index, sizeof(SchemaField))) {
                        SchemaField field = schema->_tagged_union.fields[index];

                        void *offsetValue = TaggedUnionDataPointer(value);
                        if (field.serializable &&
                            DeserializeSchemaDataValue(parser, field.type, offsetValue)) {
                            unionPtr->type = index;
                            if (ParsePopObjectValue(parser)) {
                                success = true;
                            }
                        }
                    }
                }
            }
        } break;

        case SchemaType_PersistentFunctionHandle: {
            FuncHandle *funcHandle = (FuncHandle *)value;
            if (ParsePushObjectValue(parser)) {
                if (ParseIdentifier(parser, NULL, "funcType")) {
                    PersistentFuncType funcType;
                    if (ParseEnumValue(parser, PersistentFuncType, &funcType) &&
                        funcType == schema->_func_handle.funcType) {
                        if (ParseIdentifier(parser, NULL, "functionName")) {
                            const char *functionName;
                            if (ParseStringValue(parser, &functionName) && functionName != NULL) {
                                    
                                funcHandle->funcType = funcType;
                                funcHandle->tableIndex = GetPersistentFunctionIndex(&Game.persistentFunctionTable, functionName, funcType);
                            }
                            else {
                                funcHandle->funcType = PersistentFuncType_Invalid;
                                funcHandle->tableIndex = 0;
                            }

                            if (ParsePopObjectValue(parser)) {
                                success = true;
                            }
                        }
                        else if (!ParseSkipValue(parser)) {
                            funcHandle->funcType = PersistentFuncType_Invalid;
                            funcHandle->tableIndex = 0;
                        }
                    }
                    else {
                        funcHandle->funcType = PersistentFuncType_Invalid;
                        funcHandle->tableIndex = 0;
                    }
                }
            }
        } break;

        case SchemaType_PersistentFunctionDataPair: {
            if (ParsePushObjectValue(parser)) {

                while (!ParseHasError(parser)) {

                    const char *fieldName;
                    if (ParseIdentifier(parser, &fieldName)) {

                        uint32 fieldIndex;
                        if (FindString(&schema->_func_data_pair.fields[0].name, 2, fieldName, &fieldIndex, sizeof(SchemaField))) {
                            SchemaField field = schema->_func_data_pair.fields[fieldIndex];

                            void *offsetValue = (void *)((uint8 *)value + field.byteOffset);
                            if (field.serializable &&
                                DeserializeSchemaDataValue(parser, field.type, offsetValue)) {
                            }
                            else if (!ParseSkipValue(parser)) {
                                break;
                            }
                        }
                        else if (!ParseSkipValue(parser)) {
                            break;
                        }
                    }
                    else {
                        if (ParsePopObjectValue(parser)) {
                            success = true;
                        }

                        break;
                    }
                }
            }
        } break;

        case SchemaType_AssetHandle: {

            if (ParsePushObjectValue(parser)) {
                const char *identifier;
                if (ParseIdentifier(parser, &identifier, "type")) {
                    AssetType assetType;
                    if (ParseEnumValue(parser, AssetType, (uint32 *)&assetType)) {

                        AssetType schemaAssetType;
                        if (ENUM_FINDVALUE(AssetType, schema->_assetHandle.assetTypeName, &schemaAssetType) &&
                            assetType == schemaAssetType) {

                            if (ParseIdentifier(parser, &identifier, "name")) {

                                const char *assetName;
                                if (ParseStringValue(parser, &assetName)) {
                                    AssetHandle *assetHandle = (AssetHandle *)value;

                                    if (assetType != AssetType_Invalid) {
                                        *assetHandle = LoadAsset(Game.assets, assetType, assetName);
                                    }

                                    if (ParsePopObjectValue(parser)) {
                                        success = true;
                                    }
                                }
                            }
                        }
                    }
                }
            }
        } break;
    }

    if (!success) {
        SetCursor(parser, startCursor);
    }

    return success;
}


// Schema Comparison
b32 StructCompareField(Schema *schema, void *data, StructCompareData structCompareData) {
    for (int i = 0; i < schema->_struct.fields.count; i++) {
        if (StringEquals(structCompareData.fieldName, schema->_struct.fields[i].name)) {

            CompareData compareData = structCompareData.compareData;
            
            SchemaField field = schema->_struct.fields[i];
            void *value = (void *)((u8 *)data + field.byteOffset);
            
            switch (structCompareData.compareData.type) {
                case CompareData_real32Equality: {
                    ASSERT(field.type.dataType == SchemaDataType_real32);
                    return *(real32 *)value == compareData._data.real32Equality.value;
                } break;
                case CompareData_real32Range: {
                    ASSERT(field.type.dataType == SchemaDataType_real32);
                    return (*(real32 *)value > compareData._data.real32Range.min &&
                            *(real32 *)value <= compareData._data.real32Range.max);
                } break;

                case CompareData_real32Greater: {
                    ASSERT(field.type.dataType == SchemaDataType_real32);
                    return *(real32 *)value > compareData._data.real32Greater.min;
                } break;
                    
                default: {} break;
            }

            break;
        }
    }

    return false;
}

