

STRUCT()
struct AABB {
    vec3 center;
    vec3 halfExtents;
};


STRUCT()
struct MinMaxAABB {
    vec3 min;
    vec3 max;
};


inline MinMaxAABB AABBToMinMaxAABB(AABB aabb) {
    MinMaxAABB result = {};
    result.min = aabb.center - aabb.halfExtents;
    result.max = aabb.center + aabb.halfExtents;

    return result;
}


STRUCT()
struct Sphere {
    vec3 center;
    real32 radius;
};

STRUCT()
struct OBB {
    vec3 center;
    quaternion rot;
    vec3 dimensions;
};

TAGGED_UNION(Shape, {
    Sphere sphere;
    OBB obb;
    AABB aabb;
});


vec3i IndexTo3D(uint32 i, uint32 maxX, uint32 maxY, uint32 maxZ);
uint32 IndexTo1D(vec3i index3D, int32 width, int32 height);
vec3 RelativeAABBPosition(vec3 pos, MinMaxAABB aabb);
int32 PositionToGridCellIndex(vec3 pos, MinMaxAABB aabb, real32 cellDimensions, uint32 width, uint32 height, uint32 depth, r32 epsilon = 0.0f, vec3i *index3D = NULL);
MinMaxAABB ComputeAABBForGridCell(MinMaxAABB aabb, uint32 cellDimension, vec3i index3D);
uint32 RaycastGrid(Ray ray, MinMaxAABB aabb, u32 width, u32 height, u32 depth, r32 cellSize, ChunkedArray<u32> *hitCells);


COMPONENT(Shape) {
    Shape shape;
};

SYSTEM(Shape) {
    HashTable<EntityID, uint32> entityMap;

    DynamicArray<ShapeComponent> shapes;
};

bool TestAABBAABB(AABB a, AABB b);

BITFIELD(DebugShape, {
    DebugShape_DrawShape,
    DebugShape_DrawCenter,
    DebugShape_Raycast,
});

#if ETC_DEBUG
void DebugDrawShape(Shape shape, vec3 position, quaternion rotation, vec3 scale, vec4 color);
void DebugDrawShape(Shape shape, Transform trans, vec4 color);
#endif


DEFINE_FUNCTIONTYPE(OnTriggerEnterFunc, void, EntityDB *entityDB, EntityID triggerID, EntityID entityID, SchemaData data);
DEFINE_FUNCTIONTYPE(OnTriggerExitFunc, void, EntityDB *entityDB, EntityID triggerID, EntityID entityID, SchemaData data);

STRUCT()
struct TriggerVolumeListener {
    // @TODO: array, and maybe EntityIdentifier instead of ID
    EntityID filter;

    ComponentType componentTypeFilter;

    SchemaData tagData;

    OnTriggerEnterFuncHandle enterFunc;
    OnTriggerExitFuncHandle exitFunc;

    SchemaData data;
};

COMPONENT(TriggerVolume) {
    Shape shape;
    DynamicArray<TriggerVolumeListener> listeners;
};

struct TriggerVolumeState {
     DynamicArray<EntityID> entitiesInside;
};

// struct CollidingTriggersWithEntity {
//     DynamicArray<EntityID> triggerIDs;
// };

SYSTEM(TriggerVolume) {
    HashTable<EntityID, uint32> entityMap;

    DynamicArray<TriggerVolumeComponent> triggers;
    DynamicArray<TriggerVolumeState> states;

    // HashTable<EntityID, uint32> collidingEntityMap;
    // DynamicArray<CollidingTriggersWithEntity> collidingTriggers;
};

BITFIELD(DebugTriggerVolume, {
    DebugTriggerVolume_Draw,
    DebugTriggerVolume_Shape,
});


bool PointInAABB(MinMaxAABB aabb, vec3 point, real32 epsilon = 0.0f);
r32 AABBVolume(MinMaxAABB aabb);
bool AABBsOverlap(MinMaxAABB a, MinMaxAABB b);
bool AABBContainsAABB(MinMaxAABB a, MinMaxAABB b);
MinMaxAABB AABBUnion(MinMaxAABB a, MinMaxAABB b);
bool PointInAABB(AABB shape, vec3 shapePos, vec3 point);

