
void AllocateTaskScheduler(TaskScheduler *scheduler, uint32 capacity) {
    scheduler->taskCapacity = capacity;
    scheduler->tasks = (Task *)Plat.AllocateMem(sizeof(Task) * capacity);

    scheduler->freeList.capacity = capacity + 1;
    scheduler->freeList.ringBuffer = (uint32 *)Plat.AllocateMem(sizeof(uint32) * (capacity + 1));

    for (uint32 i = 0; i < capacity; i++) {
        PushBackForce(&scheduler->freeList, i);
    }
        
    scheduler->queuedTasks.capacity = capacity + 1;
    scheduler->queuedTasks.ringBuffer = (uint32 *)Plat.AllocateMem(sizeof(uint32) * (capacity + 1));
}

b32 CompleteNextTask(TaskScheduler *scheduler, b32 block) {
    uint32 index;
    if (block) {
        PopFrontForce(&scheduler->queuedTasks, &index);
    }
    else if (!PopFront(&scheduler->queuedTasks, &index)) {
        return false;
    }

    Task *task = scheduler->tasks + index;
    task->fn(task->data);

    PushBackForce(&scheduler->freeList, index);

    return true;
}

void CompleteNextTasks(TaskScheduler *scheduler, uint32 limit) {
    for (int i = 0; i < limit; i++) {
        if (scheduler->semaPtr == NULL || Plat.SemaWait(scheduler->semaPtr, false)) {
            if (!CompleteNextTask(scheduler, false)) {
                break;
            }
        }
    }
}

// @BUG: the functino being called in CompleteNextTask hasnt necessarily finished so we could
// fall out of this while loop.
void WaitForAllTasks(TaskScheduler *scheduler) {
    while (Plat.SemaWait(scheduler->semaPtr, false)) {
        CompleteNextTask(scheduler, false);
    }
}

void PlatformTaskSchedulerThreadProc(void *ptr) {
    TaskScheduler *scheduler = (TaskScheduler *)ptr;

    TaskSchedulerTLSData tlsData = {};
    AllocateTLSData(&tlsData);

    while (true) {
        if (Plat.SemaWait(scheduler->semaPtr, true)) {
            CompleteNextTask(scheduler, true);
            ClearMemoryArena(&tlsData.arena);
        }
        else {
            break;
        }
    }
}

void AllocateTLSData(TaskSchedulerTLSData *tlsData) {
    AllocateMemoryArena(&tlsData->arena, Megabytes(128));
    Plat.SetTLS(Plat.taskSchedulerTLSKey, tlsData);
}

