
#ifndef _ETC_TASK_H
#define _ETC_TASK_H

#include "lockfree_queue.h"

typedef void TaskFunc(void *);

struct Task {
    TaskFunc *fn;
    uint8 data[24];
};

struct TaskScheduler {
    uint32 taskCapacity;
    Task *tasks;

    LockFreeQueue<uint32> freeList;
    LockFreeQueue<uint32> queuedTasks;

    void *semaPtr;
};

void AllocateTaskScheduler(TaskScheduler *scheduler, uint32 capacity);

b32 CompleteNextTask(TaskScheduler *scheduler, b32 block);
void CompleteNextTasks(TaskScheduler *scheduler, uint32 limit = 1);
void WaitForTasks(TaskScheduler *scheduler);

void PlatformTaskSchedulerThreadProc(void *ptr);

inline Task *AllocateTask(TaskScheduler *scheduler) {
    uint32 offset;
    PopFrontForce(&scheduler->freeList, &offset);

    Task *task = scheduler->tasks + offset;
    Plat.ClearMem(task, sizeof(Task));

    return task;
}

inline void SubmitTask(TaskScheduler *scheduler, Task *task) {

    int32 index = task - scheduler->tasks;
    if (index < 0 || index >= scheduler->taskCapacity) {
        ASSERT(false);
    }

    PushBackForce(&scheduler->queuedTasks, (uint32)index);
    
    if (scheduler->semaPtr != NULL) {
        Plat.SemaSignal(scheduler->semaPtr);
    }
}


struct TaskSchedulerTLSData {
    MemoryArena arena;
};

void AllocateTLSData(TaskSchedulerTLSData *data);

#endif

