
// UIStyle
FOREACH_STYLETYPE(DEFINE_STYLETYPE)

void PushStyles(UIStyle *styles, uint32 count, UIManager *manager) {
    manager = ResolveUIManager(manager);

    UIStyleGroup group = {};
    group.first = manager->styles.count;
    group.count = count;

    for (int i = 0; i < count; i++) {
        PushBack(&manager->styles, styles[i]);
    }

    PushBack(&manager->styleGroups, group);
    
    for (int i = 0; i < manager->styles.count; i++) {
        UIStyle *style = &manager->styles[i];
        manager->cachedStyle[style->tag] = style;
    }
}

void PushStyle(UIStyle style, UIManager *manager) {
    manager = ResolveUIManager(manager);

    UIStyleGroup group = {};
    group.first = manager->styles.count;
    group.count = 1;

    PushBack(&manager->styles, style);
    PushBack(&manager->styleGroups, group);

    for (int i = 0; i < manager->styles.count; i++) {
        UIStyle *style = &manager->styles[i];
        manager->cachedStyle[style->tag] = style;
    }
}

void PopStyles(UIManager *manager) {
    manager = ResolveUIManager(manager);

    UIStyleGroup group;
    if (PopBack(&manager->styleGroups, &group)) {
        for (int i = 0; i < group.count; i++) {

            UIStyle poppedStyle;
            if (PopBack(&manager->styles, &poppedStyle)) {
                manager->cachedStyle[poppedStyle.tag] = NULL;

                for (int j = manager->styles.count - 1; j >= 0; j--) {
                    UIStyle *style = &manager->styles[j];
                    if (style->tag == poppedStyle.tag) {
                        manager->cachedStyle[poppedStyle.tag] = style;
                        break;
                    }
                }
            }
        }
    }
}


// General Management
void AllocateUIManager(UIManager *manager) {
    AllocateCommandList(&manager->commandList, 32768 * 4, Megabytes(2));

    manager->styles = MakeDynamicArray<UIStyle>(256);

    glGenBuffers(1, &manager->textBufferID);
}

void DeallocateUIManager(UIManager *manager) {
    glDeleteBuffers(1, &manager->textBufferID);
}

WidgetID PushWidgetID(UIManager *manager, const char *name, b32 recordID) {
    WidgetGroup *group = PeekWidgetGroup(manager);

    void *hashInput = (void *)name;
    uint32 hashInputLength = StringLength(name);
    ASSERT(hashInput != NULL);

    if (hashInputLength == 0) {
        hashInput = &group->prevID;
        hashInputLength = sizeof(group->prevID);
    }
    
    WidgetID result = group->prevID;
    for (int i = 0; i < hashInputLength; i++) {
        result = result * 33 + ((uint8 *)hashInput)[i];
    }

    ASSERT(result != WidgetID_Invalid);

    if (result == manager->activeID) {
        manager->activeIDAccessedLastFrame = true;
    }

    if (recordID) {
        group->prevID = result;
        group->widgetCount++;
    }

    return result;
}

b32 RegisterSelectableWidget(UIManager *manager, WidgetID id) {
    if (!manager->activeIDAccessedLastFrame) {
        manager->prevSelectableWidgetID = id;
    }
    else if (id != manager->activeID && manager->nextSelectableWidgetID == WidgetID_Invalid) {
        manager->nextSelectableWidgetID = id;
    }

    b32 shouldActivate = (manager->shouldActivateNextWidgetID ||
                          id == manager->shouldActivateWidgetID);
    manager->shouldActivateNextWidgetID = false;

    return shouldActivate;
}

void ActivateNextWidget(UIManager *manager) {
    manager = ResolveUIManager(manager);
    manager->shouldActivateNextWidgetID = true;
}

void *AllocatePersistentData(UIManager *manager, WidgetID id, uint32 dataID, uint32 size) {

    if (!HashTableContains(&manager->persistentDataMap, id)) {
        PersistentWidgetData data = {};
        HashTableInsert(&manager->persistentDataMap, id, data, true);
    }

    PersistentWidgetData *data;
    if (HashTableGetPtr(&manager->persistentDataMap, id, &data)) {
        data->accessedLastFrame = true;

        PersistentAllocatedData *allocatedData = (PersistentAllocatedData *)AllocateClearedMemory(sizeof(PersistentAllocatedData) + size);
        allocatedData->next = data->dataHead;
        allocatedData->dataID = dataID;
        allocatedData->dataSize = size;

        data->dataHead = allocatedData;

        return allocatedData->data;
    }

    return NULL;
}

void DeallocatePersistentData(UIManager *manager, WidgetID id, uint32 dataID) {
    PersistentWidgetData *persistentData;
    if (HashTableGetPtr(&manager->persistentDataMap, id, &persistentData)) {

        PersistentAllocatedData *prevAllocatedData = NULL;
        PersistentAllocatedData *allocatedData = persistentData->dataHead;
        while (allocatedData != NULL) {
            if (allocatedData->dataID == dataID) {
                if (prevAllocatedData == NULL) {
                    persistentData->dataHead = allocatedData->next;
                }
                else {
                    prevAllocatedData->next = allocatedData->next;
                }

                Plat.DeallocateMem(allocatedData);

                break;
            }

            prevAllocatedData = allocatedData;
            allocatedData = allocatedData->next;
        }
    }
}

void DeallocateAllPersistentData(PersistentWidgetData *persistentData) {
    PersistentAllocatedData *allocatedData = persistentData->dataHead;
    while (allocatedData != NULL) {
        PersistentAllocatedData *nextAllocatedData = allocatedData->next;

        Plat.DeallocateMem(allocatedData);
        allocatedData = nextAllocatedData;
    }

    persistentData->dataHead = NULL;
}

void DeallocateAllPersistentData(UIManager *manager, WidgetID id) {
    PersistentWidgetData *persistentData;
    if (HashTableGetPtr(&manager->persistentDataMap, id, &persistentData)) {
        DeallocateAllPersistentData(persistentData);
    }
}

b32 GetPersistentData(UIManager *manager, WidgetID id, uint32 dataID, void **data) {
    b32 result = false;

    PersistentWidgetData *persistentData;
    if (HashTableGetPtr(&manager->persistentDataMap, id, &persistentData)) {
        persistentData->accessedLastFrame = true;

        PersistentAllocatedData *allocatedData = persistentData->dataHead;
        while (allocatedData != NULL) {
            if (allocatedData->dataID == dataID) {
                persistentData->accessedLastFrame = true;
                *data = (void *)allocatedData->data;
                result = true;
                break;
            }

            allocatedData = allocatedData->next;
        }
    }

    return result;
}

void SetHoveredWidget(UIManager *manager, WidgetGroup *group, WidgetID id, real32 depth, int32 index) {
    UIRenderPriority priority = UIRenderPriority_Default;
    if (HasFlag(group->flags, WidgetGroupFlag_OverridePriority)) {
        priority = group->overridePriority;
    }

    if (manager->hoveredID != WidgetID_Invalid && manager->hoveredID != id) {
        if (priority < manager->hoveredRenderPriority) {
            return;
        }

        if (HasFlag(group->flags, WidgetGroupFlag_Is3D)) {
            if (depth >= manager->hoveredDepth) {
                return;
            }
        }
    }

    manager->hoveredRenderPriority = priority;
    manager->hoveredGroupIndex = group->groupIndex;
    manager->hoveredGroupDepth = group->groupDepth;
    manager->hoveredDepth = depth;
    manager->hoveredIndex = index;
    manager->hoveredID = id;
    manager->hoveredIDAccessedLastFrame = true;
}

b32 IsHovered(UIManager *manager, WidgetID id) {
    b32 result = false;

    if (id == manager->hoveredID) {
        result = true;
    }

    return result;
}

void SetActiveWidget(UIManager *manager, WidgetID id, uint32 activeWidgetDataSize) {
    // @TODO: mark all groups housing the widget as "active"
    
    if (id == WidgetID_Invalid) {
        manager->activeID = WidgetID_Invalid;
        manager->activeIDAccessedLastFrame = false;
        return;
    }

    manager->activeID = id;
    manager->activeIDAccessedLastFrame = true;

    if (manager->activeWidgetData == NULL && activeWidgetDataSize > 0) {
        manager->activeWidgetData = Plat.AllocateMem(activeWidgetDataSize);
        manager->activeWidgetDataSize = activeWidgetDataSize;
    }
    else if (manager->activeWidgetDataSize < activeWidgetDataSize) {
        manager->activeWidgetData = Plat.ResizeMem(manager->activeWidgetData, activeWidgetDataSize);
        manager->activeWidgetDataSize = activeWidgetDataSize;
    }

    Plat.ClearMem(manager->activeWidgetData, manager->activeWidgetDataSize);
}

b32 IsActive(UIManager *manager, WidgetID id) {
    b32 result = false;

    if (id == manager->activeID) {
        result = true;
    }

    return result;
}

void CleanupPersistentDataMap(UIManager *manager) {
    uint32 removedWidgetCount = 0;
    WidgetID *removedWidgetIDs = PushArray(Game.frameMem, WidgetID, manager->persistentDataMap.capacity);

    for (int i = 0; i < manager->persistentDataMap.capacity; i++) {
        if (!manager->persistentDataMap.occupied[i]) {
            continue;
        }

        WidgetID id = manager->persistentDataMap.keys[i];
        PersistentWidgetData *persistentData = &manager->persistentDataMap.values[i];

        if (id != manager->activeID && !persistentData->accessedLastFrame) {
            removedWidgetIDs[removedWidgetCount++] = id;
            DeallocateAllPersistentData(persistentData);
        }

        persistentData->accessedLastFrame = false;
    }

    for (int i = 0; i < removedWidgetCount; i++) {
        b32 removed = HashTableRemove(&manager->persistentDataMap, removedWidgetIDs[i]);
        ASSERT(removed);
    }
}

void UIManagerStartFrame(UIManager *manager) {
    TIMED_BLOCK_LABEL("UI");
    
    // Cleanup previous frame
    CleanupPersistentDataMap(manager);
    ChunkedArrayClear(&manager->groupStack);

    // @GACK: ChunkedArray should support zero-is-initialization
    if (manager->groupStack.allocator == NULL) {
        manager->groupStack = MakeChunkedArray<WidgetGroup>(Game.heap, 16);
    }

    DynamicArrayClear(&manager->styleGroups);
    DynamicArrayClear(&manager->styles);
    Plat.ClearMem(manager->cachedStyle, sizeof(UIStyle *) * UIStyleTag_Count);


    // Start new frame
    if (!manager->hoveredIDAccessedLastFrame) {
        manager->hoveredID = WidgetID_Invalid;
        manager->hoveredDepth = INFINITY;
        manager->hoveredGroupDepth = 0;
    }
    manager->hoveredIDAccessedLastFrame = false;

    if (!manager->activeIDAccessedLastFrame) {
        manager->activeID = WidgetID_Invalid;
        manager->blockingKeyboard = false;
        manager->wrappingMouse = false;
    }
    manager->activeIDAccessedLastFrame = false;

    if (InputActionPressed(Game.inputActions, InputAction_EditorSelectPrevWidget, true)) {
        manager->shouldActivateWidgetID = manager->prevSelectableWidgetID;
    }
    else if (InputActionPressed(Game.inputActions, InputAction_EditorSelectNextWidget, true)) {
        manager->shouldActivateWidgetID = manager->nextSelectableWidgetID;
    }
    else {
        manager->shouldActivateWidgetID = WidgetID_Invalid;
    }

    manager->prevSelectableWidgetID = WidgetID_Invalid;
    manager->nextSelectableWidgetID = WidgetID_Invalid;

    manager->nextGroupIndex = 0;

    // Read Input
    manager->prevScreenMousePosition = manager->screenMousePosition;

    // mouse input treats lower left as (0, 0)
    vec2 mouseInput = InputAnalogue2D(Game.mouse, Etc_MousePositionXY, true);
    manager->screenMousePosition = mouseInput;
    manager->screenMousePosition.y = Game.graphics->height - manager->screenMousePosition.y;

    if (manager->wrappingMouse) {
        Plat.SetMouseWrap(true);
        
        vec2 wrapScreenSize = {(real32)Game.graphics->width, (real32)Game.graphics->height};

        vec2 mousePos = manager->screenMousePosition;

        bool wrappedX = false;
        if (mousePos.x < 0) {
            mousePos.x = Game.graphics->width + mousePos.x;
            manager->wrapScreenCoords.x--;
            wrappedX = true;
        }
        else if (mousePos.x > Game.graphics->width) {
            mousePos.x = mousePos.x - (real32)Game.graphics->width;
            manager->wrapScreenCoords.x++;
            wrappedX = true;
        }

        bool wrappedY = false;
        if (mousePos.y < 0) {
            mousePos.y = Game.graphics->height + mousePos.y;
            manager->wrapScreenCoords.y--;
            wrappedY = true;
        }
        else if (mousePos.y > Game.graphics->height) {
            mousePos.y = mousePos.y - (real32)Game.graphics->height;
            manager->wrapScreenCoords.y++;
            wrappedY = true;
        }

        manager->screenMouseVirtualPosition = {
            manager->wrapScreenCoords.x * wrapScreenSize.x + mousePos.x,
            manager->wrapScreenCoords.y * wrapScreenSize.y + mousePos.y
        };

        if (wrappedX || wrappedY) {
            Plat.MoveMouse(mousePos.x, Game.graphics->height - mousePos.y);
            manager->screenMousePosition = mousePos;
        }

        //LOGF("%d %d", manager->wrapScreenCoords.x, manager->wrapScreenCoords.y);
    }
    else {
        Plat.SetMouseWrap(false);
        
        manager->screenMouseVirtualPosition = manager->screenMousePosition;
        manager->wrapScreenCoords = V2i(0, 0);
    }

    manager->screenMouseVirtualPositionNorm.x = manager->screenMouseVirtualPosition.x / Game.graphics->width;
    manager->screenMouseVirtualPositionNorm.y = manager->screenMouseVirtualPosition.y / Game.graphics->height;

    manager->screenMousePositionNorm.x = manager->screenMousePosition.x / Game.graphics->width;
    manager->screenMousePositionNorm.y = manager->screenMousePosition.y / Game.graphics->height;

    if (manager->showMouse) {
        Plat.ShowMouse();
    }
    else {
        Plat.HideMouse();        
    }

    Game.keyboard->blocked = manager->blockingKeyboard;
    Game.mouse->blocked |= manager->blockingMouse;
    manager->blockingMouse = false;

    manager->showMouse = false;

    DynamicArray<uint32> tempInputCharacters = manager->inputCharactersBackBuffer;
    manager->inputCharactersBackBuffer = manager->inputCharactersFrontBuffer;
    manager->inputCharactersFrontBuffer = tempInputCharacters;
    DynamicArrayClear(&manager->inputCharactersFrontBuffer);


    // Setup/Teardown Rendering
    ClearCommandList(&manager->commandList);
    SetDrawRenderTarget(&manager->commandList, 0, NULL);

    DisableDepthWrite(&manager->commandList, 0);
    DisableDepthTest(&manager->commandList, 0);
    EnableBlending(&manager->commandList, 0, GL_FUNC_ADD, GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    DisableScissorTest(&manager->commandList, GenerateUIRenderKey(UIRenderPriority_Teardown));


    UIStyle defaultStyles[] = {
        MakeStyle4(UIStyle_WindowBackgroundColor, WithAlpha(BLACK, 0.65f)),

        MakeStyleShaderProgram(UIStyle_MeshShader, Game.graphics->defaultUIMeshShader),
        MakeStyleShaderProgram(UIStyle_TexturedMeshShader, Game.graphics->defaultUITexturedMeshShader),
        MakeStyleShaderProgram(UIStyle_TextShader, Game.graphics->defaultUITextShader),

        MakeStyleFont(UIStyle_Font, DebugMem->debugFontHandle),
        MakeStylef(UIStyle_FontSize, 20),
        MakeStyle4(UIStyle_FontColor, WHITE),

        MakeStyle4(UIStyle_ButtonColor, PASTEL_RED),
        MakeStyle4(UIStyle_ButtonActiveColor, PASTEL_YELLOW),

        MakeStylePadding(UIStyle_InsetMargin, {4, 4, 4, 4}),
        MakeStylePadding(UIStyle_Padding, {4, 4, 1, 1}),

        MakeStylef(UIStyle_InputStringBlinkRate, 2.5f),
        MakeStylef(UIStyle_DoubleClickThreshold, 0.2f),
    };
    PushStyles(defaultStyles, ARRAY_LENGTH(UIStyle, defaultStyles), manager);

    EtcRect screenRect = {
        V2(0, 0),
        V2(Game.graphics->width, Game.graphics->height)
    };
    PushWidgetGroup(WidgetID_Invalid, V2(0, 0), screenRect, screenRect, 0, manager);
}


// Layout
void ResizeWidgetGroup_(UIManager *manager, WidgetGroup *group, vec2 anchor, EtcRect localBounds, EtcRect localClipRect) {
    vec2 parentScreenPosition;
    vec2 parentScreenSize;
    EtcRecti parentClipRect;

    if (group->groupDepth > 0 && !HasFlag(group->flags, WidgetGroupFlag_Absolute)) {
        WidgetGroup *parentGroup = &manager->groupStack[group->groupDepth - 1];
        ASSERT(!HasFlag(parentGroup->flags, WidgetGroupFlag_Is3D));

        parentScreenPosition = parentGroup->screenPosition;
        parentScreenSize = parentGroup->screenSize;
        parentClipRect = parentGroup->clipRect;
    }
    else {
        parentScreenPosition = V2(0, 0);
        parentScreenSize = V2(Game.graphics->width, Game.graphics->height);
        parentClipRect = {V2i(0, 0), V2i(Game.graphics->width, Game.graphics->height)};
    }

    group->anchor = anchor;
    group->localBounds = localBounds;
    group->localClipRect = localClipRect;

    group->screenPosition = parentScreenPosition + anchor + localBounds.pos;
    group->screenSize = localBounds.size;

    // @TODO: compute the clip rect from the local clip rect
    EtcRecti clipRect;
    clipRect.pos = V2i(Floorf(group->screenPosition.x),
                       Floorf(group->screenPosition.y));
    clipRect.size = V2i(Ceilf(group->screenSize.x),
                        Ceilf(group->screenSize.y));

    group->clipRect = Intersection(parentClipRect, clipRect);

    group->localMousePosition = manager->screenMousePosition - group->screenPosition;
    group->isHoveringGroup = PointInRect(group->clipRect, manager->screenMousePosition);

    if (HasFlag(group->flags, WidgetGroupFlag_BlocksMouse)) {
        manager->blockingMouse |= group->isHoveringGroup;
    }
}

void CommitWidgetBounds(UIManager *manager, WidgetGroup *group, EtcRect widgetBounds) {
    group->prevWidgetBounds = widgetBounds;
    group->totalWidgetBounds = Union(group->totalWidgetBounds, widgetBounds);

    if (HasFlag(group->flags, WidgetGroupFlag_ResizeToContent)) {
        UIPadding padding = GetStylePadding(UIStyle_Padding, manager);

        EtcRect newBounds = group->localBounds;
        if (HasFlag(group->flags, WidgetGroupFlag_ResizeToContentHorizontal)) {
            real32 left = group->totalWidgetBounds.pos.x;
            real32 right = group->totalWidgetBounds.pos.x + group->totalWidgetBounds.size.x;
            newBounds.pos.x = left;
            newBounds.size.x = right - left;
        }
        if (HasFlag(group->flags, WidgetGroupFlag_ResizeToContentVertical)) {
            real32 top = group->totalWidgetBounds.pos.y;
            real32 bottom = group->totalWidgetBounds.pos.y + group->totalWidgetBounds.size.y;
            newBounds.pos.y = top;
            newBounds.size.y = bottom - top;
        }

        // Recursively resize parent
        if (group->groupDepth > 0 && !HasFlag(group->flags, WidgetGroupFlag_Absolute)) {
            WidgetGroup *parentGroup = &manager->groupStack[group->groupDepth - 1];
            EtcRect groupBounds = {group->anchor + newBounds.pos, newBounds.size};
            CommitWidgetBounds(manager, parentGroup, groupBounds);
        }

        ResizeWidgetGroup_(manager, group, group->anchor, newBounds, group->localClipRect);
    }
}

// @NOTE: will always return a height of 0 because the widget needs to set that
EtcRect GetNextWidgetBounds(UIManager *manager, WidgetGroup *group) {

    if (group->nextWidgetBoundsOverride) {
        // @TODO: I don't like that this is here, should really get reset on the next pushed ID
        group->nextWidgetBoundsOverride = false;
        return group->nextWidgetBounds;
    }

    UIFlowDirection flowDirection = GetStyleFlowDirection(UIStyle_FlowDirection, manager);
    UIPadding padding = GetStylePadding(UIStyle_Padding, manager);

    EtcRect prevBounds = group->prevWidgetBounds;
    EtcRect nextBounds;

    switch (flowDirection) {
        case UIFlowDirection_Vertical: {
            nextBounds.pos.x = padding.left;
            nextBounds.pos.y = prevBounds.pos.y + prevBounds.size.y + padding.top;

            if (group->widgetCount > 1 ||
                // Padding can change between the tree header group
                // and the first element of the content, so the
                // padding is temporarily zeroed and handled by the
                // content group
                HasFlag(group->flags, WidgetGroupFlag_IsTreeContentGroup)) {
                nextBounds.pos.y += padding.bottom;
            }

            nextBounds.size.x = Max(0.0f, group->screenSize.x - (padding.left + padding.right));
            nextBounds.size.y = 0.0f;
        } break;

        case UIFlowDirection_Horizontal: {
            nextBounds.pos.x = prevBounds.pos.x + prevBounds.size.x + padding.left;
            nextBounds.pos.y = prevBounds.pos.y;

            if (group->widgetCount > 1) {
                nextBounds.pos.x += padding.right;
            }

            nextBounds.size.x = Max(0.0f, group->screenSize.x - padding.right - nextBounds.pos.x);
            nextBounds.size.y = 0.0f;
        } break;

        default: {
            ASSERT(false);
        } break;
    }

    return nextBounds;
}

EtcRect GetNextTextWidgetBounds(UIManager *manager, WidgetGroup *group, FontTable *font, real32 emSize, EtcRect *textBounds, vec2 *textPosition) {
    UIPadding insetMargin = GetStylePadding(UIStyle_InsetMargin, manager);
    real32 textHeight = Ceilf((font->ascent - font->descent) * emSize);

    b32 widgetBoundsOverride = group->nextWidgetBoundsOverride;
    EtcRect widgetBounds = GetNextWidgetBounds(manager, group);

    real32 minWidgetBoundsHeight = Roundf(textHeight + insetMargin.top + insetMargin.bottom);

    if (widgetBoundsOverride) {
        widgetBounds.size.y = Max(widgetBounds.size.y, minWidgetBoundsHeight);
    }
    else {
        widgetBounds.size.y = minWidgetBoundsHeight;
    }

    if (textBounds != NULL) {
        real32 textWidth = Max(0.0f, widgetBounds.size.x - (insetMargin.left + insetMargin.right));

        *textBounds = {widgetBounds.pos + V2(insetMargin.left, insetMargin.top),
                       {textWidth, textHeight}};
        textBounds->pos.x = Floorf(textBounds->pos.x);
        textBounds->pos.y = Floorf(textBounds->pos.y);
        textBounds->size.x = Ceilf(textBounds->size.x);
        textBounds->size.y = Ceilf(textBounds->size.y);
    }

    if (textPosition != NULL) {
        *textPosition = textBounds->pos + V2(0, Roundf(font->ascent * emSize));
    }

    return widgetBounds;
}

EtcRect GetNextDefaultTextWidgetBounds(UIManager *manager, WidgetGroup *group, EtcRect *textBounds, vec2 *textPosition) {
    FontTable *font = GetFontTable(Game.assets, GetStyleFont(UIStyle_Font, manager));
    if (font == NULL) {
        return {};
    }

    real32 emSize = GetStylef(UIStyle_FontSize, manager);

    return GetNextTextWidgetBounds(manager, group, font, emSize, textBounds, textPosition);
}

void SetNextWidgetBounds(UIManager *manager, WidgetGroup *group, EtcRect bounds) {
    group->nextWidgetBoundsOverride = true;
    group->nextWidgetBounds = bounds;
}

// @TODO: doesn't handle UTF-8!
// @TODO: be able to confine text to a box
void UILayoutGlyphAttributes(const char *text, uint32 textLength, FontTable *font, vec2 position, vec4 color, real32 emSize, GlyphAttributes *attribs, EtcRect *bounds = NULL) {

    ASSERT(font != NULL);

    static quaternion rotation = IdentityQuaternion();

    vec2 cursor = position;

    vec2 boundsMin = V2(INFINITY, INFINITY);
    vec2 boundsMax = V2(-INFINITY, -INFINITY);

    for (int i = 0; i < textLength; i++) {
        GlyphAttributes *attrib = &attribs[i];

        uint32 unicode = text[i];
        uint32 codepoint = GetInternalCodepoint(font, unicode);
        GlyphInfo info = GetGlyphInfo(font, codepoint);

        attrib->codepoint = codepoint;
        attrib->position = V3(cursor + emSize * V2(info.lowerLeft.x, -info.lowerLeft.y), 0.0f);
        attrib->rotation = rotation;
        attrib->dimensions = V2(info.size.x, -info.size.y) * emSize;
        attrib->color = color;
        attrib->scale = V2(1, 1);
        attrib->anchor = V2(0, -0.0f);

        // upperLeft/lowerRight are in UI space, so the upperLeft
        // coordinate is the min and the lowerRight coordinate is the
        // max
        vec2 upperLeft = cursor + V2(0.0f, emSize * -(info.lowerLeft.y + info.size.y));
        vec2 lowerRight = cursor + emSize * V2(info.lowerLeft.x + info.size.x, -info.lowerLeft.y);

        boundsMin = Min(boundsMin, upperLeft);
        boundsMax = Max(boundsMax, lowerRight);

        cursor.x += emSize * info.advance;
    }

    if (bounds != NULL) {
        *bounds = RectFromAABB(boundsMin, boundsMax);
    }
}


// Widget Groups
WidgetGroup *PushWidgetGroup_(UIManager *manager, WidgetID id, vec2 anchor, EtcRect localBounds, EtcRect localClipRect, uint32 flags) {
    ASSERT(manager != NULL);

    WidgetGroup *parentGroup = PeekWidgetGroup(manager);

    if (parentGroup != NULL) {
        ASSERT(HasFlag(flags, WidgetGroupFlag_Absolute) || !HasFlag(parentGroup->flags, WidgetGroupFlag_Is3D));
    }

    WidgetGroup *group = PushBackPtr(&manager->groupStack);
    Plat.ClearMem(group, sizeof(WidgetGroup));

    group->id = id;
    group->groupIndex = manager->nextGroupIndex++;
    group->groupDepth = manager->groupStack.count - 1;
    group->widgetCount = 0;
    group->prevID = id;
    group->prevWidgetBounds = {};

    group->flags = flags;

    if (parentGroup != NULL && !HasFlag(group->flags, WidgetGroupFlag_Absolute)) {
        EtcRect bounds = {anchor + localBounds.pos, localBounds.size};
        CommitWidgetBounds(manager, parentGroup, bounds);
    }

    ResizeWidgetGroup_(manager, group, anchor, localBounds, localClipRect);

    return group;
}

WidgetGroup *PushWidgetGroup(WidgetID id, vec2 anchor, EtcRect localBounds, EtcRect localClipRect, uint32 flags, UIManager *manager) {
    manager = ResolveUIManager(manager);
    WidgetGroup *group = PushWidgetGroup_(manager, id, anchor, localBounds, localClipRect, flags);

    group->view = Translation4(V3(group->screenPosition.x, group->screenPosition.y, 0.0f));
    group->projection = Orthographic(0, Game.graphics->width, Game.graphics->height, 0, -1, 1);
    group->viewProjection = group->projection * group->view;

    return group;
}

WidgetGroup *PushWidgetGroup3D(WidgetID id, vec2 anchor, EtcRect localBounds, CameraInfo cameraInfo, uint32 flags, UIManager *manager) {
    manager = ResolveUIManager(manager);
    flags = AddFlag(flags, WidgetGroupFlag_Is3D);
    WidgetGroup *group = PushWidgetGroup_(manager, id, anchor, localBounds, localBounds, flags);

    group->cameraInfo = cameraInfo;
    group->view = cameraInfo.view;
    group->projection = cameraInfo.projection;
    group->viewProjection = cameraInfo.viewProjection;


    vec2 localMousePositionNorm = Hadamard(group->localMousePosition, V2(1.0f / group->localBounds.size.x,
                                                                         1.0f / group->localBounds.size.y));
    localMousePositionNorm.y = 1.0f - localMousePositionNorm.y;
    group->mouseWorldRay = MousePositionToWorldRay(group->cameraInfo, localMousePositionNorm);

    return group;
}

WidgetGroup *PushWidgetGroup3D(WidgetID id, CameraInfo cameraInfo, uint32 flags, UIManager *manager) {
    EtcRect group3DBounds = {
        V2(0, 0),
        V2(Game.graphics->width, Game.graphics->height),
    };

    return PushWidgetGroup3D(id, V2(0, 0), group3DBounds, cameraInfo, flags, manager);
}

void PopWidgetGroup(UIManager *manager) {
    TIMED_BLOCK_LABEL("UI");
    
    manager = ResolveUIManager(manager);
    WidgetGroup *group = PeekWidgetGroup(manager);

    if (HasFlag(group->flags, WidgetGroupFlag_ResizeToContent)) {
        UIPadding padding = GetStylePadding(UIStyle_Padding, manager);

        if (HasFlag(group->flags, WidgetGroupFlag_ResizeToContentHorizontal)) {
            group->totalWidgetBounds.size.x += padding.right;
        }
        if (HasFlag(group->flags, WidgetGroupFlag_ResizeToContentVertical)) {
            group->totalWidgetBounds.size.y += padding.bottom;
        }
    }

    CommitWidgetBounds(manager, group, group->totalWidgetBounds);

    // Rendering setup
    // @TODO: compute this priority properly!
    uint32 key = GenerateUIRenderKey(UIRenderPriority_Default, group->groupIndex, true);

    if (HasFlag(group->flags, WidgetGroupFlag_Is3D)) {
        EtcRecti clipRect = group->clipRect;
        clipRect.pos.y = Game.graphics->height - (clipRect.pos.y + clipRect.size.y);
        EnableScissorTest(&manager->commandList, key, clipRect);
        SetViewport(&manager->commandList, key, clipRect);
    }
    else {
        EtcRecti clipRect = group->clipRect;
        clipRect.pos.y = Game.graphics->height - (clipRect.pos.y + clipRect.size.y);
        EnableScissorTest(&manager->commandList, key, clipRect);
        SetViewport(&manager->commandList, key, {V2i(0, 0), V2i(Game.graphics->width, Game.graphics->height)});
    }


    if (HasFlag(group->flags, WidgetGroupFlag_DrawBackground)) {
        vec4 color = GetStyle4(UIStyle_WindowBackgroundColor, manager);
        mat4 model = TS(V3(group->localBounds.pos, 0), V3(group->localBounds.size, 0));

        RenderMesh(manager, group, &Game.graphics->glyphQuad, model, color, true);
    }


    EtcRect bounds = {group->anchor + group->localBounds.pos, group->localBounds.size};
    RemoveLast(&manager->groupStack);

    if (!HasFlag(group->flags, WidgetGroupFlag_Absolute)) {
        CommitWidgetBounds(manager, PeekWidgetGroup(manager), bounds);
    }
}

WidgetGroup *PeekWidgetGroup(UIManager *manager) {
    manager = ResolveUIManager(manager);

    WidgetGroup *result = NULL;

    if (manager->groupStack.count > 0) {
        result = LastPtr(&manager->groupStack);
    }

    return result;
}

WidgetGroup *GetParentGroup(UIManager *manager, WidgetGroup *group) {
    if (group->groupDepth > 0) {
        WidgetGroup *parent = &manager->groupStack[group->groupDepth - 1];
        return parent;
    }

    return NULL;
}

WidgetGroup *GetAncestorGroupWithFlags(UIManager *manager, WidgetGroup *group, uint32 flags) {

    while (group->groupDepth > 0) {
        WidgetGroup *parent = &manager->groupStack[group->groupDepth - 1];

        if (HasFlag(parent->flags, flags)) {
            return parent;
        }

        group = parent;
    }
    
    return NULL;
}


// Rendering
uint8 GetShaderIndex(WidgetGroup *group, ShaderProgram *shader, UIBatchType batchType, b32 *addedShader) {
    ASSERT(shader != NULL);

    uint32 index;
    for (int i = 0; i < group->shaderCount; i++) {
        if (group->shaders[i] == shader && group->shaderBatchTypes[i] == batchType) {
            index = (uint8)i;
            *addedShader = false;
            return index;
        }
    }

    ASSERT(group->shaderCount < UI_MAX_SHADERS_PER_GROUP);
    index = group->shaderCount;
    group->shaders[index] = shader;
    group->shaderBatchTypes[index] = batchType;

    group->shaderCount++;

    *addedShader = true;

    return index;
}

uint8 GetBatchAssetIndex(WidgetGroup *group, UIBatchType batchType, void *asset, b32 *addedAsset) {
    ASSERT(asset != NULL);

    uint32 assetCount = group->batchAssetCount[batchType];
    void **assets = group->batchAssets[batchType];

    uint32 index;
    for (int i = 0; i < assetCount; i++) {
        if (assets[i] == asset) {
            index = (uint8)i;
            *addedAsset = false;
            return index;
        }
    }

    ASSERT(assetCount < UI_MAX_BATCHES_PER_GROUP);
    index = (uint8)assetCount;
    group->batchAssets[batchType][group->batchAssetCount[batchType]++] = asset;
    *addedAsset = true;

    return index;
}

b32 UseShaderProgramInGroup(UIManager *manager, WidgetGroup *group, ShaderProgram *shader, UIBatchType batchType, uint8 *outShaderIndex) {
    TIMED_BLOCK_LABEL("UI");
    
    if (shader == NULL) {
        return false;
    }

    b32 addedShader;
    uint8 shaderIndex = GetShaderIndex(group, shader, batchType, &addedShader);

    if (shaderIndex < (1 << 3)) {
        if (addedShader) {
            // @TODO: compute this priority properly!
            UIRenderPriority priority = UIRenderPriority_Default;
            uint32 setUniformRenderKey = GenerateUIRenderKey(priority, group->groupIndex, false, shaderIndex, true, batchType);

            SetShader(&manager->commandList, setUniformRenderKey, shader->programID);
            SetUniform(&manager->commandList, setUniformRenderKey, UniformLocation(shader, "viewProjection"), group->viewProjection);
        }

        if (outShaderIndex != NULL) {
            *outShaderIndex = shaderIndex;
        }

        return true;
    }

    return false;
}

b32 UseMeshInGroup(UIManager *manager, WidgetGroup *group, ShaderProgram *shader, Mesh *mesh, b32 priorityRender, uint32 *renderKey = NULL) {
    TIMED_BLOCK_LABEL("UI");
    
    if (shader == NULL || mesh == NULL) {
        return false;
    }

    uint8 shaderIndex;
    if (!UseShaderProgramInGroup(manager, group, shader, UIBatchType_Mesh, &shaderIndex)) {
        return false;
    }

    b32 addedAsset;
    uint8 batchIndex = GetBatchAssetIndex(group, UIBatchType_Mesh, mesh, &addedAsset);

    if (batchIndex < (1 << 7)) {
        // @TODO: compute this priority properly!
        UIRenderPriority priority = UIRenderPriority_Default;

        b32 newShaderForAsset = !(group->batchAssetUsedShader[UIBatchType_Mesh][batchIndex] & (1 << shaderIndex));

        if (addedAsset || newShaderForAsset) {
            uint32 setupKey = GenerateUIRenderKey(priority, group->groupIndex, false, shaderIndex, false, UIBatchType_Mesh, batchIndex, UIBatchRenderMode_Setup);

            uint32 teardownKey = GenerateUIRenderKey(priority, group->groupIndex, false, shaderIndex, false, UIBatchType_Mesh, batchIndex, UIBatchRenderMode_Teardown);


            BindBuffer(&manager->commandList, setupKey, GL_ARRAY_BUFFER, mesh->vboID);
            BindBuffer(&manager->commandList, setupKey, GL_ELEMENT_ARRAY_BUFFER, mesh->iboID);

            if (mesh->positions != NULL) {
                EnableVertexAttribute(&manager->commandList, setupKey, Position_Attrib, 3, GL_FLOAT, GL_FALSE, 0,
                                      POINTER_OFFSET(mesh->vertices, mesh->positions));
                DisableVertexAttribute(&manager->commandList, teardownKey, Position_Attrib);
            }

            if (mesh->texcoords != NULL) {
                EnableVertexAttribute(&manager->commandList, setupKey, TexCoord_Attrib, 2, GL_FLOAT, GL_FALSE, 0,
                                      POINTER_OFFSET(mesh->vertices, mesh->texcoords));
                DisableVertexAttribute(&manager->commandList, teardownKey, TexCoord_Attrib);
            }

            if (mesh->normals != NULL) {
                EnableVertexAttribute(&manager->commandList, setupKey, Normal_Attrib, 3, GL_FLOAT, GL_FALSE, 0,
                                      POINTER_OFFSET(mesh->vertices, mesh->normals));
                DisableVertexAttribute(&manager->commandList, teardownKey, Normal_Attrib);
            }

            if (mesh->colors != NULL) {
                EnableVertexAttribute(&manager->commandList, setupKey, Color_Attrib, 4, GL_FLOAT, GL_FALSE, 0,
                                      POINTER_OFFSET(mesh->vertices, mesh->colors));
                DisableVertexAttribute(&manager->commandList, teardownKey, Color_Attrib);
            }

            group->batchAssetUsedShader[UIBatchType_Mesh][batchIndex] |= (1 << shaderIndex);
        }

        if (renderKey != NULL) {
            UIBatchRenderMode renderMode = UIBatchRenderMode_Render;
            if (priorityRender) {
                renderMode = UIBatchRenderMode_PriorityRender;
            }
            *renderKey = GenerateUIRenderKey(priority, group->groupIndex, false, shaderIndex, false, UIBatchType_Mesh, batchIndex, renderMode);
        }

        return true;
    }

    return false;
}

b32 UseFontTableInGroup(UIManager *manager, WidgetGroup *group, ShaderProgram *shader, FontTable *font, uint32 *renderKey = NULL) {
    TIMED_BLOCK_LABEL("UI");
    
    if (shader == NULL || font == NULL) {
        return false;
    }

    uint8 shaderIndex;
    if (!UseShaderProgramInGroup(manager, group, shader, UIBatchType_Font, &shaderIndex)) {
        return false;
    }

    b32 addedAsset;
    uint8 batchIndex = GetBatchAssetIndex(group, UIBatchType_Font, font, &addedAsset);

    if (batchIndex < (1 << 7)) {
        // @TODO: compute this priority properly!
        UIRenderPriority priority = UIRenderPriority_Default;

        b32 newShaderForAsset = !(group->batchAssetUsedShader[UIBatchType_Font][batchIndex] & (1 << shaderIndex));

        if (addedAsset || newShaderForAsset) {
            uint32 setupKey = GenerateUIRenderKey(priority, group->groupIndex, false, shaderIndex, false, UIBatchType_Font, batchIndex, UIBatchRenderMode_Setup);

            SetUniform(&manager->commandList, setupKey, UniformLocation(shader, "model"), Identity4());

            SetUniform(&manager->commandList, setupKey, UniformLocation(shader, "texcoordsMap"), 0, font->texcoordsMapID, GL_TEXTURE_1D);
            SetSampler(&manager->commandList, setupKey, 0, Game.graphics->nearestClampEdge);

            SetUniform(&manager->commandList, setupKey, UniformLocation(shader, "atlas"), 1, font->atlas.textureID, GL_TEXTURE_2D);
            SetSampler(&manager->commandList, setupKey, 1, Game.graphics->linearClampEdge);


            // Bind GlyphQuad
            BindBuffer(&manager->commandList, setupKey, GL_ARRAY_BUFFER, Game.graphics->glyphQuad.vboID);
            BindBuffer(&manager->commandList, setupKey, GL_ELEMENT_ARRAY_BUFFER, Game.graphics->glyphQuad.iboID);

            EnableVertexAttribute(&manager->commandList, setupKey, Position_Attrib, 3, GL_FLOAT, GL_FALSE, 0,
                                  POINTER_OFFSET(Game.graphics->glyphQuad.vertices, Game.graphics->glyphQuad.positions));

            EnableVertexAttribute(&manager->commandList, setupKey, TexCoord_Attrib, 2, GL_FLOAT, GL_FALSE, 0,
                                  POINTER_OFFSET(Game.graphics->glyphQuad.vertices, Game.graphics->glyphQuad.texcoords));

            // Bind GlyphAttributes buffer
            BindBuffer(&manager->commandList, setupKey, GL_ARRAY_BUFFER, manager->textBufferID);
            EnableVertexIAttribute(&manager->commandList, setupKey, GlyphCodepoint_Attrib, 1, GL_UNSIGNED_INT,
                                   sizeof(GlyphAttributes), FIELD_OFFSET(GlyphAttributes, codepoint), 1);

            EnableVertexAttribute(&manager->commandList, setupKey, GlyphPosition_Attrib, 3, GL_FLOAT, GL_FALSE,
                                  sizeof(GlyphAttributes), FIELD_OFFSET(GlyphAttributes, position), 1);

            EnableVertexAttribute(&manager->commandList, setupKey, Rotation_Attrib, 4, GL_FLOAT, GL_FALSE,
                                  sizeof(GlyphAttributes), FIELD_OFFSET(GlyphAttributes, rotation), 1);

            EnableVertexAttribute(&manager->commandList, setupKey, GlyphDimensions_Attrib, 2, GL_FLOAT, GL_FALSE,
                                  sizeof(GlyphAttributes), FIELD_OFFSET(GlyphAttributes, dimensions), 1);

            EnableVertexAttribute(&manager->commandList, setupKey, Color_Attrib, 4, GL_FLOAT, GL_FALSE,
                                  sizeof(GlyphAttributes), FIELD_OFFSET(GlyphAttributes, color), 1);

            
            EnableVertexAttribute(&manager->commandList, setupKey, Scale_Attrib, 2, GL_FLOAT, GL_FALSE,
                                  sizeof(GlyphAttributes), FIELD_OFFSET(GlyphAttributes, scale), 1);

            EnableVertexAttribute(&manager->commandList, setupKey, GlyphAnchor_Attrib, 2, GL_FLOAT, GL_FALSE,
                                  sizeof(GlyphAttributes), FIELD_OFFSET(GlyphAttributes, anchor), 1);


            uint32 teardownKey = GenerateUIRenderKey(priority, group->groupIndex, false, shaderIndex, false, UIBatchType_Font, batchIndex, UIBatchRenderMode_Teardown);

            DisableVertexAttribute(&manager->commandList, teardownKey, GlyphCodepoint_Attrib);
            DisableVertexAttribute(&manager->commandList, teardownKey, GlyphPosition_Attrib);
            DisableVertexAttribute(&manager->commandList, teardownKey, Rotation_Attrib);
            DisableVertexAttribute(&manager->commandList, teardownKey, GlyphDimensions_Attrib);
            DisableVertexAttribute(&manager->commandList, teardownKey, Color_Attrib);
            DisableVertexAttribute(&manager->commandList, teardownKey, Scale_Attrib);
            DisableVertexAttribute(&manager->commandList, teardownKey, GlyphAnchor_Attrib);

            DisableVertexAttribute(&manager->commandList, teardownKey, Position_Attrib);
            DisableVertexAttribute(&manager->commandList, teardownKey, TexCoord_Attrib);


            group->batchAssetUsedShader[UIBatchType_Font][batchIndex] |= (1 << shaderIndex);
        }

        if (renderKey != NULL) {
            *renderKey = GenerateUIRenderKey(priority, group->groupIndex, false, shaderIndex, false, UIBatchType_Font, batchIndex, UIBatchRenderMode_Render);
        }

        return true;
    }

    return false;
}

b32 RenderMesh(UIManager *manager, WidgetGroup *group, Mesh *mesh, Texture *texture0, mat4 model, vec4 color, b32 priorityRender) {
    TIMED_BLOCK_LABEL("UI");
    
    ShaderProgram *shader;
    if (texture0 != NULL) {
        shader = GetShaderProgram(Game.assets, GetStyleShaderProgram(UIStyle_TexturedMeshShader, manager));
    }
    else {
        shader = GetShaderProgram(Game.assets, GetStyleShaderProgram(UIStyle_MeshShader, manager));
    }

    uint32 renderKey;
    if (UseMeshInGroup(manager, group, shader, mesh, priorityRender, &renderKey)) {
        SetUniform(&manager->commandList, renderKey, UniformLocation(shader, "model"), model);
        SetUniform(&manager->commandList, renderKey, UniformLocation(shader, "color"), color);

        if (texture0 != NULL) {
            SetUniform(&manager->commandList, renderKey, UniformLocation(shader, "texture0"), 0, texture0->textureID, GL_TEXTURE_2D);
            SetSampler(&manager->commandList, renderKey, 0, Game.graphics->linearClampEdge);
        }

        DrawIndexed(&manager->commandList, renderKey, mesh->primitiveType, 0, mesh->indexCount);

        return true;
    }

    return false;
}

b32 RenderMesh(UIManager *manager, WidgetGroup *group, Mesh *mesh, mat4 model, vec4 color, b32 priorityRender) {
    return RenderMesh(manager, group, mesh, NULL, model, color, priorityRender);
}


// @NOTE: glyphs buffer has to be valid until the end of the frame because the render commands are deferred
b32 RenderText(UIManager *manager, WidgetGroup *group, FontTable *font, GlyphAttributes *glyphs, uint32 glyphCount) {
    TIMED_BLOCK_LABEL("UI");
    
    if (glyphCount == 0) {
        return false;
    }

    ShaderProgram *shader = GetShaderProgram(Game.assets, GetStyleShaderProgram(UIStyle_TextShader, manager));
    uint32 renderKey;
    if (UseFontTableInGroup(manager, group, shader, font, &renderKey)) {

        uint32 glyphBufferSize = glyphCount * sizeof(GlyphAttributes);

        // There's no need to bind the manager->textBufferID because
        // the setup for the batch already took care of that
        CopyBufferData(&manager->commandList, renderKey, GL_ARRAY_BUFFER, glyphBufferSize, glyphs, GL_STREAM_DRAW);

        // The batch setup also set up the glyphQuad for position/texcoord attributes
        DrawIndexedInstanced(&manager->commandList, renderKey, PrimitiveType_Triangles, 0, 6, glyphCount);

        return true;
    }

    return false;
}


#if ETC_DEBUG
void UIDebugRenderPrevWidgetBounds(UIManager *manager, WidgetGroup *group, vec4 color) {
    EtcRect bounds = group->prevWidgetBounds;
    mat4 model = TS(V3(bounds.pos, 0), V3(bounds.size, 0));
    RenderMesh(manager, group, &Game.graphics->glyphQuad, model, color);
}

void UIDebugRenderLocalBounds(UIManager *manager, WidgetGroup *group, vec4 color) {
    EtcRect bounds = group->localBounds;
    mat4 model = TS(V3(bounds.pos, 0), V3(bounds.size, 0));
    RenderMesh(manager, group, &Game.graphics->glyphQuad, model, color);
}
#endif


////////////////////////////////////////////////////////////////////////////////
//                                   WIDGETS                                  //
////////////////////////////////////////////////////////////////////////////////


// Windows
void PushWindow(const char *name, vec2 *position, vec2 *size, bool *collapsed, uint32 flags, const char *title) {
    UIManager *manager = GetCurrentUIManager();
    WidgetGroup *parentGroup = PeekWidgetGroup(manager);

    manager->showMouse = true;

    WidgetID id = HasFlag(flags, UIWindowFlag_ReuseWidgetID_) ?
        parentGroup->prevID :
        PushWidgetID(manager, name, false);

    uint32 groupFlags = WidgetGroupFlag_BlocksMouse | WidgetGroupFlag_DrawBackground;
    if (HasFlag(flags, UIWindowFlag_ResizeToContentVertical)) {
        groupFlags = AddFlag(groupFlags, WidgetGroupFlag_ResizeToContentVertical);
    }
    if (HasFlag(flags, UIWindowFlag_ResizeToContentHorizontal)) {
        groupFlags = AddFlag(groupFlags, WidgetGroupFlag_ResizeToContentHorizontal);
    }

    if (HasFlag(flags, UIWindowFlag_Absolute)) {
        groupFlags = AddFlag(groupFlags, WidgetGroupFlag_Absolute);
    }

    EtcRect localBounds = {
        V2(0, 0),
        *size
    };

    if (collapsed != NULL && *collapsed) {
        real32 emSize = GetStylef(UIStyle_FontSize, manager);
        FontTable *font = GetFontTable(Game.assets, GetStyleFont(UIStyle_Font, manager));
        EtcRect textBounds;
        vec2 textPosition;
        EtcRect titleBarBounds = GetNextTextWidgetBounds(manager, parentGroup, font, emSize, &textBounds, &textPosition);
        
        localBounds.size.y = titleBarBounds.size.y;
    }
    
    WidgetGroup *group = PushWidgetGroup(id, *position, localBounds, localBounds, groupFlags, manager);


    if (HasFlag(flags, UIWindowFlag_ShowTitle) || HasFlag(flags, UIWindowFlag_Freemove) || HasFlag(flags, UIWindowFlag_Collapsable)) {
        UIStyle titleBarStyles[] = {
            MakeStylePadding(UIStyle_Padding, {0, 0, 0, 0}),
            MakeStyle4(UIStyle_WindowBackgroundColor, {0.25f, 0.25f, 0.25f, 1.0f}),
        };
        PushStyles(titleBarStyles, ARRAY_LENGTH(UIStyle, titleBarStyles));

        WidgetID titleBarID = PushWidgetID(manager, name);
        EtcRect titleBarBounds = {
            V2(0, 0),
            V2(localBounds.size.x, 0)
        };
        uint32 titleBarFlags = WidgetGroupFlag_ResizeToContentVertical | WidgetGroupFlag_DrawBackground;
        WidgetGroup *barGroup = PushWidgetGroup(titleBarID, V2(0, 0), titleBarBounds, titleBarBounds, titleBarFlags);

        if (HasFlag(flags, UIWindowFlag_ShowTitle)) {
            if (title != NULL) {
                PushText(title);
            }
            else {
                PushText(name);
            }
        }
        else {
            PushText("");
        }

        struct WindowActiveData {
            vec2 startMousePositionParentGroup;
            vec2 startPosition;
            real32 prevMouseReleaseTime;
        };

        if (HasFlag(flags, UIWindowFlag_Freemove) || HasFlag(flags, UIWindowFlag_Collapsable)) {
            EtcRect barGroupBounds = barGroup->localBounds;
            
            b32 isMouseOver = barGroup->isHoveringGroup && PointInRect(barGroupBounds, barGroup->localMousePosition);
            bool hold = InputActionHold(Game.inputActions, InputAction_EditorPrimaryClick, true);
            bool clicked = InputActionPressed(Game.inputActions, InputAction_EditorPrimaryClick, true);
            bool released = InputActionReleased(Game.inputActions, InputAction_EditorPrimaryClick, true);

            bool doubleClickThreshold = GetStylef(UIStyle_DoubleClickThreshold);

            if (IsActive(manager, titleBarID)) {
                WindowActiveData *windowData = (WindowActiveData *)manager->activeWidgetData;
                
                if (released) {
                    windowData->prevMouseReleaseTime = Game.time;
                }

                bool missedDoubleClickChance = (real32)(Game.time - windowData->prevMouseReleaseTime) > (real32)doubleClickThreshold;
                
                if (hold && !clicked && HasFlag(flags, UIWindowFlag_Freemove)) {
                    vec2 diff = parentGroup->localMousePosition - windowData->startMousePositionParentGroup;
                    *position = windowData->startPosition + diff; 
                }
                else if (HasFlag(flags, UIWindowFlag_Collapsable) && clicked && !missedDoubleClickChance) {
                    *collapsed = !(*collapsed);
                    SetActiveWidget(manager, WidgetID_Invalid);
                }

                if (!hold && missedDoubleClickChance) {
                    SetActiveWidget(manager, WidgetID_Invalid);
                }
            }
            else if (isMouseOver) {
                if (IsHovered(manager, titleBarID)) {
                    if (clicked) {
                        SetActiveWidget(manager, titleBarID, sizeof(WindowActiveData));
                        WindowActiveData *windowData = (WindowActiveData *)manager->activeWidgetData;
                        windowData->startPosition = *position;
                        windowData->startMousePositionParentGroup = parentGroup->localMousePosition;
                    }
                }

                SetHoveredWidget(manager, group, titleBarID);
            }

        }
        

        PopWidgetGroup();
        PopStyles();

        // @TODO: render a draggable handle at the bottom of a window. Currently no idea how to do this
#if 0        
        if (HasFlag(flags, UIWindowFlag_Freescale)) {
            UIStyle scaleNubStyles[] = {
                MakeStylePadding(UIStyle_Padding, {0, 0, 0, 0}),
                MakeStyle4(UIStyle_WindowBackgroundColor, {0.25f, 0.25f, 0.25f, 1.0f}),
            };
            PushStyles(scaleNubStyles, ARRAY_LENGTH(UIStyle, scaleNubStyles));
        
            WidgetID scaleNubID = PushWidgetID(manager, "Scale");
            EtcRect scaleNubBounds = {
                V2(0, 0),
                V2(100, 100)
            };
            uint32 scaleNubFlags = WidgetGroupFlag_ResizeToContentVertical | WidgetGroupFlag_DrawBackground;
            WidgetGroup *scaleGroup = PushWidgetGroup(scaleNubID, V2(0, 0), scaleNubBounds, scaleNubBounds, scaleNubFlags);

            Button("YEHA");

            UIDebugRenderPrevWidgetBounds(manager, scaleGroup, YELLOW);
            //UIDebugRenderLocalBounds(manager, scaleGroup, RED);

            PopStyles();
            PopWidgetGroup();
        }
#endif
    }
}

void PushWindow(const char *name, vec2 *position, vec2 size, bool *collapsed, uint32 flags, const char *title) {
    PushWindow(name, position, &size, collapsed, flags, title);
}

void PushWindow(const char *name, vec2 position, vec2 size, uint32 flags, const char *title) {
    bool collapsed = false;
    PushWindow(name, &position, &size, &collapsed, flags, title);
}

void PushWindow(const char *name, vec2 *position, vec2 size, uint32 flags, const char *title) {
    bool collapsed = false;
    PushWindow(name, position, &size, &collapsed, flags, title);
}

void PushWindow(const char *name, vec2 *position, vec2 *size, uint32 flags, const char *title) {
    bool collapsed = false;
    PushWindow(name, position, size, &collapsed, flags, title);
}



void PopWindow() {
    UIManager *manager = GetCurrentUIManager();
    WidgetGroup *group = PeekWidgetGroup(manager);

    PopWidgetGroup(manager);
}


struct TreeGroupPersistentData {
    b32 collapsed;
};

b32 DefaultTreeHeader(const char *name, b32 collapsed, void *data) {
    UIManager *manager = GetCurrentUIManager();
    WidgetGroup *currentGroup = PeekWidgetGroup(manager);
    WidgetID id = PushWidgetID(manager, name);

    char displayText[128];
    stbsp_snprintf(displayText, ARRAY_LENGTH(char, displayText), "%s %s", collapsed ? "+" : "-", name);

    vec4 buttonColor = PASTEL_BLUE;

    PushStyle(MakeStyle4(UIStyle_ButtonColor, buttonColor));

    if (Button(name, 0, displayText)) {
        collapsed = !collapsed;
    }

    PopStyles();

    return collapsed;
}

b32 PushTreeGroup(const char *name, InputTreeHeaderFunc *headerFunc, void *headerData, uint32 flags) {
    TIMED_BLOCK_LABEL("UI");
    
    UIManager *manager = GetCurrentUIManager();
    WidgetGroup *currentGroup = PeekWidgetGroup(manager);
    WidgetID id = PushWidgetID(manager, name);

    static uint32 TreeGroupDataID = GenerateDataID("TreeGroupPersistentData");

    TreeGroupPersistentData *persistentData = NULL;
    if (!GetPersistentData(manager, id, TreeGroupDataID, (void **)&persistentData)) {
        persistentData = (TreeGroupPersistentData *)AllocatePersistentData(manager, id, TreeGroupDataID, sizeof(TreeGroupPersistentData));

        persistentData->collapsed = HasFlag(flags, UITreeGroupFlag_StartCollapsed);
    }

    UIPadding padding = GetStylePadding(UIStyle_Padding);

    // Group header
    WidgetGroup *headerGroup;
    {
        EtcRect bounds = GetNextWidgetBounds(manager, currentGroup);
        EtcRect localBounds = {V2(0, 0), bounds.size};

        bounds.pos.x -= padding.left;
        localBounds.size.x += padding.left + padding.right;

        uint32 flags = WidgetGroupFlag_IsTreeHeaderGroup | WidgetGroupFlag_ResizeToContentVertical;
        headerGroup = PushWidgetGroup(id, bounds.pos, localBounds, localBounds, flags, manager);


        UIPadding tempPadding = GetStylePadding(UIStyle_Padding);
        tempPadding.top = 0.0f;

        PushStyle(MakeStylePadding(UIStyle_Padding, tempPadding));

        persistentData->collapsed = headerFunc(name, persistentData->collapsed, headerData);

        PopStyles();
    }

    // Content Group
    {
        UIPadding tempPadding = padding;
        tempPadding.top = 0.0f;
        tempPadding.bottom = 0.0f;
        PushStyle(MakeStylePadding(UIStyle_Padding, tempPadding));


        EtcRect bounds = GetNextWidgetBounds(manager, headerGroup);
        EtcRect localBounds = {V2(0, 0), bounds.size};

        localBounds.size.x += padding.right;


        uint32 flags = WidgetGroupFlag_IsTreeContentGroup | WidgetGroupFlag_ResizeToContentVertical;
        WidgetGroup *contentGroup = PushWidgetGroup(id, bounds.pos, localBounds, localBounds, flags, manager);

        PopStyles();
    }

    // If the group is collapsed no other widgets can go into the group
    if (persistentData->collapsed) {
        UIPadding tempPadding = GetStylePadding(UIStyle_Padding);
        tempPadding.top = 0.0f;
        tempPadding.bottom = 0.0f;

        PushStyle(MakeStylePadding(UIStyle_Padding, tempPadding));

        PopWidgetGroup(manager);
        PopWidgetGroup(manager);

        PopStyles();
    }
    
    return !persistentData->collapsed;
}

b32 PushTreeGroup(const char *name, uint32 flags) {
    return PushTreeGroup(name, DefaultTreeHeader, NULL, flags);
}


void PopTreeGroup() {
    UIManager *manager = GetCurrentUIManager();

    WidgetGroup *contentGroup = PeekWidgetGroup(manager);
    ASSERT(HasFlag(contentGroup->flags, WidgetGroupFlag_IsTreeContentGroup));

    WidgetGroup *headerGroup = GetParentGroup(manager, contentGroup);
    ASSERT(HasFlag(headerGroup->flags, WidgetGroupFlag_IsTreeHeaderGroup));

    WidgetGroup *headerParent = GetParentGroup(manager, headerGroup);

    UIPadding tempPadding = GetStylePadding(UIStyle_Padding);
    tempPadding.bottom = 0.0f;

    PushStyle(MakeStylePadding(UIStyle_Padding, tempPadding));

    // Pop content group
    PopWidgetGroup(manager);

    // Pop header group
    PopWidgetGroup(manager);

    PopStyles();

#if 0
    UIDebugRenderLocalBounds(manager, headerGroup, WithAlpha(RED, 0.125f));
    UIDebugRenderLocalBounds(manager, contentGroup, WithAlpha(BLUE, 0.125f));
#endif
}


b32 PushContextMenu(const char *name, WidgetID ownerID, ContextMenuData *menuData, uint32 flags) {
    UIManager *manager = GetCurrentUIManager();
    b32 isOpen = false;

    if (ownerID != WidgetID_Invalid && ownerID == menuData->ownerID && !menuData->renderedThisFrame) {
        EtcRect windowBounds;
        windowBounds.pos = menuData->pos;
        windowBounds.size = menuData->size;

        uint32 windowFlags = UIWindowFlag_Absolute | UIWindowFlag_ResizeToContentVertical;
        PushWindow(name, windowBounds.pos, windowBounds.size, windowFlags);

        WidgetGroup *windowGroup = PeekWidgetGroup(manager);
        menuData->windowID = windowGroup->id;

        isOpen = true;
        menuData->renderedThisFrame = true;
    }

    return isOpen;
}

b32 PushContextMenuPrevWidget(const char *name, ContextMenuData *menuData, uint32 flags) {
    UIManager *manager = GetCurrentUIManager();
    WidgetGroup *group = PeekWidgetGroup(manager);
    WidgetID prevID = group->prevID;

    if (IsHovered(manager, prevID)) {
        if (InputActionPressed(Game.inputActions, InputAction_EditorSecondaryClick, true)) {
            menuData->ownerID = prevID;
            menuData->pos = manager->screenMousePosition;
            menuData->size = {256, 0};
        }
    }

    return PushContextMenu(name, prevID, menuData, flags);
}

void CloseContextMenu(ContextMenuData *menuData) {
    menuData->ownerID = WidgetID_Invalid;
}

void PopContextMenu(ContextMenuData *menuData) {
    UIManager *manager = GetCurrentUIManager();
    WidgetGroup *windowGroup = PeekWidgetGroup(manager);

    ASSERT(windowGroup->id == menuData->windowID);

    if ((InputActionPressed(Game.inputActions, InputAction_EditorPrimaryClick, true) ||
         InputActionPressed(Game.inputActions, InputAction_EditorSecondaryClick, true)) &&
        !windowGroup->isHoveringGroup) {
        menuData->ownerID = WidgetID_Invalid;
    }

    PopWindow();
}


EtcRect ClampScrollGroupContentBounds(EtcRect visibleBounds, EtcRect contentBounds) {
    visibleBounds.pos.x = Clamp(visibleBounds.pos.x,
                                contentBounds.pos.x,
                                Max(0.0f, contentBounds.pos.x + contentBounds.size.x - visibleBounds.size.x));

    // @BUG: this prevents us from actually reaching the bottom of the scroll group and causes this terrible
    // oscillation when we keep scrolling where its at the bottom for a frame (from the PopScrollGroup compute)
    // and then bounces back up here...
    visibleBounds.pos.y = Clamp(visibleBounds.pos.y,
                                contentBounds.pos.y,
                                Max(0.0f, contentBounds.pos.y + contentBounds.size.y - visibleBounds.size.y));

    return visibleBounds;
}

struct UIScrollBarActiveData {
    vec2 mouseDeltaFromNubCenter;
};

b32 ScrollBar(const char *name, EtcRect scrollbarBounds, EtcRect contentBounds, EtcRect *visibleBounds, uint32 flags) {
    UIManager *manager = GetCurrentUIManager();
    WidgetGroup *group = PeekWidgetGroup(manager);
    WidgetID id = PushWidgetID(manager, name);

    b32 viewChanged = false;

    vec2 posRatio = V2(0, 0);
    vec2 sizeRatio = V2(1, 1);
    vec2 posMax = V2(0, 0);

    if (HasFlag(flags, UIScrollBarFlag_Vertical)) {
        if (contentBounds.size.y > 0.0f) {
            posRatio.y = (visibleBounds->pos.y - contentBounds.pos.y) / contentBounds.size.y;
            sizeRatio.y = Min(visibleBounds->size.y / contentBounds.size.y, 1.0f);

            posMax.y = scrollbarBounds.size.y * (1 - sizeRatio.y);
        }
    }

    if (HasFlag(flags, UIScrollBarFlag_Horizontal)) {
        if (contentBounds.size.x > 0.0f) {
            posRatio.x = (visibleBounds->pos.x - contentBounds.pos.x) / contentBounds.size.x;
            sizeRatio.x = Min(visibleBounds->size.x / contentBounds.size.x, 1.0f);

            posMax.x = scrollbarBounds.size.x * (1 - sizeRatio.x);
        }
    }

    EtcRect nubBounds = {Hadamard(scrollbarBounds.size, posRatio),
                         Hadamard(scrollbarBounds.size, sizeRatio)};
    vec2 nubCenter = nubBounds.pos + 0.5f * nubBounds.size;

    b32 isActive = false;
    b32 mouseOverNub = group->isHoveringGroup && PointInRect(nubBounds, group->localMousePosition);

    if (IsActive(manager, id)) {
        if (InputActionReleased(Game.inputActions, InputAction_EditorPrimaryClick, true)) {
            SetActiveWidget(manager, WidgetID_Invalid);
        }
        else {
            UIScrollBarActiveData *activeData = (UIScrollBarActiveData *)manager->activeWidgetData;

            vec2 newCenter = group->localMousePosition - activeData->mouseDeltaFromNubCenter;
            vec2 newPos = newCenter - 0.5f * nubBounds.size;
            newPos = Clamp(newPos, V2(0, 0), posMax);

            vec2 newPosRatio = Hadamard(1.0f / scrollbarBounds.size, newPos);
            newPos = contentBounds.pos + Hadamard(newPosRatio, contentBounds.size);

            if (newPos != visibleBounds->pos) {
                viewChanged = true;
                
                if (HasFlag(flags, UIScrollBarFlag_Horizontal) && !
                    HasFlag(flags, UIScrollBarFlag_Vertical)) {
                    newPos.y = visibleBounds->pos.y;
                }

                if (HasFlag(flags, UIScrollBarFlag_Vertical) &&
                    !HasFlag(flags, UIScrollBarFlag_Horizontal)) {
                    newPos.x = visibleBounds->pos.x;
                }
                
                visibleBounds->pos = newPos;
                
                *visibleBounds = ClampScrollGroupContentBounds(*visibleBounds, contentBounds);
            }

            isActive = true;
        }
    }
    // @BUG: it's possible that the scrollbar is rendered over another widget, in which case
    // IsHovered() will not be true
    // @THEFIX: (probably) introduce a priority system for widgets so some widgets always override (or maybe the override should be on the group?)
    else if (mouseOverNub) {
        if (IsHovered(manager, id)) {
            if (InputActionPressed(Game.inputActions, InputAction_EditorPrimaryClick, true)) {
                isActive = true;
                SetActiveWidget(manager, id, sizeof(UIScrollBarActiveData));

                UIScrollBarActiveData *activeData = (UIScrollBarActiveData *)manager->activeWidgetData;
                activeData->mouseDeltaFromNubCenter = group->localMousePosition - nubCenter;
            }
        }

        SetHoveredWidget(manager, group, id);
    }


    vec4 backgroundColor = V4(0.0f, 0.0f, 0.0f, 0.75f);
    vec4 nubColor = isActive ? PASTEL_YELLOW : PASTEL_RED;

    UIDebugRenderLocalBounds(manager, group, backgroundColor);
    RenderMesh(manager, group, &Game.graphics->glyphQuad, TS(nubBounds.pos, nubBounds.size), nubColor);

    return viewChanged;
}

struct ScrollGroupPersistentData {
    EtcRect visibleBounds;

    b32 childIsScrolling;
};

static uint32 ScrollGroupDataID = GenerateDataIDString("ScrollGroupPersistentData");


void PushScrollGroup(const char *name, EtcRect localBounds, uint32 flags) {
    UIManager *manager = GetCurrentUIManager();
    WidgetID id = PushWidgetID(manager, name);


    ScrollGroupPersistentData *persistentData = NULL;
    if (!GetPersistentData(manager, id, ScrollGroupDataID, (void **)&persistentData)) {
        persistentData = (ScrollGroupPersistentData *)AllocatePersistentData(manager, id, ScrollGroupDataID, sizeof(ScrollGroupPersistentData));
        persistentData->visibleBounds = {};
    }

    persistentData->childIsScrolling = false;

    uint32 groupFlags = AddFlag(flags, WidgetGroupFlag_IsScrollGroup);
    EtcRect scrollGroupBounds = localBounds;
    EtcRect scrollGroupLocalBounds = {V2(0, 0), scrollGroupBounds.size};
    WidgetGroup *scrollGroup = PushWidgetGroup(id, scrollGroupBounds.pos, scrollGroupLocalBounds, scrollGroupLocalBounds, groupFlags, manager);

    persistentData->visibleBounds.size = scrollGroupBounds.size;


    // @TODO: style style style
    const real32 nubWidth = 16.0f;

    EtcRect contentBounds = scrollGroupBounds;
    if (HasFlag(flags, UIScrollBarFlag_Vertical)) {
        contentBounds.size.x -= nubWidth;
        persistentData->visibleBounds.size.x -= nubWidth;
    }
    if (HasFlag(flags, UIScrollBarFlag_Horizontal)) {
        contentBounds.size.y -= nubWidth;
        persistentData->visibleBounds.size.y -= nubWidth;
    }

    WidgetID contentID = PushWidgetID(manager, "Content");
    uint32 contentFlags = WidgetGroupFlag_IsScrollContentGroup | WidgetGroupFlag_ResizeToContentVertical;
    WidgetGroup *content = PushWidgetGroup(id,
                                           -persistentData->visibleBounds.pos,
                                           {V2(0, 0), V2(contentBounds.size.x, 0)},
                                           persistentData->visibleBounds,
                                           contentFlags, manager);
}

void PushScrollGroup(const char *name, uint32 flags) {
    UIManager *manager = GetCurrentUIManager();
    WidgetGroup *currentGroup = PeekWidgetGroup(manager);

    // Use the rest of the space in the parent group
    EtcRect scrollGroupBounds = GetNextWidgetBounds(manager, currentGroup);
    scrollGroupBounds.pos.x = 0;
    scrollGroupBounds.size.x = currentGroup->localBounds.size.x;
    scrollGroupBounds.size.y = (currentGroup->localBounds.size.y - scrollGroupBounds.pos.y);

    PushScrollGroup(name, scrollGroupBounds, flags);
}

void PopScrollGroup() {
    UIManager *manager = GetCurrentUIManager();
    WidgetGroup *content = PeekWidgetGroup(manager);

    WidgetGroup *scrollGroup = GetParentGroup(manager, content);
    ASSERT(HasFlag(scrollGroup->flags, WidgetGroupFlag_IsScrollGroup));

    ScrollGroupPersistentData *persistentData = NULL;
    b32 foundData = GetPersistentData(manager, scrollGroup->id, ScrollGroupDataID, (void **)&persistentData);
    ASSERT(foundData);


    // Pop content group
    EtcRect contentLocalBounds = content->localBounds;
    EtcRect contentBounds = content->totalWidgetBounds;
    PopWidgetGroup(manager);


    // Render scrollbars
    WidgetID verticalID = PushWidgetID(manager, "Vertical");
    WidgetID horizontalID = PushWidgetID(manager, "Horizontal");

    persistentData->visibleBounds = ClampScrollGroupContentBounds(persistentData->visibleBounds, contentBounds);
    EtcRect prevVisibleBounds = persistentData->visibleBounds;

    b32 viewChanged = false;

    if (HasFlag(scrollGroup->flags, UIScrollBarFlag_Vertical)) {
        float nubWidth = scrollGroup->localBounds.size.x - contentLocalBounds.size.x;
            
        EtcRect scrollbarBounds = {};
        scrollbarBounds.pos = V2(scrollGroup->localBounds.size.x - nubWidth, 0);
        scrollbarBounds.size = V2(nubWidth, scrollGroup->localBounds.size.y);
        EtcRect scrollbarLocalBounds = {V2(0, 0), scrollbarBounds.size};

        WidgetGroup *scrollbar = PushWidgetGroup(verticalID, scrollbarBounds.pos, scrollbarLocalBounds, scrollbarLocalBounds, 0, manager);

        uint32 flags = UIScrollBarFlag_Vertical;
        viewChanged |= ScrollBar("Vertical Scrollbar", scrollbar->localBounds, contentBounds, &persistentData->visibleBounds, flags);
        
        PopWidgetGroup(manager);
    }

    if (HasFlag(scrollGroup->flags, UIScrollBarFlag_Horizontal)) {
        float nubWidth = scrollGroup->localBounds.size.y - contentLocalBounds.size.y;

        EtcRect scrollbarBounds = {};
        scrollbarBounds.pos = V2(0, scrollGroup->localBounds.size.y - nubWidth);

        if (HasFlag(scrollGroup->flags, UIScrollBarFlag_Vertical)) {
            scrollbarBounds.size = V2(scrollGroup->localBounds.size.x - nubWidth, nubWidth);
        }
        else {
            scrollbarBounds.size = V2(scrollGroup->localBounds.size.x, nubWidth);
        }

        EtcRect scrollbarLocalBounds = {V2(0, 0), scrollbarBounds.size};

        WidgetGroup *scrollbar = PushWidgetGroup(verticalID, scrollbarBounds.pos, scrollbarLocalBounds, scrollbarLocalBounds, 0, manager);

        uint32 flags = UIScrollBarFlag_Horizontal;
        viewChanged |= ScrollBar("Horizontal Scrollbar", scrollbar->localBounds, contentBounds, &persistentData->visibleBounds, flags);
        
        PopWidgetGroup(manager);
    }

    // Mousewheel scrolling
    b32 isScrolling = false;
    if (!persistentData->childIsScrolling && scrollGroup->isHoveringGroup && !viewChanged) {
        float scrollDirection = InputAnalogue(Game.mouse, Etc_ScrollDirection, true);
        if (scrollDirection != 0.0f) {
            persistentData->visibleBounds.pos.y -= 48 * scrollDirection;

            real32 minY = contentBounds.pos.y;
            real32 maxY = contentBounds.pos.y + Max(0, contentBounds.size.y - scrollGroup->localBounds.size.y);
            
            //persistentData->visibleBounds.pos.y -= 24 * scrollDirection;
            persistentData->visibleBounds.pos.y = Clamp(persistentData->visibleBounds.pos.y, minY, maxY);
        
            viewChanged = prevVisibleBounds.pos != persistentData->visibleBounds.pos;
            isScrolling = viewChanged;
        }
    }

    if (isScrolling || persistentData->childIsScrolling) {
        // Find parent scroll group and don't let it mousewheel scroll
        if (WidgetGroup *parentContentGroup = GetAncestorGroupWithFlags(manager, scrollGroup, WidgetGroupFlag_IsScrollContentGroup)) {
            WidgetGroup *parentScrollGroup = GetParentGroup(manager, parentContentGroup);
            ASSERT(parentScrollGroup != NULL);
            ASSERT(HasFlag(parentScrollGroup->flags, WidgetGroupFlag_IsScrollGroup));

            ScrollGroupPersistentData *parentPersistentData = NULL;
            b32 foundData = GetPersistentData(manager, parentScrollGroup->id, ScrollGroupDataID, (void **)&parentPersistentData);
            ASSERT(foundData);

            parentPersistentData->childIsScrolling = viewChanged;
        }
    }

    PopWidgetGroup(manager);
}


void ListWidget(const char *name, real32 entryHeight, uint32 *listCount, void *userData, ListEntryWidgetFunc *widgetFunc) {
    UIManager *manager = GetCurrentUIManager();

    PushScrollGroup(name);
    WidgetGroup *contentGroup = PeekWidgetGroup(manager);
    WidgetGroup *scrollGroup = GetParentGroup(manager, contentGroup);
    ASSERT(HasFlag(scrollGroup->flags, WidgetGroupFlag_IsScrollGroup));

    ScrollGroupPersistentData *persistentData = NULL;
    b32 foundData = GetPersistentData(manager, scrollGroup->id, ScrollGroupDataID, (void **)&persistentData);
    ASSERT(foundData);

    EtcRect entryBounds = GetNextWidgetBounds(manager, contentGroup);
    UIPadding padding = GetStylePadding(UIStyle_Padding, manager);


    EtcRect prevVisibleBounds = persistentData->visibleBounds;

    uint32 firstIndex = 0;
    uint32 lastIndex = *listCount - 1;

    real32 entryHeightWithPadding = (entryHeight + padding.top + padding.bottom);
    if (entryHeightWithPadding > 0) {
        firstIndex = (uint32)Floorf(prevVisibleBounds.pos.y / entryHeightWithPadding);
        lastIndex = (uint32)Ceilf((prevVisibleBounds.pos.y + prevVisibleBounds.size.y) / entryHeightWithPadding);
    }

    u32 successfulWidgetFuncs = 0;
    //for (uint32 i = firstIndex; i <= lastIndex; i++) {
    for (uint32 i = firstIndex; successfulWidgetFuncs < *listCount; i++) {
        if (i >= *listCount) {
            break;
        }

        EtcRect prevBounds = entryBounds;
        
        entryBounds.pos.y = padding.top + successfulWidgetFuncs * entryHeightWithPadding;
        entryBounds.size.y = entryHeight;
        SetNextWidgetBounds(manager, contentGroup, entryBounds);

        if (widgetFunc(i, userData)) {
            successfulWidgetFuncs++;
        }
    }

    contentGroup->totalWidgetBounds.pos.y = 0;
    contentGroup->totalWidgetBounds.size.y = successfulWidgetFuncs * entryHeightWithPadding;

    PopScrollGroup();
}

void DefaultDisplayNameWidget(const char *name, void *userData) {
    PushText(name);
}

void PushText(const char *text) {
    PushText(text, StringLength(text));
}

void PushText(const char *text, uint32 textLength) {
    UIManager *manager = GetCurrentUIManager();
    WidgetGroup *group = PeekWidgetGroup(manager);
    WidgetID id = PushWidgetID(manager, text);

    FontTable *font = GetFontTable(Game.assets, GetStyleFont(UIStyle_Font, manager));
    if (font == NULL) {
        return;
    }

    real32 emSize = GetStylef(UIStyle_FontSize, manager);

    EtcRect textBounds;
    vec2 textPosition;
    EtcRect widgetBounds = GetNextTextWidgetBounds(manager, group, font, emSize, &textBounds, &textPosition);

    vec4 color = GetStyle4(UIStyle_FontColor, manager);

    GlyphAttributes *attribs = PushArray(Game.frameMem, GlyphAttributes, textLength);
    ASSERT(attribs != NULL);

    EtcRect renderedTextBounds;
    UILayoutGlyphAttributes(text, textLength, font, textPosition, color, emSize, attribs, &renderedTextBounds);

    UIPadding insetMargin = GetStylePadding(UIStyle_InsetMargin, manager);
    widgetBounds.size.x = renderedTextBounds.size.x + (insetMargin.left + insetMargin.right);

    CommitWidgetBounds(manager, group, widgetBounds);

    RenderText(manager, group, font, attribs, textLength);

#if 0
    UIDebugRenderPrevWidgetBounds(manager, group, WithAlpha(BLACK, 0.75f));
#endif
}

void PushTextf(const char *fmt, ...) {
    va_list args;
    va_start(args, fmt);
    char *text = (char *)PushSize(Game.frameMem, 128);
    vsnprintf(text, 128, fmt, args);
    PushText(text);

    va_end(args);
}

b32 Button(const char *name, uint32 flags, const char *displayText) {
    UIManager *manager = GetCurrentUIManager();
    WidgetGroup *group = PeekWidgetGroup(manager);
    WidgetID id = PushWidgetID(manager, name);

    if (displayText == NULL) {
        displayText = name;
    }

    FontTable *font = GetFontTable(Game.assets, GetStyleFont(UIStyle_Font, manager));
    if (font == NULL) {
        return false;
    }

    real32 emSize = GetStylef(UIStyle_FontSize, manager);

    EtcRect textBounds;
    vec2 textPosition;
    EtcRect widgetBounds = GetNextTextWidgetBounds(manager, group, font, emSize, &textBounds, &textPosition);
    CommitWidgetBounds(manager, group, widgetBounds);

    b32 isActive = false;
    b32 clicked = false;
    b32 released = false;
    b32 isMouseOver = group->isHoveringGroup && PointInRect(widgetBounds, group->localMousePosition);

    if (IsActive(manager, id)) {
        if (InputActionReleased(Game.inputActions, InputAction_EditorPrimaryClick, true)) {
            SetActiveWidget(manager, WidgetID_Invalid);
            released = true;
        }
        else {
            isActive = true;
        }
    }
    else if (isMouseOver) {
        if (IsHovered(manager, id)) {
            if (InputActionPressed(Game.inputActions, InputAction_EditorPrimaryClick, true)) {
                SetActiveWidget(manager, id);

                isActive = true;
                clicked = true;
            }
        }

        SetHoveredWidget(manager, group, id);
    }


    // Render
    vec4 textColor = GetStyle4(UIStyle_FontColor, manager);

    uint32 textLength = StringLength(displayText);
    GlyphAttributes *attribs = PushArray(Game.frameMem, GlyphAttributes, textLength);
    ASSERT(attribs != NULL);

    UILayoutGlyphAttributes(displayText, textLength, font, textPosition, textColor, emSize, attribs);

    vec4 buttonColor;

    if (isActive) {
        buttonColor = GetStyle4(UIStyle_ButtonActiveColor);
    }
    else {
        buttonColor = GetStyle4(UIStyle_ButtonColor);
    }

    RenderMesh(manager, group, &Game.graphics->glyphQuad, TS(widgetBounds.pos, widgetBounds.size), buttonColor);
    RenderText(manager, group, font, attribs, textLength);


    if (HasFlag(flags, UIButtonFlag_TrueOnRelease)) {
        return released;
    }

    return clicked;
}

struct ButtonConfirmPersistentData {
    b32 needConfirmation;
};

b32 ButtonConfirm(const char *text) {
    b32 result = false;
    
    UIManager *manager = GetCurrentUIManager();
    WidgetGroup *group = PeekWidgetGroup(manager);
    WidgetID id = PushWidgetID(manager, text);

    static uint32 ButtonConfirmDataID = GenerateDataID("ButtonConfirmBuffer");

    ButtonConfirmPersistentData *persistentData = NULL;
    if (!GetPersistentData(manager, id, ButtonConfirmDataID, (void **)&persistentData)) {
        persistentData = (ButtonConfirmPersistentData *)AllocatePersistentData(manager, id, ButtonConfirmDataID, sizeof(ButtonConfirmPersistentData));
    }

    b32 confirmed = false;

    if (!persistentData->needConfirmation) {
        persistentData->needConfirmation = Button(text);
    }
    else {
        if (Button("Cancel")) {
            persistentData->needConfirmation = false;
        }

        PushStyle(MakeStyle4(UIStyle_ButtonColor, PASTEL_GREEN), manager);
        if (Button("Confirm")) {
            confirmed = true;
        }
        PopStyles(manager);
    }

    return confirmed;
}

b32 Toggle(const char *name, uint32 nameLength, bool *value, uint32 flags) {
    UIManager *manager = GetCurrentUIManager();
    WidgetGroup *group = PeekWidgetGroup(manager);
    WidgetID id = PushWidgetID(manager, name);

    FontTable *font = GetFontTable(Game.assets, GetStyleFont(UIStyle_Font, manager));
    if (font == NULL) {
        return false;
    }

    real32 emSize = GetStylef(UIStyle_FontSize, manager);
    vec4 nameColor = GetStyle4(UIStyle_FontColor, manager);

    EtcRect textBounds;
    vec2 textPosition;
    EtcRect widgetBounds = GetNextTextWidgetBounds(manager, group, font, emSize, &textBounds, &textPosition);

    GlyphAttributes *attribs = PushArray(Game.frameMem, GlyphAttributes, nameLength);
    EtcRect renderedNameBounds;
    UILayoutGlyphAttributes(name, nameLength, font, textPosition, nameColor, emSize, attribs, &renderedNameBounds);

    UIPadding insetMargin = GetStylePadding(UIStyle_InsetMargin, manager);
    if (!HasFlag(flags, UIToggleFlag_UseAllHorizontalSpace)) {
        widgetBounds.size.x = renderedNameBounds.size.x + (insetMargin.left + insetMargin.right);
    }
    else {
        real32 textPositionX = Roundf(widgetBounds.pos.x + (widgetBounds.size.x - renderedNameBounds.size.x) / 2.0f);
        real32 textPositionXDiff = textPositionX - textPosition.x;
        for (int i = 0; i < nameLength; i++) {
            attribs[i].position.x = Roundf(attribs[i].position.x + textPositionXDiff);
        }
    }
    
    CommitWidgetBounds(manager, group, widgetBounds);

    b32 isActive = false;
    b32 clicked = false;
    b32 isMouseOver = group->isHoveringGroup && PointInRect(widgetBounds, group->localMousePosition);

    if (IsActive(manager, id)) {
        isActive = true;

        if (InputActionReleased(Game.inputActions, InputAction_EditorPrimaryClick, true)) {
            clicked = isMouseOver;

            SetActiveWidget(manager, WidgetID_Invalid);
            isActive = false;
        }
    }
    else if (isMouseOver) {
        if (IsHovered(manager, id)) {
            if (InputActionPressed(Game.inputActions, InputAction_EditorPrimaryClick, true)) {
                isActive = true;
                SetActiveWidget(manager, id);

                *value = !(*value);
                clicked = true;
            }
        }

        SetHoveredWidget(manager, group, id);
    }

    // Render
    vec4 buttonColor;

    if (*value) {
        buttonColor = PASTEL_GREEN;
    }
    else {
        buttonColor = PASTEL_RED;
    }

    RenderMesh(manager, group, &Game.graphics->glyphQuad, TS(widgetBounds.pos, widgetBounds.size), buttonColor);
    RenderText(manager, group, font, attribs, nameLength);

    return clicked;
}

b32 Toggle(const char *name, b32 *value, uint32 flags) {
    // @WARNING: if value is not 0 or 1 this will produce weird results because
    // toggling (value = !value) for non binary values will not toggle between true and false
    return Toggle(name, StringLength(name), (bool *)value, flags);
}

b32 Toggle(const char *name, bool *value, uint32 flags) {
    return Toggle(name, StringLength(name), value, flags);
}

b32 RadioToggle(const char *name, uint32 valueCount, uint32 *value, const char **valueNames, uint32 strOffset, uint32 flags) {
    UIManager *manager = GetCurrentUIManager();
    WidgetGroup *group = PeekWidgetGroup(manager);

    b32 displayName = GetStyleb(UIStyle_DisplayName, manager);
    if (displayName) {
        UIStyle tempStyles[] = {
            MakeStyleb(UIStyle_DisplayName, false),
            MakeStyleFlowDirection(UIStyle_FlowDirection, UIFlowDirection_Horizontal),
        };

        DisplayNameWidgetFunc *displayNameWidgetFunc = GetDisplayNameWidgetFunc(manager);
        displayNameWidgetFunc(name, NULL);

        PushStyles(tempStyles, ARRAY_LENGTH(UIStyle, tempStyles));
    }


    FontTable *font = GetFontTable(Game.assets, GetStyleFont(UIStyle_Font, manager));
    if (font == NULL) {
        return false;
    }

    real32 emSize = GetStylef(UIStyle_FontSize, manager);
    vec4 nameColor = GetStyle4(UIStyle_FontColor, manager);

    EtcRect widgetBounds = GetNextWidgetBounds(manager, group);
    UIPadding internalPadding = {1, 1, 1, 1};

    real32 *textWidths = PushArray(Game.frameMem, real32, valueCount);
    for (int i = 0; i < valueCount; i++) {
        uint32 nameLength = StringLength(valueNames[i]);
        GlyphAttributes *attribs = PushArray(Game.frameMem, GlyphAttributes, nameLength);

        EtcRect renderedNameBounds;
        UILayoutGlyphAttributes(valueNames[i], nameLength, font, V2(0), nameColor, emSize, attribs, &renderedNameBounds);
        textWidths[i] = renderedNameBounds.size.x;

        // @TODO: pop framemem!
    }


    b32 result = false;
    real32 widgetHeight = 0;
    uint32 row = 0;

    uint32 valueIndex = 0;

    while (valueIndex < valueCount) {
        uint32 elementsInRow = 0;
        real32 rowTextWidth = 0;

        for (int i = valueIndex; i < valueCount; i++) {
            if (elementsInRow == 0) {
                elementsInRow++;
                rowTextWidth = textWidths[i];
            }
            else {
                real32 targetWidth = rowTextWidth + internalPadding.left + internalPadding.right + textWidths[i];
                if (targetWidth <= widgetBounds.size.x) {
                    elementsInRow++;
                    rowTextWidth = targetWidth;
                }
            }
        }

        real32 spaceLeft = widgetBounds.size.x - rowTextWidth;
        real32 extraSpacePerElement = spaceLeft / elementsInRow;

        real32 usedSpaceX = 0;

        for (int i = 0; i < elementsInRow; i++) {
            EtcRect elementBounds = {};
            elementBounds.pos.x = widgetBounds.pos.x + usedSpaceX;
            elementBounds.pos.y = widgetBounds.pos.y + row * (widgetHeight + internalPadding.bottom + internalPadding.top);
            elementBounds.size.x = textWidths[valueIndex + i] + extraSpacePerElement;

            bool selected = (valueIndex + i) == *value;
            SetNextWidgetBounds(manager, group, elementBounds);
            result |= Toggle(valueNames[valueIndex + i] + strOffset, &selected, UIToggleFlag_UseAllHorizontalSpace);

            if (selected) {
                *value = valueIndex + i;
            }

            usedSpaceX += group->prevWidgetBounds.size.x + internalPadding.right + internalPadding.left;
            widgetHeight = group->prevWidgetBounds.size.y;
        }

        row++;
        valueIndex += elementsInRow;
    }

    if (displayName) {
        PopStyles();
    }

    return result;
}

b32 RadioToggleEnum(const char *name, const char *enumName, uint32 *value, uint32 flags) {
    if (Schema *schema = Game.GetSchema(enumName)) {
        if (schema->type == SchemaType_Enum) {
            bool useShortNames = GetStyleb(UIStyle_UseShortEnumNames);
            
            if (useShortNames) {
                return RadioToggle(name, schema->_enum.count, value, schema->_enum.stringValues, StringLength(enumName) + 1, flags);
            }
            else {
                return RadioToggle(name, schema->_enum.count, value, schema->_enum.stringValues, 0, flags);    
            }
            
        }
    }

    return false;
}


b32 Selectable(const char *name, u32 nameLength, b32 *selected, uint32 flags) {
    UIManager *manager = GetCurrentUIManager();
    WidgetGroup *group = PeekWidgetGroup(manager);
    WidgetID id = PushWidgetID(manager, name);

    FontTable *font = GetFontTable(Game.assets, GetStyleFont(UIStyle_Font, manager));
    if (font == NULL) {
        return false;
    }

    real32 emSize = GetStylef(UIStyle_FontSize, manager);
    vec4 nameColor = GetStyle4(UIStyle_FontColor, manager);

    EtcRect textBounds;
    vec2 textPosition;
    EtcRect widgetBounds = GetNextTextWidgetBounds(manager, group, font, emSize, &textBounds, &textPosition);

    CommitWidgetBounds(manager, group, widgetBounds);

    b32 isActive = false;
    b32 clicked = false;
    b32 isMouseOver = group->isHoveringGroup && PointInRect(widgetBounds, group->localMousePosition);

    if (IsActive(manager, id)) {
        isActive = true;

        if (InputActionReleased(Game.inputActions, InputAction_EditorPrimaryClick, true)) {
            SetActiveWidget(manager, WidgetID_Invalid);
        }
        else {
            isActive = true;
        }
    }
    else if (isMouseOver) {
        if (IsHovered(manager, id)) {
            if (InputActionPressed(Game.inputActions, InputAction_EditorPrimaryClick, true)) {
                isActive = true;

                if (selected != NULL) {
                    *selected = true;
                }

                if (!HasFlag(flags, UISelectableFlag_TrueOnDoubleClick)) {
                    clicked = true;
                }
                
                SetActiveWidget(manager, id);
            }
        }

        SetHoveredWidget(manager, group, id);
    }

    // Render
    GlyphAttributes *attribs = PushArray(Game.frameMem, GlyphAttributes, nameLength);
    ASSERT(attribs != NULL);

    UILayoutGlyphAttributes(name, nameLength, font, textPosition, nameColor, emSize, attribs);
    
    vec4 buttonColor;

    if (isActive) {
        buttonColor = PASTEL_YELLOW;
    }
    else {
        buttonColor = PASTEL_BLUE;
    }

    RenderMesh(manager, group, &Game.graphics->glyphQuad, TS(widgetBounds.pos, widgetBounds.size), buttonColor);
    RenderText(manager, group, font, attribs, nameLength);

    return clicked;
}

b32 Selectable(const char *name, b32 *selected, uint32 flags) {
    return Selectable(name, StringLength(name), selected, flags);
}

int32 FindClickedCursorIndex(vec2 localMousePosition, GlyphAttributes *attribs, uint32 stringLength) {
    int32 result = stringLength;

    for (int i = 0; i < stringLength; i++) {

        real32 targetX = attribs[i].position.x + 0.75f * attribs[i].dimensions.x;
        if (targetX > localMousePosition.x) {
            result = i;
            break;
        }
    }

    return result;
}

vec2 GetCursorPositionAtIndex(int32 index, FontTable *font, real32 emSize, GlyphAttributes *attribs, uint32 stringLength) {
    ASSERT(index >= 0);
    ASSERT(index <= stringLength);

    // @NOTE: this is all assuming glyph anchors are (0, 0)
    
    vec2 position;
    if (index == 0) {
        GlyphAttributes attribute = attribs[0];
        GlyphInfo info = GetGlyphInfo(font, attribute.codepoint);

        position.x = attribute.position.x;
        position.y = attribute.position.y + info.lowerLeft.y * emSize;
    }
    else {
        GlyphAttributes attribute = attribs[index - 1];
        GlyphInfo info = GetGlyphInfo(font, attribute.codepoint);

        position.x = attribute.position.x;
        position.y = attribute.position.y + info.lowerLeft.y * emSize;
        position.x += info.advance * emSize;
    }

    position.y -= font->ascent * emSize;

    return position;
}

void InputStringDeleteSelection_(InputStringActiveData *activeData, char *buffer, int32 *stringLength) {
    int32 selectionStart = Min(activeData->selectionStart, activeData->selectionEnd);
    int32 selectionEnd = Max(activeData->selectionStart, activeData->selectionEnd);
    int32 selectionLength = selectionEnd - selectionStart;

    if (selectionLength == 0) {
        return;
    }

    uint32 charactersToMove = *stringLength - selectionEnd;
    if (charactersToMove > 0) {
        Plat.MoveMem(buffer + selectionStart, buffer + selectionEnd, charactersToMove);
    }

    activeData->cursorIndex = selectionStart;
    *stringLength -= selectionLength;
    buffer[*stringLength] = '\0';

    activeData->hasSelection = false;
}

b32 InputString(const char *name, char *buffer, uint32 bufferLength, uint32 flags, ValidateInputTextFunc *validationFunc, void *validateData) {
    b32 modified = false;
    b32 pressedEnter = false;

    UIManager *manager = GetCurrentUIManager();
    WidgetGroup *group = PeekWidgetGroup(manager);
    WidgetID id = HasFlag(flags, InputStringFlag_ReuseWidgetID_) ? group->prevID : PushWidgetID(manager, name);

    static real32 pressedActionTime = 0.1f;
    static real32 holdActionTime = 0.05f;

    static vec4 validTextColor = GetStyle4(UIStyle_FontColor, manager);
    static vec4 invalidTextColor = PASTEL_RED;

    static uint32 InputStringDataID = GenerateDataID("InputStringPersistentData");
    static uint32 InitialBufferDataID = GenerateDataID("InputStringInitialBuffer");


    InputStringPersistentData *persistentData;
    if (!GetPersistentData(manager, id, InputStringDataID, (void **)&persistentData)) {
        persistentData = (InputStringPersistentData *)AllocatePersistentData(manager, id, InputStringDataID, sizeof(InputStringPersistentData));

        if (validationFunc != NULL) {
            persistentData->bufferIsValid = validationFunc(buffer, NULL, validateData);
        }
        else {
            persistentData->bufferIsValid = true;
        }
    }
    

    b32 displayName = GetStyleb(UIStyle_DisplayName, manager);

    if (displayName) {
        DisplayNameWidgetFunc *displayNameWidgetFunc = GetDisplayNameWidgetFunc(manager);
        displayNameWidgetFunc(name, NULL);

        UIStyle tempStyles[] = {
            MakeStyleFlowDirection(UIStyle_FlowDirection, UIFlowDirection_Horizontal),
        };

        PushStyles(tempStyles, ARRAY_LENGTH(UIStyle, tempStyles));
    }
                  

    FontTable *font = GetFontTable(Game.assets, GetStyleFont(UIStyle_Font, manager));
    if (font == NULL) {
        return false;
    }

    real32 emSize = GetStylef(UIStyle_FontSize, manager);

    EtcRect textBounds;
    vec2 textPosition;
    EtcRect widgetBounds = GetNextTextWidgetBounds(manager, group, font, emSize, &textBounds, &textPosition);

    CommitWidgetBounds(manager, group, widgetBounds);


    GlyphAttributes *attribs = PushArray(Game.frameMem, GlyphAttributes, bufferLength);
    ASSERT(attribs != NULL);

    int32 stringLength = StringLength(buffer, bufferLength);

    vec4 textColor = persistentData->bufferIsValid ? validTextColor : invalidTextColor;
    UILayoutGlyphAttributes(buffer, stringLength, font, textPosition, textColor, emSize, attribs);


    b32 isActive = IsActive(manager, id);

    b32 isMouseOver = group->isHoveringGroup && PointInRect(widgetBounds, group->localMousePosition);
    b32 wasClicked = false;
    b32 shouldRelayoutText = false;

    InputStringActiveData *activeData = (InputStringActiveData *)manager->activeWidgetData;

    if (isActive) {
        ASSERT(activeData != NULL);

        b32 shouldDeactivate = false;


        // Mark mode
        b32 exitSelectionOnInput = false;
        if (!activeData->hasSelection) {
            if (InputPressed(Game.keyboard, Etc_Shift, true) || InputPressed(Game.keyboard, Etc_RightShift, true)) {
                activeData->hasSelection = true;
                activeData->selectionStart = activeData->cursorIndex;
                activeData->selectionEnd = activeData->cursorIndex;
            }
        }
        else {
            if (!InputHold(Game.keyboard, Etc_Shift, true) && !InputHold(Game.keyboard, Etc_RightShift, true)) {
                exitSelectionOnInput = true;
            }
        }


        // Cursor Movement
        if (InputActionHold(Game.inputActions, InputAction_EditorMoveCursorUp, true)) {
            activeData->state = InputStringState_MoveUp;
        }
        else if (activeData->state == InputStringState_MoveUp) {
            activeData->state = InputStringState_None;
        }

        if (InputActionHold(Game.inputActions, InputAction_EditorMoveCursorDown, true)) {
            activeData->state = InputStringState_MoveDown;
        }
        else if (activeData->state == InputStringState_MoveDown) {
            activeData->state = InputStringState_None;
        }
        
        if (InputActionHold(Game.inputActions, InputAction_EditorMoveCursorRight, true)) {
            b32 pressed = InputActionPressed(Game.inputActions, InputAction_EditorMoveCursorRight, true);
            if (pressed && activeData->state == InputStringState_None) {
                activeData->state = InputStringState_MoveRight;
            }

            if (activeData->state == InputStringState_MoveRight &&
                (pressed || Game.mem->time >= activeData->nextActionTime)) {
                activeData->cursorIndex = Min(stringLength, activeData->cursorIndex + 1);

                if (exitSelectionOnInput) {
                    activeData->hasSelection = false;
                }
                else if (activeData->hasSelection) {
                    activeData->selectionEnd = activeData->cursorIndex;
                }

                activeData->nextActionTime = Game.mem->time + (pressed ? pressedActionTime : holdActionTime);
                activeData->blinkStartTime = Game.mem->time;
            }
        }
        else if (activeData->state == InputStringState_MoveRight) {
            activeData->state = InputStringState_None;
        }

        if (InputActionHold(Game.inputActions, InputAction_EditorMoveCursorLeft, true)) {
            b32 pressed = InputActionPressed(Game.inputActions, InputAction_EditorMoveCursorLeft, true);
            if (pressed && activeData->state == InputStringState_None) {
                activeData->state = InputStringState_MoveLeft;
            }

            if (activeData->state == InputStringState_MoveLeft &&
                (pressed || Game.mem->time >= activeData->nextActionTime)) {
                activeData->cursorIndex = Max(0, activeData->cursorIndex - 1);

                if (exitSelectionOnInput) {
                    activeData->hasSelection = false;
                }
                else if (activeData->hasSelection) {
                    activeData->selectionEnd = activeData->cursorIndex;
                }

                activeData->nextActionTime = Game.mem->time + (pressed ? pressedActionTime : holdActionTime);
                activeData->blinkStartTime = Game.mem->time;
            }
        }
        else if (activeData->state == InputStringState_MoveLeft) {
            activeData->state = InputStringState_None;
        }

        if (InputActionHold(Game.inputActions, InputAction_EditorMoveBeginningOfLine, true)) {
            if (InputActionPressed(Game.inputActions, InputAction_EditorMoveBeginningOfLine, true) &&
                activeData->state == InputStringState_None) {
                activeData->state = InputStringState_MoveBeginningOfLine;
            }

            if (activeData->state == InputStringState_MoveBeginningOfLine) {
                activeData->cursorIndex = 0;

                if (exitSelectionOnInput) {
                    activeData->hasSelection = false;
                }
                else if (activeData->hasSelection) {
                    activeData->selectionEnd = activeData->cursorIndex;
                }

                activeData->blinkStartTime = Game.mem->time;
            }
        }
        else if (activeData->state == InputStringState_MoveBeginningOfLine) {
            activeData->state = InputStringState_None;
        }

        if (InputActionHold(Game.inputActions, InputAction_EditorMoveEndOfLine, true)) {
            if (InputActionPressed(Game.inputActions, InputAction_EditorMoveEndOfLine, true) &&
                activeData->state == InputStringState_None) {
                activeData->state = InputStringState_MoveEndOfLine;
            }

            if (activeData->state == InputStringState_MoveEndOfLine) {
                activeData->cursorIndex = stringLength;

                if (exitSelectionOnInput) {
                    activeData->hasSelection = false;
                }
                else if (activeData->hasSelection) {
                    activeData->selectionEnd = activeData->cursorIndex;
                }

                activeData->blinkStartTime = Game.mem->time;
            }
        }
        else if (activeData->state == InputStringState_MoveEndOfLine) {
            activeData->state = InputStringState_None;
        }

        
        if (InputPressed(&Game.mem->keyboard, Etc_Tab, true)) {
            activeData->state = InputStringState_Completion;
        }


        // Clicking
        if (InputActionPressed(Game.inputActions, InputAction_EditorPrimaryClick, true)) {
            if (!isMouseOver) {
                shouldDeactivate = true;
            }
            else if (activeData->state == InputStringState_None) {

                int32 clickedIndex = FindClickedCursorIndex(group->localMousePosition, attribs, stringLength);
                if ((Game.time - activeData->prevMouseReleaseTime) <= GetStylef(UIStyle_DoubleClickThreshold)) {

                    activeData->hasSelection = true;
                    activeData->selectionStart = 0;
                    activeData->selectionEnd = stringLength;

                    activeData->cursorIndex = stringLength;
                }
                else {
                    activeData->hasSelection = false;
                    activeData->cursorIndex = clickedIndex;
                }
                
                activeData->blinkStartTime = Game.time;

                activeData->initialClickedIndex = clickedIndex;
                activeData->state = InputStringState_MouseClick;
            }
        }
        else if (InputActionHold(Game.inputActions, InputAction_EditorPrimaryClick, true)) {
            int32 hoveredIndex = FindClickedCursorIndex(group->localMousePosition, attribs, stringLength);

            switch (activeData->state) {
                case InputStringState_MouseClick: {
                    if (hoveredIndex != activeData->initialClickedIndex) {
                        activeData->state = InputStringState_MouseDrag;

                        activeData->hasSelection = true;
                        activeData->selectionStart = activeData->initialClickedIndex;
                        activeData->selectionEnd = hoveredIndex;
                        activeData->cursorIndex = activeData->selectionEnd;
                    }
                } break;

                case InputStringState_MouseDrag: {
                    activeData->selectionEnd = hoveredIndex;
                    activeData->cursorIndex = activeData->selectionEnd;
                } break;

                default: {
                } break;
            }
        }
        else if (InputActionReleased(Game.inputActions, InputAction_EditorPrimaryClick, true)) {

            switch (activeData->state) {
                case InputStringState_MouseClick: {
                    activeData->prevMouseReleaseTime = Game.time;
                } break;

                case InputStringState_MouseDrag: {
                    if (activeData->selectionStart == activeData->selectionEnd) {
                        activeData->hasSelection = false;
                        activeData->blinkStartTime = Game.time;
                    }

                    // Make sure you can't double click to select all immediately after dragging
                    activeData->prevMouseReleaseTime = 0.0f;
                };

                default: {
                } break;
            }

            activeData->state = InputStringState_None;
        }

        // Delete Characters
        if (InputHold(Game.keyboard, Etc_Backspace, true)) {
            b32 pressed = InputPressed(Game.keyboard, Etc_Backspace, true);
            if (pressed && activeData->state == InputStringState_None) {
                activeData->state = InputStringState_Backspace;
            }

            if (activeData->state == InputStringState_Backspace &&
                (pressed || Game.mem->time >= activeData->nextActionTime)) {

                // If there is a selection delete the selected region
                if (activeData->hasSelection && activeData->selectionStart != activeData->selectionEnd) {
                    InputStringDeleteSelection_(activeData, buffer, &stringLength);
                }
                // Otherwise delete backwards one character
                else if (activeData->cursorIndex - 1 >= 0) {
                    uint32 charactersToMove = stringLength - activeData->cursorIndex;
                    if (charactersToMove > 0) {
                        Plat.MoveMem(buffer + activeData->cursorIndex - 1, buffer + activeData->cursorIndex, charactersToMove);
                    }

                    activeData->cursorIndex -= 1;
                    stringLength -= 1;
                    buffer[stringLength] = '\0';

                }

                activeData->hasSelection = false;
                modified = true;
                activeData->nextActionTime = Game.mem->time + (pressed ? pressedActionTime : holdActionTime);
            }
        }
        else if (activeData->state == InputStringState_Backspace) {
            activeData->state = InputStringState_None;
        }

        if (InputHold(Game.keyboard, Etc_ForwardDelete, true)) {
            b32 pressed = InputPressed(Game.keyboard, Etc_ForwardDelete, true);
            if (pressed && activeData->state == InputStringState_None) {
                activeData->state = InputStringState_ForwardDelete;
            }

            if (activeData->state == InputStringState_ForwardDelete &&
                (pressed || Game.mem->time >= activeData->nextActionTime)) {

                // If there is a selection delete the selected region
                if (activeData->hasSelection && activeData->selectionStart != activeData->selectionEnd) {
                    InputStringDeleteSelection_(activeData, buffer, &stringLength);
                }
                // Otherwise delete forward one character
                else if (activeData->cursorIndex < stringLength) {

                    uint32 charactersToMove = stringLength - activeData->cursorIndex;
                    if (charactersToMove > 0) {
                        Plat.MoveMem(buffer + activeData->cursorIndex, buffer + activeData->cursorIndex + 1, charactersToMove);
                    }

                    stringLength -= 1;
                    buffer[stringLength] = '\0';
                }

                activeData->hasSelection = false;
                modified = true;
                activeData->nextActionTime = Game.mem->time + (pressed ? pressedActionTime : holdActionTime);
            }
        }
        else if (activeData->state == InputStringState_ForwardDelete) {
            activeData->state = InputStringState_None;
        }

        // Insert typed characters
        uint32 bytesInserted = 0;
        
        if (activeData->state == InputStringState_None) {
            DynamicArray<uint32> insertedCharacters = manager->inputCharactersBackBuffer;

            if (insertedCharacters.count > 0 && activeData->hasSelection) {
                InputStringDeleteSelection_(activeData, buffer, &stringLength);
            }

            uint32 bytesLeftInBuffer = bufferLength - 1 - stringLength; 
            
            uint32 bytesAfterCursor = 0;
            if (stringLength >= activeData->cursorIndex) {
                bytesAfterCursor = stringLength - activeData->cursorIndex;
            }

            for (int i = 0; i < insertedCharacters.count; i++) {
                uint32 codepoint = insertedCharacters[i];

                if (!IsPrintable(codepoint)) {
                    continue;
                }

                uint32 bytesNeededForCodepoint = 1;

                if (bytesInserted + bytesNeededForCodepoint > bytesLeftInBuffer) {
                    break;
                }

                if (bytesAfterCursor > 0) {
                    Plat.MoveMem(buffer + activeData->cursorIndex + bytesNeededForCodepoint,
                                 buffer + activeData->cursorIndex,
                                 bytesAfterCursor);
                }
                buffer[activeData->cursorIndex] = (uint8)codepoint;
                activeData->cursorIndex++;

                bytesInserted += bytesNeededForCodepoint;
            }

            if (bytesInserted > 0) {
                activeData->state = InputStringState_InsertedChar;
                
                stringLength += bytesInserted;
                buffer[stringLength] = 0;
                modified = true;
            }
        }

        if (activeData->state == InputStringState_InsertedChar && bytesInserted == 0) {
            activeData->state = InputStringState_None;
        }

        
        if (validationFunc != NULL) {
            persistentData->bufferIsValid = validationFunc(buffer, activeData, validateData);
        }


        if (modified) {
            activeData->blinkStartTime = Game.mem->time;
            shouldRelayoutText = true;

#if 0
            // @TODO: copy over prev value
            ASSERT(false);
#endif
        }


        // Exiting
        if (activeData->state == InputStringState_None) {
            if (InputPressed(Game.keyboard, Etc_Return, true)) {
                pressedEnter = true;
                shouldDeactivate = true;
            }
            else if (InputActionPressed(Game.inputActions, InputAction_EditorDeselect, true)) {
                shouldDeactivate = true;
            }
        }

        if (shouldDeactivate) {
            SetActiveWidget(manager, WidgetID_Invalid);
            isActive = false;
            manager->blockingKeyboard = false;
        }
        else {
            isActive = true;
            manager->blockingKeyboard = true;
        }
    }
    else if (isMouseOver) {
        if (IsHovered(manager, id)) {
            if (InputActionPressed(Game.inputActions, InputAction_EditorPrimaryClick, true)) {
                wasClicked = true;
            }
        }

        SetHoveredWidget(manager, group, id);
    }


    b32 shouldActivate = RegisterSelectableWidget(manager, id);

    if (wasClicked || shouldActivate) {
        SetActiveWidget(manager, id, sizeof(InputStringActiveData));
        manager->blockingKeyboard = true;
        isActive = true;

        activeData = (InputStringActiveData *)manager->activeWidgetData;

        activeData->blinkStartTime = Game.mem->time;

        activeData->cursorIndex = FindClickedCursorIndex(group->localMousePosition, attribs, stringLength);

        persistentData->initialBuffer = (char *)AllocatePersistentData(manager, id, InitialBufferDataID, sizeof(char) * bufferLength);
        Plat.ClearMem(persistentData->initialBuffer, sizeof(char) * persistentData->initialBufferLength);
        Plat.CopyMem(persistentData->initialBuffer, buffer, sizeof(char) * bufferLength);
        persistentData->initialBufferLength = bufferLength;


        if (stringLength > 0 && ((shouldActivate && !GetStyleb(UIStyle_InputStringNoSelectAllWhenActivated)) || GetStyleb(UIStyle_InputStringSelectAllOnActive))) {
            activeData->hasSelection = true;
            activeData->selectionStart = 0;
            activeData->selectionEnd = stringLength;
        }
        else {
            activeData->initialClickedIndex = activeData->cursorIndex;
            if (InputHold(Game.mouse, Etc_MouseLeft, true)) {
                activeData->state = InputStringState_MouseClick;
            }
            else {
                activeData->state = InputStringState_None;
            }
        }
    }


    // If the InputString became inactive check if the initial buffer should be restored
    if (persistentData->wasActive && !isActive) {
        // Reset to initial buffer
        if (!persistentData->bufferIsValid) {
            uint32 charactersToCopy = Min(bufferLength, persistentData->initialBufferLength);
            if (charactersToCopy > 0) {
                Plat.ClearMem(buffer, sizeof(char) * bufferLength);
                Plat.CopyMem(buffer, persistentData->initialBuffer, sizeof(char) * charactersToCopy);
                buffer[charactersToCopy - 1] = 0;
            }

            stringLength = StringLength(buffer, bufferLength);

            if (validationFunc != NULL) {
                persistentData->bufferIsValid = validationFunc(buffer, activeData, validateData);
            }
            else {
                persistentData->bufferIsValid = true;
            }

            shouldRelayoutText = true;
        }

        DeallocatePersistentData(manager, id, InitialBufferDataID);
    }

    persistentData->wasActive = isActive;


    // Render
    if (shouldRelayoutText) {
        textColor = persistentData->bufferIsValid ? validTextColor : invalidTextColor;
        UILayoutGlyphAttributes(buffer, stringLength, font, textPosition, textColor, emSize, attribs);
    }

    vec4 backgroundColor = V4(0.0f, 0.0f, 0.0f, 0.75f);
    RenderMesh(manager, group, &Game.graphics->glyphQuad, TS(widgetBounds.pos, widgetBounds.size), backgroundColor);

    RenderText(manager, group, font, attribs, stringLength);


    if (isActive) {

        // Render text selection
        if (activeData->hasSelection && (activeData->selectionStart != activeData->selectionEnd)) {

            int32 selectionStart = Min(activeData->selectionStart, activeData->selectionEnd);
            int32 selectionEnd = Max(activeData->selectionStart, activeData->selectionEnd);
            
            // @TODO: selectionColor should be a UIStyle
            vec4 selectionColor = WithAlpha(PASTEL_BLUE, 1.0f);

            vec2 selectionStartPosition = GetCursorPositionAtIndex(selectionStart, font, emSize, attribs, stringLength);
            vec2 selectionEndPosition = GetCursorPositionAtIndex(selectionEnd, font, emSize, attribs, stringLength);
            real32 selectionWidth = selectionEndPosition.x - selectionStartPosition.x;

            EtcRect selectionBounds = {selectionStartPosition, {selectionWidth, textBounds.size.y}};
            RenderMesh(manager, group, &Game.graphics->glyphQuad, TS(selectionBounds.pos, selectionBounds.size), selectionColor);
        }

        
        real32 blinkRate = GetStylef(UIStyle_InputStringBlinkRate);
        b32 shouldRenderCursor = ((int32)Floorf((Game.mem->time - activeData->blinkStartTime) * blinkRate)) % 2 == 0;

        if (!activeData->hasSelection && shouldRenderCursor) {
            real32 cursorWidth = 2.0f;
            real32 cursorHeight = textBounds.size.y;

            vec2 cursorPosition;
            if (stringLength > 0) {
                cursorPosition = GetCursorPositionAtIndex(activeData->cursorIndex, font, emSize, attribs, stringLength);
            }
            else {
                cursorPosition = textBounds.pos;
            }
            cursorPosition.x -= 0.5f * cursorWidth;

            EtcRect cursorBounds = {{cursorPosition}, {cursorWidth, cursorHeight}};
            RenderMesh(manager, group, &Game.graphics->glyphQuad, TS(cursorBounds.pos, cursorBounds.size), textColor);
        }
    }

    if (displayName) {
        PopStyles();
    }

    if (HasFlag(flags, InputStringFlag_ReturnTrueOnEnter)) {
        return pressedEnter;
    }

    return modified;
}


struct InputIntPersistentData {
    b32 wasActive;
    int32 initialValue;
    int32 prevValue;
    char buffer[32];
};

b32 ValidateInputInt(const char *buffer, InputStringActiveData *activeData = NULL, void *data = NULL) {
    const char *at = buffer;

    if (*at == '\0') {
        return false;
    }

    bool foundSign = false;

    if (*at == '-') {
        at++;
    }

    while (*at != '\0') {

        if (IsNumeric(*at)) {
            at++;
            continue;
        }

        return false;
    }

    return true;
}

b32 InputInt(const char *name, int32 *value, uint32 flags) {
    UIManager *manager = GetCurrentUIManager();
    WidgetGroup *group = PeekWidgetGroup(manager);
    WidgetID id = PushWidgetID(manager, name);
    
    static uint32 InputIntDataID = GenerateDataID("InputIntBuffer");

    InputIntPersistentData *persistentData = NULL;
    if (!GetPersistentData(manager, id, InputIntDataID, (void **)&persistentData)) {
        persistentData = (InputIntPersistentData *)AllocatePersistentData(manager, id, InputIntDataID, sizeof(InputIntPersistentData));

        stbsp_snprintf(persistentData->buffer, sizeof(persistentData->buffer), "%d", *value);
    }

    if (!IsActive(manager, id) && *value != persistentData->prevValue) {
        persistentData->prevValue = *value;
        stbsp_snprintf(persistentData->buffer, sizeof(persistentData->buffer), "%d", *value);
    }

    b32 modified = false;
    flags |= InputStringFlag_ReuseWidgetID_;
    if (InputString(name, persistentData->buffer, ARRAY_LENGTH(char, persistentData->buffer), flags, &ValidateInputInt)) {
        int32 newValue;
        int32 result = sscanf(persistentData->buffer, "%d", &newValue);

        if (result == 1 && newValue != *value) {
            *value = newValue;
            persistentData->prevValue = newValue;

            modified = true;
        }
    }

    b32 isActive = IsActive(manager, id);

    if (!persistentData->wasActive && isActive) {
        persistentData->initialValue = persistentData->prevValue;
    }
    else if (persistentData->wasActive && !isActive) {
        if (!ValidateInputInt(persistentData->buffer)) {
            *value = persistentData->initialValue;
        }

        stbsp_snprintf(persistentData->buffer, sizeof(persistentData->buffer), "%d", *value);
    }

    persistentData->wasActive = isActive;

    return modified;
}


struct InputUIntPersistentData {
    b32 wasActive;
    uint32 initialValue;
    uint32 prevValue;
    char buffer[32];
};

b32 ValidateInputUInt(const char *buffer, InputStringActiveData *activeData, void *data = NULL) {
    const char *at = buffer;

    if (*at == '\0') {
        return false;
    }

    bool foundSign = false;

    while (*at != '\0') {

        if (IsNumeric(*at)) {
            at++;
            continue;
        }

        return false;
    }

    return true;
}

b32 InputUInt(const char *name, uint32 *value, uint32 flags) {
    UIManager *manager = GetCurrentUIManager();
    WidgetGroup *group = PeekWidgetGroup(manager);
    WidgetID id = PushWidgetID(manager, name);
    
    static uint32 InputUIntDataID = GenerateDataID("InputUIntBuffer");

    InputUIntPersistentData *persistentData = NULL;
    if (!GetPersistentData(manager, id, InputUIntDataID, (void **)&persistentData)) {
        persistentData = (InputUIntPersistentData *)AllocatePersistentData(manager, id, InputUIntDataID, sizeof(InputUIntPersistentData));

        stbsp_snprintf(persistentData->buffer, sizeof(persistentData->buffer), "%u", *value);
    }

    if (!IsActive(manager, id) && *value != persistentData->prevValue) {
        persistentData->prevValue = *value;
        stbsp_snprintf(persistentData->buffer, sizeof(persistentData->buffer), "%u", *value);
    }

    b32 modified = false;
    flags |= InputStringFlag_ReuseWidgetID_;
    if (InputString(name, persistentData->buffer, ARRAY_LENGTH(char, persistentData->buffer), flags, &ValidateInputUInt)) {

        int32 newValue;
        int32 result = sscanf(persistentData->buffer, "%u", &newValue);

        if (result == 1 && newValue != *value) {
            *value = newValue;
            persistentData->prevValue = newValue;

            modified = true;
        }
    }

    b32 isActive = IsActive(manager, id);

    if (!persistentData->wasActive && isActive) {
        persistentData->initialValue = persistentData->prevValue;
    }
    else if (persistentData->wasActive && !isActive) {
        if (!ValidateInputInt(persistentData->buffer)) {
            *value = persistentData->initialValue;
        }

        stbsp_snprintf(persistentData->buffer, sizeof(persistentData->buffer), "%u", *value);
    }

    persistentData->wasActive = isActive;

    return modified;
}


struct InputFloatPersistentData {
    b32 wasActive;
    real32 initialValue;

    real32 prevValue;
    char buffer[32];
};

b32 ValidateInputFloat(const char *buffer, InputStringActiveData *activeData, void *data = NULL) {
    const char *at = buffer;

    if (*at == '\0') {
        return false;
    }

    bool foundDecimal = false;
    bool foundExponent = false;

    if (*at == '-') {
        at++;
    }

    while (*at != '\0') {
        if (*at == '.') {
            if (!foundDecimal && !foundExponent) {
                foundDecimal = true;
                at++;
                continue;
            }
            else {
                return false;
            }
        }

        if (*at == 'e' || *at == 'E') {
            at++;

            if (*at == '-') {
                at++;
            }

            continue;
        }

        if (IsNumeric(*at)) {
            at++;
            continue;
        }

        return false;
    }

    return true;
}

b32 InputFloat(const char *name, real32 *value, uint32 flags) {
    UIManager *manager = GetCurrentUIManager();
    WidgetGroup *group = PeekWidgetGroup(manager);
    WidgetID id = PushWidgetID(manager, name);
    
    static uint32 InputFloatDataID = GenerateDataID("InputFloatBuffer");

    InputFloatPersistentData *persistentData = NULL;
    if (!GetPersistentData(manager, id, InputFloatDataID, (void **)&persistentData)) {
        persistentData = (InputFloatPersistentData *)AllocatePersistentData(manager, id, InputFloatDataID, sizeof(InputFloatPersistentData));

        PrettyPrintFloat(persistentData->buffer, sizeof(persistentData->buffer), *value);
    }

    if (!IsActive(manager, id) && *value != persistentData->prevValue) {
        persistentData->prevValue = *value;
        PrettyPrintFloat(persistentData->buffer, sizeof(persistentData->buffer), *value);
    }


    b32 modified = false;
    flags |= InputStringFlag_ReuseWidgetID_;
    if (InputString(name, persistentData->buffer, ARRAY_LENGTH(char, persistentData->buffer), flags, &ValidateInputFloat)) {

        real32 newValue;
        int32 result = sscanf(persistentData->buffer, "%f", &newValue);

        if (result == 1 && newValue != *value) {
            *value = newValue;
            persistentData->prevValue = newValue;

            modified = true;
        }
    }

    b32 isActive = IsActive(manager, id);

    if (!persistentData->wasActive && isActive) {
        persistentData->initialValue = persistentData->prevValue;
    }
    else if (persistentData->wasActive && !IsActive(manager, id)) {
        if (!ValidateInputFloat(persistentData->buffer, NULL, NULL)) {
            *value = persistentData->initialValue;
        }

        PrettyPrintFloat(persistentData->buffer, sizeof(persistentData->buffer), *value);
    }

    persistentData->wasActive = isActive;

    return modified;
}


struct UISliderActiveData {
    vec2 mouseDeltaFromNubCenter;
};

b32 Sliderf(const char *name, real32 *value, real32 minValue, real32 maxValue, real32 snapValue, uint32 flags) {
    ASSERT(minValue <= maxValue);

    UIManager *manager = GetCurrentUIManager();
    WidgetGroup *group = PeekWidgetGroup(manager);
    WidgetID id = PushWidgetID(manager, name);

    EtcRect originalWidgetBounds = GetNextWidgetBounds(manager, group);
    EtcRect widgetBounds = originalWidgetBounds;

    b32 displayName = GetStyleb(UIStyle_DisplayName, manager);
    if (displayName) {
        UIStyle tempStyles[] = {
            MakeStyleb(UIStyle_DisplayName, false),
            MakeStyleFlowDirection(UIStyle_FlowDirection, UIFlowDirection_Horizontal),
        };

        SetNextWidgetBounds(manager, group, originalWidgetBounds);

        DisplayNameWidgetFunc *displayNameWidgetFunc = GetDisplayNameWidgetFunc(manager);
        displayNameWidgetFunc(name, NULL);


        PushStyles(tempStyles, ARRAY_LENGTH(UIStyle, tempStyles));

        widgetBounds = GetNextWidgetBounds(manager, group);
        widgetBounds.size.x = originalWidgetBounds.size.x - widgetBounds.pos.x;
    }

    widgetBounds.size.y = 21.0f;

    if (displayName) {
        widgetBounds.pos.y += (group->prevWidgetBounds.size.y - widgetBounds.size.y) / 2.0f;
    }

    CommitWidgetBounds(manager, group, widgetBounds);


    real32 valueRange = maxValue - minValue;

    minValue = SnapDown(minValue, snapValue);
    maxValue = SnapUp(maxValue, snapValue);


    real32 minNubWidth = 16.0f;
    real32 nubWidth = 16.0f;
    real32 nubTrackWidth = widgetBounds.size.x - nubWidth;
    real32 nubTrackLeftBound = widgetBounds.pos.x + 0.5f * (widgetBounds.size.x - nubTrackWidth);
    real32 nubTrackRightBound = nubTrackLeftBound + nubTrackWidth;

    real32 nubCenterX = nubTrackLeftBound + nubTrackWidth * SafeRatio((*value - minValue), valueRange, 0.0f);
    real32 nubCenterY = widgetBounds.pos.y + 0.5f * widgetBounds.size.y;
    vec2 nubCenter = V2(nubCenterX, nubCenterY);

    real32 nubPositionX = nubCenterX - 0.5f * nubWidth;
    real32 nubPositionY = widgetBounds.pos.y;
    vec2 nubSize = {nubWidth, widgetBounds.size.y};
    EtcRect nubBounds = {{nubPositionX, nubPositionY}, nubSize};


    b32 modified = false;
    b32 isActive = false;
    b32 mouseOverNub = group->isHoveringGroup && PointInRect(nubBounds, group->localMousePosition);
    b32 mouseOverTrack = group->isHoveringGroup && PointInRect(widgetBounds, group->localMousePosition);

    if (IsActive(manager, id)) {
        if (InputActionHold(Game.inputActions, InputAction_EditorPrimaryClick, true)) {
            UISliderActiveData *activeData = (UISliderActiveData *)manager->activeWidgetData;

            real32 newNubCenterX = group->localMousePosition.x - activeData->mouseDeltaFromNubCenter.x;
            newNubCenterX = Clamp(newNubCenterX, nubTrackLeftBound, nubTrackRightBound);
            nubBounds.pos.x = newNubCenterX - 0.5f * nubWidth;            

            real32 newValue = LinearRemap(nubTrackLeftBound, nubTrackRightBound, newNubCenterX, minValue, maxValue);
            newValue = Snap(newValue, snapValue);
            if (newValue != *value) {
                *value = newValue;
                modified = true;
            }

            newNubCenterX = nubTrackLeftBound + nubTrackWidth * SafeRatio(newValue - minValue, valueRange, 0.0f);
            newNubCenterX = Clamp(newNubCenterX, nubTrackLeftBound, nubTrackRightBound);

            nubBounds.pos.x = newNubCenterX - 0.5f * nubWidth;

            isActive = true;
        }
        else {
            SetActiveWidget(manager, WidgetID_Invalid);
        }
    }
    else if (mouseOverNub || mouseOverTrack) {
        if (IsHovered(manager, id)) {
            if (InputActionPressed(Game.inputActions, InputAction_EditorPrimaryClick, true)) {
                isActive = true;
                SetActiveWidget(manager, id, sizeof(UISliderActiveData));

                UISliderActiveData *activeData = (UISliderActiveData *)manager->activeWidgetData;
                if (mouseOverNub) {
                    activeData->mouseDeltaFromNubCenter = group->localMousePosition - nubCenter;
                }
            }
        }

        SetHoveredWidget(manager, group, id);
    }

    // Render
    vec4 backgroundColor = V4(0.0f, 0.0f, 0.0f, 0.75f);
    vec4 nubColor = isActive ? PASTEL_YELLOW : PASTEL_RED;

    RenderMesh(manager, group, &Game.graphics->glyphQuad, TS(widgetBounds.pos, widgetBounds.size), backgroundColor);
    RenderMesh(manager, group, &Game.graphics->glyphQuad, TS(nubBounds.pos, nubBounds.size), nubColor);

    if (displayName) {
        PopStyles();
    }

    return modified;
}

b32 InputVecNf_(const char *name, real32 *values, uint32 valueCount, uint32 flags) {
    UIManager *manager = GetCurrentUIManager();
    WidgetGroup *group = PeekWidgetGroup(manager);

    b32 displayName = GetStyleb(UIStyle_DisplayName, manager);
    if (displayName) {
        UIStyle tempStyles[] = {
            MakeStyleb(UIStyle_DisplayName, false),
            MakeStyleFlowDirection(UIStyle_FlowDirection, UIFlowDirection_Horizontal),
        };

        DisplayNameWidgetFunc *displayNameWidgetFunc = GetDisplayNameWidgetFunc(manager);
        displayNameWidgetFunc(name, NULL);

        PushStyles(tempStyles, ARRAY_LENGTH(UIStyle, tempStyles));
    }

    EtcRect widgetBounds = GetNextWidgetBounds(manager, group);
    UIPadding internalPadding = {1, 1, 0, 0};

    real32 widthPerComponent = (widgetBounds.size.x - (valueCount - 1) * (internalPadding.left + internalPadding.right)) / (real32)valueCount;
    
    b32 modified = false;
    
    EtcRect componentBounds = {};
    componentBounds.pos = widgetBounds.pos;
    componentBounds.size.x = widthPerComponent;

    for (int i = 0; i < valueCount; i++) {
        SetNextWidgetBounds(manager, group, componentBounds);
        modified |= InputFloat("vec_value", &values[i], flags);

        componentBounds.pos.x += widthPerComponent + internalPadding.left + internalPadding.right;
    }

    if (displayName) {
        PopStyles();
    }

    return modified;
}

b32 InputVec2(const char *name, vec2 *value, uint32 flags) {
    return InputVecNf_(name, (real32 *)value, 2, flags);
}

b32 InputVec3(const char *name, vec3 *value, uint32 flags) {
    return InputVecNf_(name, (real32 *)value, 3, flags);
}

b32 InputVec4(const char *name, vec4 *value, uint32 flags) {
    return InputVecNf_(name, (real32 *)value, 4, flags);
}

b32 InputQuaternion(const char *name, quaternion *value, uint32 flags) {
    if (LengthSq(*value) <= 0) {
        *value = IdentityQuaternion();
    }

    vec3 euler = ToEulerAngles(*value);

    euler.x = RadToDeg(euler.x);
    euler.y = RadToDeg(euler.y);
    euler.z = RadToDeg(euler.z);
    
    b32 modified = InputVec3(name, &euler, flags);

    euler.x = DegToRad(euler.x);
    euler.y = DegToRad(euler.y);
    euler.z = DegToRad(euler.z);

    if (modified) {
        *value = FromEulerAngles(euler);
    }

    return modified;
}

b32 InputVecNi_(const char *name, s32 *values, uint32 valueCount, uint32 flags) {
    UIManager *manager = GetCurrentUIManager();
    WidgetGroup *group = PeekWidgetGroup(manager);

    b32 displayName = GetStyleb(UIStyle_DisplayName, manager);
    if (displayName) {
        UIStyle tempStyles[] = {
            MakeStyleb(UIStyle_DisplayName, false),
            MakeStyleFlowDirection(UIStyle_FlowDirection, UIFlowDirection_Horizontal),
        };

        DisplayNameWidgetFunc *displayNameWidgetFunc = GetDisplayNameWidgetFunc(manager);
        displayNameWidgetFunc(name, NULL);

        PushStyles(tempStyles, ARRAY_LENGTH(UIStyle, tempStyles));
    }

    EtcRect widgetBounds = GetNextWidgetBounds(manager, group);
    UIPadding internalPadding = {1, 1, 0, 0};

    real32 widthPerComponent = (widgetBounds.size.x - (valueCount - 1) * (internalPadding.left + internalPadding.right)) / (real32)valueCount;
    
    b32 modified = false;
    
    EtcRect componentBounds;
    componentBounds.pos = widgetBounds.pos;
    componentBounds.size.x = widthPerComponent;

    for (int i = 0; i < valueCount; i++) {
        SetNextWidgetBounds(manager, group, componentBounds);
        modified |= InputInt("vec_value", &values[i], flags);

        componentBounds.pos.x += widthPerComponent + internalPadding.left + internalPadding.right;
    }

    if (displayName) {
        PopStyles();
    }

    return modified;
}

b32 InputVec2i(const char *name, vec2i *value, uint32 flags) {
    return InputVecNi_(name, (s32 *)value, 2, flags);
}


b32 InputColor(const char *name, Color *value, uint32 flags) {
    UIManager *manager = GetCurrentUIManager();
    WidgetGroup *group = PeekWidgetGroup(manager);

    b32 displayName = GetStyleb(UIStyle_DisplayName, manager);
    if (displayName) {
        UIStyle tempStyles[] = {
            MakeStyleb(UIStyle_DisplayName, false),
            MakeStyleFlowDirection(UIStyle_FlowDirection, UIFlowDirection_Horizontal),
        };

        DisplayNameWidgetFunc *displayNameWidgetFunc = GetDisplayNameWidgetFunc(manager);
        displayNameWidgetFunc(name, NULL);

        PushStyles(tempStyles, ARRAY_LENGTH(UIStyle, tempStyles));
    }
    
    vec4 color = WithAlpha(value->rgb, 1);
    if (HasFlag(flags, UIColorPickerFlag_RenderRYB)) {
        color.rgb = RYBToRGB(color.rgb);
    }
    PushStyle(MakeStyle4(UIStyle_ButtonColor, color));

    if (Button("")) {
        ColorPickerWindowData *windowData = &DebugMem->colorPickerWindowData;

        if (windowData->ownerID == group->prevID) {
            windowData->ownerID = WidgetID_Invalid;
        }
        else {
            windowData->ownerID = group->prevID;
            windowData->originalColor = *value;
            windowData->prevHSV = RGBToHSV(value->rgb);
        }
    }

    PopStyles();

    if (displayName) {
        PopStyles();
    }

    b32 modified = ColorPickerWindow("Color Picker", value, group->prevID, &DebugMem->colorPickerWindowData, flags);

    return modified;
}


void AllocateColorPickerRing(Mesh *mesh) {
    uint32 segments = 64;

    uint32 vertexBufferSize = sizeof(vec3) * 2 * (segments + 1);
    ReallocateMesh(mesh, 2 * (segments + 1), vertexBufferSize, 2 * (segments + 1), 1);
    mesh->primitiveType = PrimitiveType_TriangleStrip;

    mesh->positions = (vec3 *)mesh->vertices;

    for (int i = 0; i <= segments; i++) {
        real32 theta = 2 * _PI * i / (real32)segments;
        vec3 position = V3(cosf(-theta),
                           sinf(-theta),
                           i / (real32)segments * 360.0f);

        mesh->positions[2 * i + 0] = position;
        mesh->positions[2 * i + 1] = position;

        mesh->indices[2 * i + 0] = 2 * i + 0;
        mesh->indices[2 * i + 1] = 2 * i + 1;
    }

    mesh->vertexCount = 2 * (segments + 1);
    mesh->indexCount = 2 * (segments + 1);
    mesh->submeshLengths[0] = mesh->indexCount;
}

void AllocateColorPickerTriangle(Mesh *mesh) {
    uint32 vertexBufferSize = (sizeof(vec3) + sizeof(vec2)) * 3;
    ReallocateMesh(mesh, 3, vertexBufferSize, 3, 1);
    mesh->primitiveType = PrimitiveType_Triangles;
    mesh->vertexCount = 3;
    mesh->indexCount = 3;
    mesh->submeshLengths[0] = 3;

    mesh->positions = (vec3 *)BUFFER_OFFSET(mesh->vertices, 0);
    mesh->texcoords = (vec2 *)BUFFER_OFFSET(mesh->positions, sizeof(vec3) * 3);

    mesh->positions[0] = {cosf(DegToRad(0)), sinf(DegToRad(0))};
    mesh->positions[1] = {cosf(DegToRad(-120)), sinf(DegToRad(-120))};
    mesh->positions[2] = {cosf(DegToRad(-240)), sinf(DegToRad(-240))};

    mesh->texcoords[0] = {1, 1};
    mesh->texcoords[1] = {0, 1};
    mesh->texcoords[2] = {0, 0};

    mesh->indices[0] = 0;
    mesh->indices[1] = 1;
    mesh->indices[2] = 2;
}

b32 ColorPickerHueWidget_(const char *name, vec3 *hsv, vec2 center, real32 outerRadius, uint32 flags) {
    UIManager *manager = GetCurrentUIManager();
    WidgetGroup *group = PeekWidgetGroup(manager);
    WidgetID id = PushWidgetID(manager, name);

    real32 innerRadius = outerRadius - 32.0f;

    real32 mouseDistance = Distance(center, group->localMousePosition);
    b32 isMouseOver = group->isHoveringGroup && innerRadius <= mouseDistance && mouseDistance <= outerRadius;

    b32 isActive = false;
    b32 modified = false;

    if (IsActive(manager, id)) {
        if (InputActionHold(Game.inputActions, InputAction_EditorPrimaryClick, true)) {
            isActive = true;
        }
        else {
            SetActiveWidget(manager, WidgetID_Invalid);
        }
    }
    else if (isMouseOver) {
        if (InputActionPressed(Game.inputActions, InputAction_EditorPrimaryClick, true)) {
            SetActiveWidget(manager, id);
            isActive = true;
        }

        SetHoveredWidget(manager, group, id);
    }

    if (isActive) {
        real32 hue = NormalizeAngleRad(atan2f(-(group->localMousePosition.y - center.y),
                                              group->localMousePosition.x - center.x));
        hue = RadToDeg(hue);

        if (hue != hsv->h) {
            hsv->h = hue;
            modified = true;
        }
    }

    Mesh *mesh = GetMesh(Game.assets, DebugMem->debugUIColorPickerRing);
    ShaderProgram *shader = NULL;
    if (HasFlag(flags, UIColorPickerFlag_RenderRYB)) {
        shader = GetShaderProgram(Game.assets, DebugMem->debugUIColorPickerRingShaderRYB);
    }
    else {
        shader = GetShaderProgram(Game.assets, DebugMem->debugUIColorPickerRingShaderRGB);
    }

    if (mesh != NULL && shader != NULL) {
        uint32 renderKey;
        if (UseMeshInGroup(manager, group, shader, mesh, false, &renderKey)) {
            mat4 model = TS(V3(center, 0), 1);
            SetUniform(&manager->commandList, renderKey, UniformLocation(shader, "model"), model);
            SetUniform(&manager->commandList, renderKey, UniformLocation(shader, "innerRadius"), innerRadius);
            SetUniform(&manager->commandList, renderKey, UniformLocation(shader, "outerRadius"), outerRadius);

            DrawIndexed(&manager->commandList, renderKey, mesh->primitiveType, 0, mesh->indexCount);
        }
    }


    {
        vec2 direction = V2(cosf(DegToRad(-hsv->h)), sinf(DegToRad(-hsv->h)));
        vec2 directionPerp = LeftHandPerp(direction);
        quaternion rotation = AxisAngle(V3(0, 0, 1), DegToRad(-hsv->h));

        real32 innerBoxHeight = 3;
        real32 outerBoxHeight = 5;

        mat4 model0 = TRS(V3(center + direction * innerRadius - outerBoxHeight * directionPerp, 0),
                          rotation,
                          V3(outerRadius - innerRadius, 2 * outerBoxHeight, 1.0f));
        RenderMesh(manager, group, &Game.graphics->glyphQuad, model0, WHITE);


        vec3 rgb = HSVToRGB(V3(hsv->h, 1, 1));
        if (HasFlag(flags, UIColorPickerFlag_RenderRYB)) {
            rgb = RYBToRGB(rgb);
        }

        mat4 model1 = TRS(V3(center + direction * (innerRadius + 2) - innerBoxHeight * directionPerp, 0),
                          rotation,
                          V3(outerRadius - innerRadius - 4, 2 * innerBoxHeight, 1.0f));
        RenderMesh(manager, group, &Game.graphics->glyphQuad, model1, V4(rgb, 1));
    }

    return modified;
}

b32 ColorPickerSVWidget_(const char *name, vec3 *hsv, vec2 center, real32 outerRadius, uint32 flags) {
    UIManager *manager = GetCurrentUIManager();
    WidgetGroup *group = PeekWidgetGroup(manager);
    WidgetID id = PushWidgetID(manager, name);

    real32 triangleRadius = outerRadius - 40.0f;
    real32 pointRadius = 8.0f;


    b32 modified = false;
    b32 isActive = false;

    vec2 trianglePoints[] = {
        center + triangleRadius * V2(cosf(-DegToRad(hsv->h)), sinf(-DegToRad(hsv->h))),
        center + triangleRadius * V2(cosf(-DegToRad(hsv->h + 120)), sinf(-DegToRad(hsv->h + 120))),
        center + triangleRadius * V2(cosf(-DegToRad(hsv->h + 240)), sinf(-DegToRad(hsv->h + 240))),
    };

    vec3 mouseBarycentric = CartesianToBarycentric(trianglePoints[0], trianglePoints[1], trianglePoints[2],
                                                   group->localMousePosition);
    b32 isMouseOver = (0.0f <= mouseBarycentric.x && mouseBarycentric.x <= 1.0f &&
                       0.0f <= mouseBarycentric.y && mouseBarycentric.y <= 1.0f &&
                       0.0f <= mouseBarycentric.z && mouseBarycentric.z <= 1.0f);

    if (IsActive(manager, id)) {
        if (InputActionHold(Game.inputActions, InputAction_EditorPrimaryClick, true)) {
            isActive = true;
        }
        else {
            SetActiveWidget(manager, WidgetID_Invalid);
        }
    }
    else if (isMouseOver) {
        if (InputActionPressed(Game.inputActions, InputAction_EditorPrimaryClick, true)) {
            SetActiveWidget(manager, id);
            isActive = true;
        }

        SetHoveredWidget(manager, group, id);
    }

    real32 chroma = hsv->v * hsv->s;

    if (isActive) {
        vec2 closestPoint = ClosestPointOnTriangle(trianglePoints[0], trianglePoints[1], trianglePoints[2],
                                                   group->localMousePosition);
        mouseBarycentric = CartesianToBarycentric(trianglePoints[0], trianglePoints[1], trianglePoints[2],
                                                  closestPoint);

        vec2 cartesian = BarycentricToCartesian(V2(1, 1), V2(0, 1), V2(0, 0), mouseBarycentric);
        cartesian = Clamp(cartesian, V2(0), V2(1));

        if (cartesian.x != chroma || cartesian.y != hsv->v) {
            chroma = cartesian.x;
            hsv->v = cartesian.y;
            hsv->s = SafeRatio(cartesian.x, hsv->v, 0.0f);
            modified = true;
        }
    }

    Mesh *mesh = GetMesh(Game.assets, DebugMem->debugUIColorPickerTriangle);
    ShaderProgram *shader = NULL;
    if (HasFlag(flags, UIColorPickerFlag_RenderRYB)) {
        shader = GetShaderProgram(Game.assets, DebugMem->debugUIColorPickerTriangleShaderRYB);
    }
    else {
        shader = GetShaderProgram(Game.assets, DebugMem->debugUIColorPickerTriangleShaderRGB);
    }

    if (mesh != NULL && shader != NULL) {
        uint32 renderKey;
        if (UseMeshInGroup(manager, group, shader, mesh, false, &renderKey)) {
            mat4 model = TRS(V3(center, 0), AxisAngle(V3(0, 0, 1), DegToRad(-hsv->h)), triangleRadius);
            SetUniform(&manager->commandList, renderKey, UniformLocation(shader, "model"), model);
            SetUniform(&manager->commandList, renderKey, UniformLocation(shader, "hue"), hsv->h);

            DrawIndexed(&manager->commandList, renderKey, mesh->primitiveType, 0, mesh->indexCount);
        }
    }


    // @GACK: this group is only here to resolve mesh batch ordering problems!
    WidgetID circleGroup = PushWidgetID(manager, "CircleGroup");
    PushWidgetGroup(circleGroup, V2(0), group->localBounds, group->localBounds);

    vec3 barycentric = CartesianToBarycentric(V2(1, 1), V2(0, 1), V2(0, 0), V2(chroma, hsv->v));
    barycentric = Clamp(barycentric, V3(0), V3(1));
    vec2 colorPoint = BarycentricToCartesian(trianglePoints[0], trianglePoints[1], trianglePoints[2],
                                             barycentric);
    if (Mesh *mesh = GetMesh(Game.assets, DebugMem->circle)) {
        mat4 model0 = TS(V3(colorPoint, 0), pointRadius + 2);
        RenderMesh(manager, PeekWidgetGroup(manager), mesh, model0, WHITE);

        mat4 model1 = TS(V3(colorPoint, 0), pointRadius);
        vec3 rgb = HSVToRGB(*hsv);
        if (HasFlag(flags, UIColorPickerFlag_RenderRYB)) {
            rgb = RYBToRGB(rgb);
        }

        RenderMesh(manager, PeekWidgetGroup(), mesh, model1, V4(rgb, 1));
    }

    PopWidgetGroup();

    return modified;
}

b32 ColorPicker(const char *name, vec3 *value, EtcRect bounds, uint32 flags) {
    vec4 value4 = V4(*value, 1.0f);
    if (ColorPicker(name, &value4, bounds, flags)) {
        *value = value4.rgb;
        return true;
    }

    return false;
}

b32 ColorPicker(const char *name, vec4 *value, EtcRect bounds, uint32 flags) {
    UIManager *manager = GetCurrentUIManager();
    WidgetGroup *parentGroup = PeekWidgetGroup(manager);
    WidgetID id = PushWidgetID(manager, name);

    EtcRect localBounds = {{}, bounds.size};
    PushWidgetGroup(id, bounds.pos, localBounds, localBounds);

    vec3 hsv;
    if (HasFlag(flags, UIColorPickerFlag_InputHSV)) {
        hsv = value->rgb;
    }
    else {
        hsv = RGBToHSV(value->rgb);
    }


    UIPadding padding = GetStylePadding(UIStyle_Padding, manager);
    real32 outerRadius = 0.5f * Min(localBounds.size.x - (padding.left + padding.right),
                                    localBounds.size.y - (padding.top + padding.bottom));
    vec2 center = localBounds.pos + 0.5f * localBounds.size;

    b32 modified = false;

    modified |= ColorPickerHueWidget_("HueWidget", &hsv, center, outerRadius, flags);
    modified |= ColorPickerSVWidget_("SVWidget", &hsv, center, outerRadius, flags);

    if (modified) {
        if (HasFlag(flags, UIColorPickerFlag_InputHSV)) {
            *value = WithAlpha(hsv, value->a);
        }
        else {
            vec3 rgb = HSVToRGB(hsv);
            *value = WithAlpha(rgb, value->a);
        }
    }

    PopWidgetGroup();

    return modified;
}


struct InputIntOrNormFloatPersistentData {
    b32 wasActive;
    int32 initialValue;
    int32 prevValue;
    char buffer[32];
};

enum InputIntOrNormFloatFlags {
    InputIntOrNormFloatFlag_DisplayFloat = 1 << 0,
};


b32 ValidateInputIntOrFloat(const char *buffer, InputStringActiveData *activeData = NULL, void *data = NULL) {
    return ValidateInputFloat(buffer, activeData, data) || ValidateInputInt(buffer, activeData, data);
        
}

b32 InputIntOrNormFloat(const char *name, real32 *value, int32 intMin, int32 intMax, uint32 flags) {
    UIManager *manager = GetCurrentUIManager();
    WidgetGroup *group = PeekWidgetGroup(manager);
    WidgetID id = PushWidgetID(manager, name);
    
    static uint32 InputIntDataID = GenerateDataID("InputIntOrNormFloatBuffer");

    InputIntPersistentData *persistentData = NULL;
    if (!GetPersistentData(manager, id, InputIntDataID, (void **)&persistentData)) {
        persistentData = (InputIntPersistentData *)AllocatePersistentData(manager, id, InputIntDataID, sizeof(InputIntPersistentData));

        if (HasFlag(flags, InputIntOrNormFloatFlag_DisplayFloat)) {
            stbsp_snprintf(persistentData->buffer, sizeof(persistentData->buffer), "%f", *value);
        }
        else {
            // @TODO: make sure value is normalized
            int32 intValue = (int32)Roundf(intMin + *value * (intMax - intMin));
            intValue = Clamp(intValue, intMin, intMax);
            stbsp_snprintf(persistentData->buffer, sizeof(persistentData->buffer), "%d", intValue);
        }
    }

    if (!IsActive(manager, id) && *value != persistentData->prevValue) {
        persistentData->prevValue = *value;
        if (HasFlag(flags, InputIntOrNormFloatFlag_DisplayFloat)) {
            stbsp_snprintf(persistentData->buffer, sizeof(persistentData->buffer), "%f", *value);
        }
        else {
            // @TODO: make sure value is normalized
            int32 intValue = (int32)Roundf(intMin + *value * (intMax - intMin));
            intValue = Clamp(intValue, intMin, intMax);
            stbsp_snprintf(persistentData->buffer, sizeof(persistentData->buffer), "%d", intValue);
        }
    }

    b32 modified = false;
    flags |= InputStringFlag_ReuseWidgetID_;
    if (InputString(name, persistentData->buffer, ARRAY_LENGTH(char, persistentData->buffer), flags, &ValidateInputIntOrFloat)) {

        r32 newValue;
        int32 result = 0;
        
        if (ValidIntString(persistentData->buffer)) {
            int32 intValue;
            result = sscanf(persistentData->buffer, "%d", &intValue);

            intValue = Clamp(intValue, intMin, intMax);
            newValue = (intValue - intMin) / (real32)(intMax - intMin);
        }
        else {
            result = sscanf(persistentData->buffer, "%f", &newValue);
        }

        if (result == 1 && newValue != *value) {
            *value = newValue;
            persistentData->prevValue = newValue;

            modified = true;
        }
    }

    b32 isActive = IsActive(manager, id);

    if (!persistentData->wasActive && isActive) {
        persistentData->initialValue = persistentData->prevValue;
    }
    else if (persistentData->wasActive && !isActive) {
        if (!ValidateInputIntOrFloat(persistentData->buffer)) {
            *value = persistentData->initialValue;
        }
        if (HasFlag(flags, InputIntOrNormFloatFlag_DisplayFloat)) {
            stbsp_snprintf(persistentData->buffer, sizeof(persistentData->buffer), "%f", *value);
        }
        else {
            int32 intValue = (int32)Roundf(intMin + *value * (intMax - intMin));
            intValue = Clamp(intValue, intMin, intMax);
            stbsp_snprintf(persistentData->buffer, sizeof(persistentData->buffer), "%d", intValue);
        }
    }

    if (modified) {
        *value = Clamp(*value, 0.0f, 1.0f);
    }

    persistentData->wasActive = isActive;

    return modified;
}



b32 ColorSliderf_(const char *name, real32 *value, real32 min, real32 max, int32 intMin, int32 intMax, uint32 flags = 0) {
    UIManager *manager = GetCurrentUIManager();
    WidgetGroup *group = PeekWidgetGroup(manager);
    
    b32 modified = false;

    EtcRect widgetBounds = GetNextWidgetBounds(manager, group);

    EtcRect sliderBounds = {};
    sliderBounds.pos = widgetBounds.pos;
    sliderBounds.size.x = Roundf(0.75f * widgetBounds.size.x);
    SetNextWidgetBounds(manager, group, sliderBounds);
    modified |= Sliderf(name, value, min, max);


    EtcRect fieldBounds = {};
    fieldBounds.pos = sliderBounds.pos;
    fieldBounds.pos.x += sliderBounds.size.x + 2;
    fieldBounds.size.x = widgetBounds.size.x - sliderBounds.size.x - 2;
    SetNextWidgetBounds(manager, group, fieldBounds);
    
    PushStyle(MakeStyleb(UIStyle_DisplayName, false));
    real32 value01 = (*value - min) / (max - min);
    int32 intValue = (int32)Roundf(intMin + value01 * (intMax - intMin));
    // if (InputInt(name, &intValue)) {
    //     intValue = Clamp(intValue, intMin, intMax);
    //     value01 = (intValue - intMin) / (real32)(intMax - intMin);
    //     *value = min + value01 * (max - min);
    //     modified = true;
    // }
    if (InputIntOrNormFloat(name, value, intMin, intMax)) {
        modified = true;
    }

    PopStyles();

    return modified;
}

b32 ColorPickerWindow(const char *name, vec4 *value, WidgetID ownerID, ColorPickerWindowData *windowData, uint32 flags) {
    UIManager *manager = GetCurrentUIManager();
    b32 modified = false;

    UIPadding padding = GetStylePadding(UIStyle_Padding, manager);
    const real32 colorPickerWidth = 256;
    const real32 slidersWidth = 256;
        
    if (ownerID != WidgetID_Invalid && ownerID == windowData->ownerID && !windowData->renderedThisFrame) {
        EtcRect windowBounds = {
            V2(Game.graphics->width * 0.25f + 8, Game.graphics->height * 0.5f + 2),
            V2(colorPickerWidth + slidersWidth + 2 * (padding.left + padding.right), 0)
        };

        UIStyle styles[] = {
            MakeStyleb(UIStyle_DisplayName, true),
        };
        PushStyles(styles, ARRAY_LENGTH(UIStyle, styles));

        PushWindow("Color Picker", windowBounds.pos, windowBounds.size,
                   UIWindowFlag_ShowTitle | UIWindowFlag_Absolute | UIWindowFlag_ResizeToContentVertical);
        WidgetGroup *windowGroup = PeekWidgetGroup(manager);

        EtcRect colorPickerBounds = GetNextWidgetBounds(manager, windowGroup);
        colorPickerBounds.size.x = colorPickerWidth;

        vec2 sliderAnchor = colorPickerBounds.pos;
        sliderAnchor.x += colorPickerBounds.size.x + padding.left + padding.right;
        EtcRect sliderBounds = {{0, 0}, {slidersWidth, 0}};
        WidgetGroup *sliderGroup = PushWidgetGroup(PushWidgetID(manager, "Slider Group"),
                                                   sliderAnchor, sliderBounds, sliderBounds,
                                                   WidgetGroupFlag_ResizeToContentVertical);


        vec3 hsv;
        if (HasFlag(flags, UIColorPickerFlag_InputHSV)) {
            hsv = value->hsv;
        }
        else {
            hsv = windowData->prevHSV;
        }

        b32 hsvModified = false;
        hsvModified |= ColorSliderf_("H", &hsv.h, 0.0f, 360.0f, 0, 259);
        hsvModified |= ColorSliderf_("S", &hsv.s, 0.0f, 1.0f, 0, 255);
        hsvModified |= ColorSliderf_("V", &hsv.v, 0.0f, 1.0f, 0, 255);

        if (hsvModified) {
            hsv.h = NormalizeAngleDeg(hsv.h);

            if (HasFlag(flags, UIColorPickerFlag_InputHSV)) {
                value->hsv = hsv;
            }
            else {
                value->rgb = HSVToRGB(hsv);
            }

            windowData->prevHSV = hsv;
            modified = true;
        }

        EtcRect spacerBounds = GetNextWidgetBounds(manager, sliderGroup);
        spacerBounds.size.y = 20;
        CommitWidgetBounds(manager, sliderGroup, spacerBounds);

        b32 rgbaModified = false;
        vec4 rgba;
        if (HasFlag(flags, UIColorPickerFlag_InputHSV)) {
            rgba = WithAlpha(HSVToRGB(value->rgb), value->a);
        }
        else {
            rgba = *value;
        }

        if (HasFlag(flags, UIColorPickerFlag_RenderRYB)) {
            rgbaModified |= ColorSliderf_("R", &rgba.r, 0.0f, 1.0f, 0, 255);
            rgbaModified |= ColorSliderf_("Y", &rgba.g, 0.0f, 1.0f, 0, 255);
            rgbaModified |= ColorSliderf_("B", &rgba.b, 0.0f, 1.0f, 0, 255);
            rgbaModified |= ColorSliderf_("A", &rgba.a, 0.0f, 1.0f, 0, 255);
        }
        else {
            rgbaModified |= ColorSliderf_("R", &rgba.r, 0.0f, 1.0f, 0, 255);
            rgbaModified |= ColorSliderf_("G", &rgba.g, 0.0f, 1.0f, 0, 255);
            rgbaModified |= ColorSliderf_("B", &rgba.b, 0.0f, 1.0f, 0, 255);
            rgbaModified |= ColorSliderf_("A", &rgba.a, 0.0f, 1.0f, 0, 255);
        }

        if (rgbaModified) {
            if (HasFlag(flags, UIColorPickerFlag_InputHSV)) {
                value->rgb = RGBToHSV(rgba.rgb);
                value->a = rgba.a;
            }
            else {
                *value = rgba;
            }

            windowData->prevHSV = RGBToHSV(value->rgb);
            modified = true;
        }


        vec3 originalSwatchRGB = windowData->originalColor.rgb;
        vec3 swatchRGB = value->rgb;

        if (HasFlag(flags, UIColorPickerFlag_RenderRYB)) {
            originalSwatchRGB = RYBToRGB(originalSwatchRGB);
            swatchRGB = RYBToRGB(swatchRGB);
        }

        UIPadding internalPadding = {1, 1, 0, 0};
        EtcRect swatchesBounds = GetNextWidgetBounds(manager, sliderGroup);
        swatchesBounds.size.y = 48.0f;

        {
            EtcRect swatchBounds;
            swatchBounds.pos = swatchesBounds.pos;
            swatchBounds.size.x = swatchesBounds.size.x / 2.0f - 1.0f;
            swatchBounds.size.y = swatchesBounds.size.y;
            SetNextWidgetBounds(manager, sliderGroup, swatchBounds);

            vec4 color = WithAlpha(originalSwatchRGB, 1.0f);
            PushStyle(MakeStyle4(UIStyle_ButtonColor, color));

            if (Button("")) {
                *value = windowData->originalColor;
                windowData->prevHSV = RGBToHSV(value->rgb);
            }

            PopStyles();
        }

        {
            EtcRect swatchBounds;
            swatchBounds.pos = swatchesBounds.pos;
            swatchBounds.pos.x = swatchBounds.pos.x + swatchesBounds.size.x / 2.0f + 1.0f;
            swatchBounds.size.x = swatchesBounds.size.x / 2.0f - 1.0f;
            swatchBounds.size.y = swatchesBounds.size.y;
            SetNextWidgetBounds(manager, sliderGroup, swatchBounds);

            vec4 color = WithAlpha(swatchRGB, 1.0f);
            PushStyle(MakeStyle4(UIStyle_ButtonColor, color));

            if (Button("")) {
            }

            PopStyles();
        }

        PopWidgetGroup();

        
        colorPickerBounds.size.y = windowGroup->prevWidgetBounds.size.y;
        if (ColorPicker("Color Picker", &windowData->prevHSV, colorPickerBounds,
                        AddFlag(flags, UIColorPickerFlag_InputHSV))) {
            if (HasFlag(flags, UIColorPickerFlag_InputHSV)) {
                value->hsv = windowData->prevHSV;
            }
            else {
                value->rgb = HSVToRGB(windowData->prevHSV);
            }
            
            modified = true;
        }

        if (InputActionPressed(Game.inputActions, InputAction_EditorPrimaryClick) &&
            !windowGroup->isHoveringGroup) {
            windowData->ownerID = WidgetID_Invalid;
        }

        PopWindow();
        PopStyles();

        windowData->renderedThisFrame = true;
    }

    return modified;
}

b32 InputTransform(const char *name, Transform *value, uint32 flags) {
    UIManager *manager = GetCurrentUIManager();
    WidgetID id = PushWidgetID(manager, name);

    b32 modified = false;

    modified |= InputVec3("Position", &value->position, flags);
    modified |= InputQuaternion("Rotation", &value->rotation, flags);
    modified |= InputVec3("Scale", &value->scale, flags);

    return modified;
}

struct InputDynamicArrayPersistentData {
    uint32 editIndex;
};

b32 InputDynamicArray_(const char *name, EntityDB *entityDB, SchemaTypeDescription templateType, uint32 *arrayCapacity, uint32 *arrayCount, void **arrayData, uint32 flags) {
    b32 result = false;
    
    UIManager *manager = GetCurrentUIManager();
    WidgetGroup *group = PeekWidgetGroup(manager);
    WidgetID id = PushWidgetID(manager, name);
    
    static uint32 InputDynamicArrayDataID = GenerateDataID("InputDynamicArrayBuffer");

    InputDynamicArrayPersistentData *persistentData = NULL;
    if (!GetPersistentData(manager, id, InputDynamicArrayDataID, (void **)&persistentData)) {
        persistentData = (InputDynamicArrayPersistentData *)AllocatePersistentData(manager, id, InputDynamicArrayDataID, sizeof(InputDynamicArrayPersistentData));

        persistentData->editIndex = *arrayCount;
    }
        
    char *indexStr = PushArray(Game.frameMem, char, 8);
    stbsp_snprintf(indexStr, 8, "%d", persistentData->editIndex);
    PushWidgetID(manager, indexStr);
    
    uint32 elemSize = templateType.sizeInBytes;

    // @NOTE: this means we cant have editors for DynamicArrays<DynamicArray<>>, or
    // of tables, because we cant know their size without the templates
    if (elemSize == 0) {
        return false;
    }
    // @TODO: RemoveAtIndex

    if (PushTreeGroup(name)) {
        int nameLength = 64;
        char *buttonName = PushArray(Game.frameMem, char, nameLength);
        Plat.ClearMem(buttonName, nameLength);
        stbsp_snprintf(buttonName, nameLength, "PushBack_%s", name);

        if (Button(buttonName)) {
            void *data = PushSize(Game.frameMem, elemSize);
            Plat.ClearMem(data, elemSize);
            PushBack(arrayCapacity, arrayCount, arrayData, elemSize, data);
            result = true;
        }
        
        for (int i = 0; i < *arrayCount; i++) {
            char *indexName = PushArray(Game.frameMem, char, nameLength);
            Plat.ClearMem(indexName, nameLength);
            stbsp_snprintf(indexName, nameLength, "%s[%d]", name, i);

            char *buttonName = PushArray(Game.frameMem, char, nameLength);
            Plat.ClearMem(buttonName, nameLength);
            stbsp_snprintf(buttonName, nameLength, "Remove[%d]", i);

            if (ButtonConfirm(buttonName)) {
                RemoveAtIndex(arrayCapacity, arrayCount, arrayData, elemSize, i);
                result = true;
            }

            void *elemPtr = (void *)((uint8 *)*arrayData + (elemSize * i));
            result |= InputSchemaDataValue_(indexName, entityDB, templateType, elemPtr, flags);
        }

        PopTreeGroup();

        return result;
    }

    return false;
}

b32 InputDynamicArray(const char *name, EntityDB *entityDB, Schema *schema, uint32 *arrayCapacity, uint32 *arrayCount, void **arrayData, uint32 flags) {
    b32 result = false;
    
    UIManager *manager = GetCurrentUIManager();
    WidgetGroup *group = PeekWidgetGroup(manager);
    WidgetID id = PushWidgetID(manager, name);
    
    static uint32 InputDynamicArrayDataID = GenerateDataID("InputDynamicArrayBuffer");

    InputDynamicArrayPersistentData *persistentData = NULL;
    if (!GetPersistentData(manager, id, InputDynamicArrayDataID, (void **)&persistentData)) {
        persistentData = (InputDynamicArrayPersistentData *)AllocatePersistentData(manager, id, InputDynamicArrayDataID, sizeof(InputDynamicArrayPersistentData));

        persistentData->editIndex = *arrayCount;
    }
        
    char *indexStr = PushArray(Game.frameMem, char, 8);
    stbsp_snprintf(indexStr, 8, "%d", persistentData->editIndex);
    PushWidgetID(manager, indexStr);
    
    uint32 elemSize = schema->sizeInBytes;

    // @NOTE: this means we cant have editors for DynamicArrays<DynamicArray<>>, or
    // of tables, because we cant know their size without the templates
    if (elemSize == 0) {
        return false;
    }
    // @TODO: RemoveAtIndex

    if (PushTreeGroup(name)) {
        int nameLength = 64;
        char *buttonName = PushArray(Game.frameMem, char, nameLength);
        Plat.ClearMem(buttonName, nameLength);
        stbsp_snprintf(buttonName, nameLength, "PushBack_%s", name);

        if (Button(buttonName)) {
            void *data = PushSize(Game.frameMem, elemSize);
            Plat.ClearMem(data, elemSize);
            PushBack(arrayCapacity, arrayCount, arrayData, elemSize, data);
            result = true;
        }
        
        for (int i = 0; i < *arrayCount; i++) {
            char *indexName = PushArray(Game.frameMem, char, nameLength);
            Plat.ClearMem(indexName, nameLength);
            stbsp_snprintf(indexName, nameLength, "%s[%d]", name, i);

            char *buttonName = PushArray(Game.frameMem, char, nameLength);
            Plat.ClearMem(buttonName, nameLength);
            stbsp_snprintf(buttonName, nameLength, "Remove[%d]", i);

            if (ButtonConfirm(buttonName)) {
                RemoveAtIndex(arrayCapacity, arrayCount, arrayData, elemSize, i);
                result = true;
            }

            void *elemPtr = (void *)((uint8 *)*arrayData + (elemSize * i));
            result |= InputSchemaValue(indexName, entityDB, schema, elemPtr, flags);
        }

        PopTreeGroup();

        return result;
    }

    return false;
}

b32 InputDynamicArray(const char *name, EntityDB *entityDB, Schema *schema, DynamicArray_Untyped *array, uint32 flags) {
    return InputDynamicArray(name, entityDB, schema, &array->capacity, &array->count, &array->data, flags);
}


struct InputFuncHandlePersistentData {
    char buffer[128];
};

b32 ValidateInputFuncHandle(const char *buffer, InputStringActiveData *activeData, void *data = NULL) {
    if (StringLength(buffer) == 0) {
        return true;
    }

    PersistentFuncType funcType = *(PersistentFuncType *)data;

    if (funcType == PersistentFuncType_Invalid) {
        return false;
    }

    uint32 tableIndex = GetPersistentFunctionIndex(&Game.persistentFunctionTable, buffer, funcType);
    return tableIndex != PersistentFunctionIndex_Invalid;
}

b32 InputFuncHandle(const char *name, PersistentFuncType funcType, FuncHandle *funcHandle, uint32 flags) {
    ASSERT(funcType != PersistentFuncType_Invalid);

    UIManager *manager = GetCurrentUIManager();
    WidgetGroup *group = PeekWidgetGroup(manager);
    WidgetID id = PushWidgetID(manager, name);
    
    static uint32 InputFuncHandleDataID = GenerateDataID("InputFuncHandleDataBuffer");

    funcHandle->funcType = funcType;

    InputFuncHandlePersistentData *persistentData = NULL;
    if (!GetPersistentData(manager, id, InputFuncHandleDataID, (void **)&persistentData)) {
        persistentData = (InputFuncHandlePersistentData *)AllocatePersistentData(manager, id, InputFuncHandleDataID, sizeof(InputFuncHandlePersistentData));

        if (FuncHandleIsValid(*funcHandle)) {
            const char *funcName = GetPersistentFunctionName(&Game.persistentFunctionTable, funcHandle->tableIndex, funcType);
            if (funcName != NULL) {
                Plat.CopyMem(persistentData->buffer, funcName, StringLength(funcName));    
            }
        }
    }

    flags |= InputStringFlag_ReuseWidgetID_;
    if (InputString(name, persistentData->buffer, 128, flags, ValidateInputFuncHandle, &funcType)) {
        uint32 tableIndex = GetPersistentFunctionIndex(&Game.persistentFunctionTable, persistentData->buffer, funcType);
        if (tableIndex != PersistentFunctionIndex_Invalid) {
            funcHandle->funcType = funcType;
            funcHandle->tableIndex = tableIndex;
            return true;
        }
        else {
            *funcHandle = {};
        }
    }

    return false;
}

b32 InputFuncDataPair(const char *name, EntityDB *entityDB, Schema *schema, FuncDataPair *funcDataPair, u32 flags) {

    Schema *funcHandleSchema = Game.GetSchema(schema->_func_data_pair.fields[0].type.typeName);

    if (funcHandleSchema == NULL) {
        return false;
    }
    
    PersistentFuncType funcType = funcHandleSchema->_func_handle.funcType;

    if (InputFuncHandle(name, funcType, (FuncHandle *)&funcDataPair->funcHandle)) {
        return true;
    }

    if (!FuncHandleIsValid_(funcDataPair->funcHandle)) {
        return false;
    }

    const char *schemaName = GetPersistentFunctionRequiredSchemaDataTypeName(&Game.persistentFunctionTable,
                                                                             funcDataPair->funcHandle.tableIndex,
                                                                             funcType);

    SchemaData *schemaData = (SchemaData *)&funcDataPair->data;
    schemaData->schemaName = (char *)schemaName;

    if (InputSchemaData(name, entityDB, schemaName, schemaData)) {
        return true;
    }

    return false;
}


b32 InputBitfield(const char *name, Schema *schema, uint32 *value, uint32 flags) {
    if (schema == NULL || schema->type != SchemaType_Bitfield) {
        return false;
    }
    
    b32 result = false;
    const char **fieldNames = schema->_bitfield.stringValues;
    uint32 fieldCount = schema->_bitfield.count;
    for (uint32 i = 0; i < fieldCount; i++) {
        bool set = HasFlag(*value, 1 << i);
        
        if (Toggle(fieldNames[i], &set, flags)) {
            result = true;

            if (set) {
                *value = AddFlag(*value, 1 << i);
            }
            else {
                *value = ClearFlag(*value, 1 << i);   
            }
        }
    }

    return result;
}


b32 InputBitarray(const char *name, Schema *schema, uint32 count, uint32 *bits, uint32 flags) {
    if (schema == NULL || schema->type != SchemaType_Bitarray) {
        return false;
    }
    
    b32 result = false;
    const char **indexNames = schema->_bitarray.stringValues;
    uint32 arrayCount = schema->_bitarray.count;
    for (uint32 i = 0; i < arrayCount; i++) {
        bool set = TestBit_(bits, i);
        if (Toggle(indexNames[i], &set, flags)) {
            result = true;

            if (set) {
                SetBit_(bits, i);
            }
            else {
                ClearBit_(bits, i);
            }
        }
    }

    return result;
}


enum DropdownFlags {
    DropdownFlag_Open = 1 << 0,
};

#define DROPDOWN_BUFFER_LENGTH 128
struct InputDropdownPersistentData {
    // currentIndex so we dont have to pass one in if we dont want to
    bool filterUsingBuffer;
    b32 dropped;
    char buffer[DROPDOWN_BUFFER_LENGTH];

    b32 inputStringWasActive;
    uint32 initialValue;
};

struct DropdownValidateData {
    const char **strings;
    u32 count;
    u32 strOffset;
};

b32 DropdownStringMatchesInput_(const char *string, const char *input, uint32 inputLength) {
    b32 matchCase = false;

    for (int i = 0; i < inputLength; i++) {
        if (input[i] == '\0') {
            inputLength = i;
            break;
        }

        if (IsUppercaseASCII(input[i])) {
            matchCase = true;
        }
    }

    // @TODO: fuzzy matching
    const char *c = string;
    for (int i = 0; i < inputLength; i++) {
        if (*c == '\0') {
            return false;
        }

        if (!IsAlphaASCII(input[i]) || matchCase) {
            if (*c != input[i]) {
                return false;
            }
        }
        else if (LowercaseASCII(*c) != LowercaseASCII(input[i])) {
            return false;
        }
            
        ++c;
    }

    return true;
}

b32 ValidateDropdown(const char *str, InputStringActiveData *activeData, void *data) {
    uint32 strLength = StringLength(str);

    DropdownValidateData dropdownData = *(DropdownValidateData *)data;
    for (int i = 0; i < dropdownData.count; i++) {
        if (DropdownStringMatchesInput_(dropdownData.strings[i] + dropdownData.strOffset, str, strLength)) {
            return true;
        }
    }

    return false;
}

b32 Dropdown_(const char *name, const char **strings, uint32 count, uint32 *index, uint32 strOffset, uint32 flags) {
    b32 result = false;

    UIManager *manager = GetCurrentUIManager();
    WidgetGroup *group = PeekWidgetGroup(manager);
    WidgetID id = PushWidgetID(manager, name);
    
    static uint32 InputDropdownDataID = GenerateDataID("InputDropdownBuffer");

    InputDropdownPersistentData *persistentData = NULL;
    if (!GetPersistentData(manager, id, InputDropdownDataID, (void **)&persistentData)) {
        persistentData = (InputDropdownPersistentData *)AllocatePersistentData(manager, id, InputDropdownDataID, sizeof(InputDropdownPersistentData));
        Plat.ClearMem(persistentData, sizeof(InputDropdownPersistentData));

        if (HasFlag(flags, DropdownFlag_Open)) {
            persistentData->dropped = true;
        }
        else {
            persistentData->dropped = false;
        }
        
        if (*index < count) {
            Plat.CopyMem(persistentData->buffer, strings[*index] + strOffset, StringLength(strings[*index]) - strOffset);
        }
    }

    FontTable *font = GetFontTable(Game.assets, GetStyleFont(UIStyle_Font, manager));
    if (font == NULL) {
        return false;
    }

    real32 emSize = GetStylef(UIStyle_FontSize, manager);

    b32 displayName = GetStyleb(UIStyle_DisplayName, manager);
    if (displayName) {
        UIStyle tempStyles[] = {
            MakeStyleb(UIStyle_DisplayName, false),
            MakeStyleFlowDirection(UIStyle_FlowDirection, UIFlowDirection_Horizontal),
        };

        DisplayNameWidgetFunc *displayNameWidgetFunc = GetDisplayNameWidgetFunc(manager);
        displayNameWidgetFunc(name, NULL);

        PushStyles(tempStyles, ARRAY_LENGTH(UIStyle, tempStyles));
    }

    EtcRect textBounds;
    vec2 textPosition;
    EtcRect widgetBounds = GetNextTextWidgetBounds(manager, group, font, emSize, &textBounds, &textPosition);

    EtcRect inputStringBounds;
    inputStringBounds.pos = widgetBounds.pos;
    inputStringBounds.size.x = widgetBounds.size.x - widgetBounds.size.y;

    SetNextWidgetBounds(manager, group, inputStringBounds);

    // @NOTE: this is used because when you click after filtering with a string,
    // the inputString will no longer be active, it'll fail the validate function,
    // and then all the filtered results will disappear, even tho they're still rendered
    // on the frame where you clicked
    char *preValidatedBuffer = PushArray(Game.frameMem, char, StringLength(persistentData->buffer) + 1);
    Plat.ClearMem(preValidatedBuffer, StringLength(persistentData->buffer) + 1 * sizeof(char));
    Plat.CopyMem(preValidatedBuffer, persistentData->buffer, sizeof(char) * StringLength(persistentData->buffer) + 1);

    DropdownValidateData validateData = {strings, count, strOffset};
    uint32 inputStringFlags = AddFlag(flags, InputStringFlag_ReturnTrueOnEnter);
    if (InputString(name, persistentData->buffer, sizeof(persistentData->buffer), inputStringFlags, &ValidateDropdown, &validateData)) {
        for (int i = 0; i < count; i++) {
            if (DropdownStringMatchesInput_(strings[i] + strOffset, persistentData->buffer, sizeof(persistentData->buffer))) {
                *index = i;
                result = true;
                persistentData->dropped = false;
                break;
            }
        }
    }

    // open the dropdown when the InputString gains focus
    WidgetID inputStringID = group->prevID;

    if (IsActive(manager, inputStringID)) {
        persistentData->dropped = true;
        persistentData->filterUsingBuffer = true;
    }
    else {
        Plat.ClearMem(persistentData->buffer, StringLength(persistentData->buffer));
        Plat.CopyMem(persistentData->buffer, preValidatedBuffer, StringLength(preValidatedBuffer));
    }

    EtcRect buttonBounds;
    buttonBounds.pos = V2(inputStringBounds.pos.x + inputStringBounds.size.x,
                          inputStringBounds.pos.y);
    buttonBounds.size = V2(widgetBounds.size.y, widgetBounds.size.y);

    SetNextWidgetBounds(manager, group, buttonBounds);

    // @TODO: ImageButton()
    if (Button("")) {
        persistentData->dropped = !persistentData->dropped;
        persistentData->filterUsingBuffer = false;

        // @TODO: set InputString active, but that requires some additional functionality!
    }

    if (Texture *buttonImage = GetTexture(Game.assets, DebugMem->debugUIArrowDown)) {
        EtcRect imgBounds;
        imgBounds.pos.x = Roundf(buttonBounds.pos.x + 0.5f * buttonBounds.size.x - 0.5f * buttonImage->width);
        imgBounds.pos.y = Roundf(buttonBounds.pos.y + 0.5f * buttonBounds.size.y - 0.5f * buttonImage->height);
        imgBounds.size.x = buttonImage->width;
        imgBounds.size.y = buttonImage->height;

        mat4 model = TS(V3(imgBounds.pos, 0), V3(imgBounds.size, 0));
        RenderMesh(manager, group, &Game.graphics->glyphQuad, buttonImage, model, WHITE);
    }


    if (displayName) {
        PopStyles();
    }

    bool activeIDIsButton = false;
    if (persistentData->dropped) {
        // @TODO: dont manually set the size here!
        EtcRect nextBounds = GetNextWidgetBounds(manager, group);
        nextBounds.size.y = 128;
        PushWindow("ScrollWindow", nextBounds.pos, nextBounds.size, 0u);

        PushScrollGroup("Scroll", flags);
        for (int i = 0; i < count; i++) {
            bool display = true;
            if (persistentData->filterUsingBuffer) {
                display = DropdownStringMatchesInput_(strings[i] + strOffset, persistentData->buffer, sizeof(persistentData->buffer));
            }

            if (display) {
                // @TODO: only return true on mouse button up (probably a flag)
                // @BUG: Cant use TrueOnRelease here, think 
                uint32 flags = 0;
                if (Button(strings[i] + strOffset)) {
                    *index = i;
                    result = true;

                    persistentData->dropped = false;
                    persistentData->filterUsingBuffer = false;
                }
            }

            activeIDIsButton |= (bool)IsActive(manager, group->prevID);
        }
        PopScrollGroup();

        PopWindow();
    }


    // close the dropdown when the InputString loses focus
    if (persistentData->filterUsingBuffer) {
        if (!IsActive(manager, inputStringID)) {
            persistentData->dropped = false;
            persistentData->filterUsingBuffer = false;
        }
    }

    if (!persistentData->inputStringWasActive && IsActive(manager, inputStringID)) {
        persistentData->initialValue = *index;
    }
    else if (!activeIDIsButton && persistentData->inputStringWasActive && !IsActive(manager, inputStringID) && !result) {
        *index = persistentData->initialValue;

        Plat.ClearMem(persistentData->buffer, sizeof(persistentData->buffer));
        Plat.CopyMem(persistentData->buffer, strings[*index] + strOffset, StringLength(strings[*index]) - strOffset);
    }


    persistentData->inputStringWasActive = IsActive(manager, inputStringID);


    if (result) {
        Plat.ClearMem(persistentData->buffer, sizeof(persistentData->buffer));
        Plat.CopyMem(persistentData->buffer, strings[*index] + strOffset, StringLength(strings[*index]) - strOffset);
    }

    return result;
}

b32 Dropdown(const char *name, const char **strings, uint32 count, uint32 *index, uint32 flags) {
    return Dropdown_(name, strings, count, index, 0, flags);
}

b32 DropdownEnum(const char *name, const char *enumName, uint32 *index, uint32 flags) {
    if (Schema *schema = Game.GetSchema(enumName)) {
        if (schema->type == SchemaType_Enum) {
            bool useShortNames = GetStyleb(UIStyle_UseShortEnumNames);
            
            if (useShortNames) {
                return Dropdown_(name, schema->_enum.stringValues, schema->_enum.count, index, StringLength(enumName) + 1, flags);
            }
            else {
                return Dropdown_(name, schema->_enum.stringValues, schema->_enum.count, index, 0, flags);
            }
        }
    }

    return false;
}

b32 ValidateAssetName(const char *str, InputStringActiveData *activeData, void *data) {
    if (StringLength(str) == 0) {
        return true;
    }

    const char *path = FindPathForAsset(Game.assets, str, *(AssetType *)data);
    return path != NULL;
}

struct InputAssetFieldPersistentData {
    char buffer[128];
};

enum InputAssetHandleFlags {
    InputAssetHandleFlag_Edit = 1 << 0,
};

b32 InputAssetHandle(const char *name, AssetType type, AssetHandle *handle, uint32 flags) {
    b32 result = false;
    
    UIManager *manager = GetCurrentUIManager();
    WidgetGroup *group = PeekWidgetGroup(manager);
    WidgetID id = PushWidgetID(manager, name);

    static uint32 InputAssetFieldDataID = GenerateDataID("InputAssetFieldBuffer");

    InputAssetFieldPersistentData *persistentData = NULL;
    if (!GetPersistentData(manager, id, InputAssetFieldDataID, (void **)&persistentData)) {
        persistentData = (InputAssetFieldPersistentData *)AllocatePersistentData(manager, id, InputAssetFieldDataID, sizeof(InputAssetFieldPersistentData));

        if (handle->asset != NULL && handle->asset->name != NULL) {
            Plat.CopyMem(persistentData->buffer, handle->asset->name, StringLength(handle->asset->name));
        }
    }

    u32 strFlags = InputStringFlag_ReuseWidgetID_ | InputStringFlag_ReturnTrueOnEnter;
    if (InputString(name, persistentData->buffer, ARRAY_LENGTH(char, persistentData->buffer), strFlags, &ValidateAssetName, &type)) {
        *handle = LoadAsset(Game.assets, type, persistentData->buffer);
        result = true;
    }

    return result;;
}

struct ValidateEntityNameData {
    EntityDB *entityDB;
    uint32 ignoredIDCount;
    EntityID *ignoredIDs;
};

b32 ValidateEntityName(const char *str, InputStringActiveData *activeData, void *data) {
    ValidateEntityNameData *validateData = (ValidateEntityNameData *)data;

    if (StringLength(str) == 0) {
        return true;
    }

    NameSystem *nameSys = GetSystem(validateData->entityDB, Name);
    EntityID entityID = EntityNamed(nameSys, str);

    if (entityID == EntityID_Invalid) {
        return false;
    }

    for (int i = 0; i < validateData->ignoredIDCount; i++) {
        if (entityID == validateData->ignoredIDs[i]) {
            return false;
        }
    }

    return true;
}

struct InputEntityNamePersistentData {
    char buffer[MAX_ENTITY_NAME_LENGTH];
};

b32 InputEntityName(const char *name, EntityDB *entityDB, EntityID *entityID, uint32 ignoredIDCount, EntityID *ignoredIDs, uint32 flags) {
    NameSystem *nameSys = GetSystem(entityDB, Name);
    
    UIManager *manager = GetCurrentUIManager();
    WidgetGroup *group = PeekWidgetGroup(manager);
    WidgetID id = PushWidgetID(manager, name);
    
    static uint32 InputEntityNameDataID = GenerateDataID("InputEntityNameBuffer");

    InputEntityNamePersistentData *persistentData = NULL;
    if (!GetPersistentData(manager, id, InputEntityNameDataID, (void **)&persistentData)) {
        persistentData = (InputEntityNamePersistentData *)AllocatePersistentData(manager, id, InputEntityNameDataID, sizeof(InputEntityNamePersistentData));
        Plat.ClearMem(persistentData, sizeof(InputEntityNamePersistentData));

        if (*entityID != EntityID_Invalid) {
            char *name = GetEntityName(nameSys, *entityID);
            if (name != NULL) {
                Plat.CopyMem(persistentData->buffer, name, MAX_ENTITY_NAME_LENGTH);    
            }
        }
    }
    
    flags |= InputStringFlag_ReuseWidgetID_;

    ValidateEntityNameData validateData;
    validateData.entityDB = entityDB;
    validateData.ignoredIDCount = ignoredIDCount;
    validateData.ignoredIDs = ignoredIDs;

    if (InputString(name, persistentData->buffer, sizeof(persistentData->buffer), flags, &ValidateEntityName, &validateData)) {
        *entityID = EntityNamed(nameSys, persistentData->buffer);
        return true;
    }

    return false;
}

b32 InputEntityName(const char *name, EntityDB *entityDB, EntityID *entityID, uint32 flags) {
    return InputEntityName(name, entityDB, entityID, 0, NULL, flags);
}

b32 InputSchemaDataValue_(const char *name, EntityDB *entityDB, SchemaTypeDescription type, void *value, uint32 flags) {
    ASSERT(value != NULL);

    UIManager *manager = GetCurrentUIManager();
    WidgetGroup *group = PeekWidgetGroup(manager);
    WidgetID id = PushWidgetID(manager, name);
    
    static uint32 InputSchemaTypeDataID = GenerateDataID("InputSchemaTypeBuffer");

    // @TODO: handle fixed-length arrays (as specified in the type description)

    switch (type.dataType) {
        case SchemaDataType_Invalid: {
        } break;

        case SchemaDataType_void: {
        } break;

        case SchemaDataType_bool32: {
            if (type.arrayLength > 0) {
                b32 modified = false;
                PushText(name);

                UIStyle styles[] = {
                    MakeStyleFlowDirection(UIStyle_FlowDirection, UIFlowDirection_Horizontal),  
                };
                PushStyles(styles, ARRAY_LENGTH(UIStyle, styles), manager);
                
                for (int i = 0; i < type.arrayLength; i++) {
                    uint32 arrayNameLength = 64;
                    char *arrayName = PushArray(Game.frameMem, char, arrayNameLength);
                    Plat.ClearMem(arrayName, arrayNameLength);
                    stbsp_snprintf(arrayName, arrayNameLength, "[%d]", i);
                    modified |= Toggle(arrayName, (b32 *)((uint8 *)value + (sizeof(b32) * i)), flags);
                }

                PopStyles();
                
                return modified;
            }
            else {
                return Toggle(name, (b32 *)value, flags);
            }
        } break;

        case SchemaDataType_bool: {
            if (type.arrayLength > 0) {
                b32 modified = false;
                PushText(name);

                UIStyle styles[] = {
                    MakeStyleFlowDirection(UIStyle_FlowDirection, UIFlowDirection_Horizontal),  
                };
                PushStyles(styles, ARRAY_LENGTH(UIStyle, styles), manager);
                
                for (int i = 0; i < type.arrayLength; i++) {
                    uint32 arrayNameLength = 64;
                    char *arrayName = PushArray(Game.frameMem, char, arrayNameLength);
                    Plat.ClearMem(arrayName, arrayNameLength);
                    stbsp_snprintf(arrayName, arrayNameLength, "[%d]", i);
                    modified |= Toggle(arrayName, (b32 *)((uint8 *)value + (sizeof(bool) * i)), flags);
                }

                PopStyles();
                
                return modified;
            }
            else {
                return Toggle(name, (bool *)value, flags);
            }
        } break;

        case SchemaDataType_char: {
            // @TODO: ...
        } break;

        case SchemaDataType_int8: {
            // @TODO: ...
        } break;

        case SchemaDataType_int16: {
            // @TODO: ...
        } break;
            
        case SchemaDataType_int32: {
            return InputInt(name, (int32 *)value, flags);
        } break;

        case SchemaDataType_int64: {
            // @TODO: ...
        } break;

        case SchemaDataType_uint8: {
            // @TODO: ...
        } break;

        case SchemaDataType_uint16: {
            // @TODO: ...
        } break;
            
        case SchemaDataType_uint32: {
            return InputUInt(name, (uint32 *)value, flags);
        } break;

        case SchemaDataType_uint64: {
            // @TODO: ...
        } break;

        case SchemaDataType_real32: {
            return InputFloat(name, (real32 *)value, flags);
        } break;

        case SchemaDataType_vec2: {
            return InputVec2(name, (vec2 *)value, flags);
        } break;
        case SchemaDataType_vec3: {
            return InputVec3(name, (vec3 *)value, flags);
        } break;
        case SchemaDataType_vec4: {
            return InputVec4(name, (vec4 *)value, flags);
        } break;
        case SchemaDataType_Color: {
            return InputColor(name, (Color *)value, flags);
        } break;

        case SchemaDataType_vec2i: {
            // @TODO: ...
        } break;
        case SchemaDataType_vec3i: {
            // @TODO: ...
        } break;
        case SchemaDataType_vec4i: {
            // @TODO: ...
        } break;

        case SchemaDataType_mat3: {
            // @TODO: ...
        } break;
        case SchemaDataType_mat4: {
            // @TODO: ...
        } break;

        case SchemaDataType_quaternion: {
            return InputQuaternion(name, (quaternion *)value, flags);
        } break;

        case SchemaDataType_Schema: {
            // @TODO: ...
        } break;

        case SchemaDataType_SchemaData: {
            return InputSchemaData(name, entityDB, (SchemaData *)value, flags);
        } break;

        case SchemaDataType_EntityID: {
            return InputEntityName(name, entityDB, (EntityID *)value, flags);
        } break;

        case SchemaDataType_DynamicArray: {
            uint32 *arrayCount = (uint32 *)value;
            uint32 *arrayCapacity = (uint32 *)((uint8 *)value + sizeof(uint32));
            void **dataPtr = (void **)((uint8 *)value + sizeof(uint32) * 2);
            return InputDynamicArray_(name, entityDB, type.templateTypes[0], arrayCapacity, arrayCount, dataPtr, flags);
        } break;

            
        case SchemaDataType_ChunkedArray: {
            // @TODO: ...
        } break;

        case SchemaDataType_HashTable: {
            // @TODO: ...
        } break;


        case SchemaDataType_String: {
            flags |= InputStringFlag_ReuseWidgetID_;

            if (type.maxStringLength > 0) {
                return InputString(name, (char *)value, type.maxStringLength, flags);
            }
            else {
                uint32 stringLength = 0;
                if (*(char **)value != NULL) {
                    stringLength = StringLength(*(char **)value);   
                }
                u32 bufferLen = Max(32u, stringLength * 2);
                char *buffer = PushArray(Game.frameMem, char, bufferLen);
                Plat.ClearMem(buffer, sizeof(char) * bufferLen);
                Plat.CopyMem(buffer, *(char **)value, stringLength);

                if (InputString(name, buffer, bufferLen, flags)) {
                    *(char **)value = (char *)StringCopy(buffer);
                    return true;
                }
            }
        } break;

            
        case SchemaDataType_Struct:
        case SchemaDataType_Enum:
        case SchemaDataType_Bitfield:
        case SchemaDataType_Bitarray:
        case SchemaDataType_TaggedUnion:
        case SchemaDataType_PersistentFunctionHandle:
        case SchemaDataType_PersistentFunctionDataPair:
        case SchemaDataType_AssetHandle: {
            if (Schema *schema = Game.GetSchema(type.typeName)) {
                b32 pushTree = GetStyleb(UIStyle_StructPushTree, manager);

                PushStyle(MakeStyleb(UIStyle_StructPushTree, true), manager);
                b32 result = InputSchemaValue(name, entityDB, schema, value, flags);
                PopStyles(manager);

                return result;
            }
        } break;
    }

    return false;
}

b32 InputSchemaStructFieldValue(const char *name, const char *fieldName, EntityDB *entityDB, Schema *schema, void *value, u32 flags) {
    ASSERT(schema->type == SchemaType_Struct);

    b32 modified = false;

    for (int i = 0; i < schema->_struct.fields.count; i++) {
        SchemaField field = schema->_struct.fields[i];

        if (!field.editable) {
            continue;
        }

        if (StringEquals(field.name, fieldName)) {
            void *offsetValue = (void *)((uint8 *)value + field.byteOffset);
            modified |= InputSchemaDataValue_(field.name, entityDB, field.type, offsetValue, flags);
            break;
        }
    }

    return modified;
}

b32 InputSchemaValue(const char *name, EntityDB *entityDB, Schema *schema, void *value, uint32 flags) {

    if (!schema->editable) {
        return false;
    }

    UIManager *manager = GetCurrentUIManager();

    b32 modified = false;

    switch (schema->type) {
        case SchemaType_Struct: {
            b32 pushTree = GetStyleb(UIStyle_StructPushTree, manager);
            if (!pushTree) {
                for (int i = 0; i < schema->_struct.fields.count; i++) {
                    SchemaField field = schema->_struct.fields[i];

                    if (!field.editable) {
                        continue;
                    }
                    
                    void *offsetValue = (void *)((uint8 *)value + field.byteOffset);
                    modified |= InputSchemaDataValue_(field.name, entityDB, field.type, offsetValue, flags);
                }
            }
            else if (PushTreeGroup(name)) {
                for (int i = 0; i < schema->_struct.fields.count; i++) {
                    SchemaField field = schema->_struct.fields[i];

                    if (!field.editable) {
                        continue;
                    }
                    
                    void *offsetValue = (void *)((uint8 *)value + field.byteOffset);
                    modified |= InputSchemaDataValue_(field.name, entityDB, field.type, offsetValue, flags);
                }

                PopTreeGroup();
            }
        } break;

        case SchemaType_Enum: {
            modified = DropdownEnum(name, schema->name, (uint32 *)value, flags);
        } break;

        case SchemaType_Bitfield: {
            b32 pushTree = GetStyleb(UIStyle_BitfieldPushTree, manager);
            if (!pushTree) {
                modified |= InputBitfield(name, schema, (uint32 *)value, flags);
            }
            else if (PushTreeGroup(name)) {
                modified |= InputBitfield(name, schema, (uint32 *)value, flags);
                PopTreeGroup();
            }
        } break;

        case SchemaType_Bitarray: {
            uint32 count = *(uint32 *)value;
            uint32 *bits = (uint32 *)((u8 *)value);
            modified |= InputBitarray(name, schema, count, bits, flags);
        } break;

        case SchemaType_TaggedUnion: {
            TaggedUnion_Untyped *union_ = (TaggedUnion_Untyped *)value;
            uint32 *type = &union_->type;
            modified |= Dropdown_(name, schema->_tagged_union.typeTagStringValues, schema->_tagged_union.fields.count, type, StringLength(schema->name), flags);

            SchemaField field = schema->_tagged_union.fields[*type];
            modified |= InputSchemaDataValue_(field.name, entityDB, field.type, (void *)((u8 *)value + field.byteOffset), flags);
        } break;

        case SchemaType_PersistentFunctionHandle: {
            FuncHandle *funcHandle = (FuncHandle *)value;
            PersistentFuncType funcType = schema->_func_handle.funcType;
            return InputFuncHandle(name, funcType, funcHandle, flags);
        } break;

        case SchemaType_PersistentFunctionDataPair: {
            return InputFuncDataPair(name, entityDB, schema, (FuncDataPair *)value, flags);
        } break;

        case SchemaType_AssetHandle: {
            uint32 handleStringIndex = FindSubstring(schema->name, "Handle");
            char *sansHandle = String(schema->name, handleStringIndex);
            const char *typeName = StringConcat(Game.frameMem, 2, "AssetType_", sansHandle);

            for (uint32 i = 0; i < ENUM_VALUECOUNT(AssetType); i++) {
                if (StringEquals(ENUM_STRINGVALUE(AssetType, i), typeName)) {
                    return InputAssetHandle(name, (AssetType)i, (AssetHandle *)value, flags);
                }
            }
        } break;
    }

    return modified;
}

b32 InputSchemaValue(const char *name, EntityDB *entityDB, const char *schemaName, void *data, uint32 flags) {
    Schema *schema = Game.GetSchema(schemaName);
    return InputSchemaValue(name, entityDB, schema, data, flags);
}


struct InputSchemaDataPersistentData {
    char buffer[128];
};

b32 InputSchemaData_(const char *name, EntityDB *entityDB, const char *schemaName, void **data, uint32 flags) {
    if (schemaName == NULL) { return false; }
    Schema *schema = Game.GetSchema(schemaName);

    if (schema != NULL) {

        if (*data == NULL) {
            *data = Plat.AllocateMem(schema->sizeInBytes);
            Plat.ClearMem(*data, schema->sizeInBytes);
            return true;
        }

        return InputSchemaValue(name, entityDB, schema, *data);
    }

    return false;
}

b32 InputSchemaData(const char *name, EntityDB *entityDB, const char *schemaName, SchemaData *schemaData, uint32 flags) {
    return InputSchemaData_(name, entityDB, schemaName, &schemaData->data, flags);
}

b32 ValidateInputSchema(const char *str, InputStringActiveData *activeData, void *schema) {
    if (Game.GetSchema(str) != NULL) {
        return true;
    }
    if (StringLength(str) == 0 || (Schema *)schema != NULL) {
        return true;
    }

    return false;
}

b32 InputSchemaData(const char *name, EntityDB *entityDB, SchemaData *schemaData, uint32 flags) {
    Schema *schema = NULL;

    if (schemaData->schemaName != NULL) {
        schema = Game.GetSchema(schemaData->schemaName);    
    }
    
    UIManager *manager = GetCurrentUIManager();
    WidgetGroup *group = PeekWidgetGroup(manager);
    WidgetID id = PushWidgetID(manager, name);
    
    static uint32 InputSchemaDataDataID = GenerateDataID("InputSchemaDataBuffer");

    InputSchemaDataPersistentData *persistentData = NULL;
    if (!GetPersistentData(manager, id, InputSchemaDataDataID, (void **)&persistentData)) {
        persistentData = (InputSchemaDataPersistentData *)AllocatePersistentData(manager, id, InputSchemaDataDataID, sizeof(InputSchemaDataPersistentData));

        if (schema != NULL) {
            Plat.CopyMem(persistentData->buffer, schema->name, StringLength(schema->name));
        }
    }

    b32 result = false;
    
    if (PushTreeGroup(name)) {
        flags |= InputStringFlag_ReuseWidgetID_;
        if (InputString(name, persistentData->buffer, 128, flags, &ValidateInputSchema, schema) &&
            StringLength(persistentData->buffer) > 0) {
            schema = Game.GetSchema(persistentData->buffer);

            if (schema != NULL) {
                schemaData->schemaName = (char *)StringCopy(persistentData->buffer);
                Plat.DeallocateMem(schemaData->data);
                schemaData->data = Plat.AllocateMem(schema->sizeInBytes);
                Plat.ClearMem(schemaData->data, schema->sizeInBytes);

                result = true;
            }
        }
        else {
            result = InputSchemaData_(name, entityDB, schemaData->schemaName, &schemaData->data, flags);    
        }

        PopTreeGroup();

        return result;
    }
    
    return false;
}

b32 InputComponentValue(const char *name, EntityDB *entityDB, ComponentType type, ComponentValue *value, uint32 flags) {
    if (Schema *schema = ComponentSchemaFromComponentType(type)) {
        return InputSchemaValue(name, entityDB, schema, value, flags);
    }

    return false;
}


void PushSchemaDataValue_(const char *name, EntityDB *entityDB, SchemaTypeDescription type, void *value) {
    switch (type.dataType) {
        case SchemaDataType_Invalid: {
        } break;

        case SchemaDataType_void: {
        } break;

        case SchemaDataType_bool32: {
        } break;

        case SchemaDataType_bool: {
        } break;

        case SchemaDataType_char: {
            if (type.arrayLength > 0) {
                PushTextf("%s %s", name, (char *)value, type.arrayLength);
            }
        } break;

        case SchemaDataType_int8: {
            // @TODO: ...
        } break;

        case SchemaDataType_int16: {
            // @TODO: ...
        } break;
            
        case SchemaDataType_int32: {
            //InputInt(name, (int32 *)value);
        } break;

        case SchemaDataType_int64: {
            // @TODO: ...
        } break;

        case SchemaDataType_uint8: {
            // @TODO: ...
        } break;

        case SchemaDataType_uint16: {
            // @TODO: ...
        } break;
            
        case SchemaDataType_uint32: {
            //return InputUInt(name, (uint32 *)value, flags);
        } break;

        case SchemaDataType_uint64: {
            // @TODO: ...
        } break;

        case SchemaDataType_real32: {
            PushTextf("%s %f", name, *(real32 *)value);
        } break;

        case SchemaDataType_vec2: {
            //return InputVec2(name, (vec2 *)value, flags);
        } break;
        case SchemaDataType_vec3: {
            vec3 v = *(vec3 *)value;
            PushTextf("%s (%f %f %f)", name, v.x, v.y, v.z);
        } break;
        case SchemaDataType_vec4: {
            vec4 v = *(vec4 *)value;
            PushTextf("%s (%f %f %f %f)", name, v.x, v.y, v.z, v.w);
        } break;
        case SchemaDataType_Color: {
            vec4 v = *(vec4 *)value;
            PushTextf("%s (%f %f %f %f)", name, v.x, v.y, v.z, v.w);
        } break;

        case SchemaDataType_vec2i: {
            // @TODO: ...
        } break;
        case SchemaDataType_vec3i: {
            // @TODO: ...
        } break;
        case SchemaDataType_vec4i: {
            // @TODO: ...
        } break;

        case SchemaDataType_mat3: {
            // @TODO: ...
        } break;
        case SchemaDataType_mat4: {
            // @TODO: ...
        } break;

        case SchemaDataType_quaternion: {
            vec3 euler = ToEulerAngles(*(quaternion *)value);

            euler.x = RadToDeg(euler.x);
            euler.y = RadToDeg(euler.y);
            euler.z = RadToDeg(euler.z);
            PushTextf("%s (%f %f %f)", name, euler.x, euler.y, euler.z);
            //return InputQuaternion(name, (quaternion *)value, flags);
        } break;

        case SchemaDataType_Schema: {
            if ((Schema *)value != NULL &&
                ((Schema *)value)->name != NULL) {
                PushTextf("%s %s", name, ((Schema *)value)->name);
            }
            else {
                PushTextf("%s Invalid Schema", name);
            }
        } break;

        case SchemaDataType_SchemaData: {
            SchemaData *schemaData = (SchemaData *)value;
            if (schemaData->schemaName != NULL) {
                Schema *schema = Game.GetSchema(schemaData->schemaName);
                if (schema != NULL) {
                    PushTextf("schema %s", schemaData->schemaName);
                    PushSchemaValue(name, entityDB, schema, schemaData->data);
                }
            }
            else {
                PushTextf("%s Invalid Schema", name);
            }
        } break;

        case SchemaDataType_EntityID: {
            //return InputEntityName(name, entityDB, (EntityID *)value, flags);
        } break;

        case SchemaDataType_DynamicArray: {
            uint32 *arrayCount = (uint32 *)value;
            uint32 *arrayCapacity = (uint32 *)((uint8 *)value + sizeof(uint32));

            uint32 elemSize = type.templateTypes[0].sizeInBytes;

            void *dataPtr = *(void **)((u8 * )value + (sizeof(uint32) * 2));
            for (int i = 0; i < *arrayCount; i++) {
                void *elemPtr = (void *)((uint8 *)dataPtr + (elemSize * i));
                char indexName[8];
                Plat.ClearMem(indexName, 8);
                stbsp_snprintf(indexName, 8, "[%d]", i);
                const char *elemName = StringConcat(Game.frameMem, 2, name, indexName);
                PushSchemaDataValue_(elemName, entityDB, type.templateTypes[0], elemPtr);
            }
            
        } break;

        case SchemaDataType_ChunkedArray: {
            // @TODO: ...
        } break;

        case SchemaDataType_HashTable: {
            // @TODO: ...
        } break;


        case SchemaDataType_String: {
            if (type.maxStringLength > 0) {
                uint32 stringLength = StringLength((char *)value, type.maxStringLength);
                PushTextf("%s: %.*s", name, stringLength, (char *)value);
            }
            else {
                PushTextf("%s: %s", name, *((char **)value));
            }
        } break;

            
        case SchemaDataType_Struct:
        case SchemaDataType_Enum:
        case SchemaDataType_Bitfield:
        case SchemaDataType_Bitarray:
        case SchemaDataType_TaggedUnion:
        case SchemaDataType_PersistentFunctionHandle:
        case SchemaDataType_PersistentFunctionDataPair:
        case SchemaDataType_AssetHandle: {
            if (Schema *schema = Game.GetSchema(type.typeName)) {
                PushSchemaValue(name, entityDB, schema, value);
            }
        } break;
    }
}

void PushSchemaValue(const char *name, EntityDB *entityDB, Schema *schema, void *value) {
    UIManager *manager = GetCurrentUIManager();
    switch (schema->type) {
        case SchemaType_Struct: {
            b32 pushTree = GetStyleb(UIStyle_StructPushTree, manager);
            if (!pushTree) {
                for (int i = 0; i < schema->_struct.fields.count; i++) {
                    SchemaField field = schema->_struct.fields[i];
                    void *offsetValue = (void *)((uint8 *)value + field.byteOffset);
                    PushSchemaDataValue_(field.name, entityDB, field.type, offsetValue);
                }
            }
            else if (PushTreeGroup(name)) {
                for (int i = 0; i < schema->_struct.fields.count; i++) {
                    SchemaField field = schema->_struct.fields[i];
                    void *offsetValue = (void *)((uint8 *)value + field.byteOffset);
                    PushSchemaDataValue_(field.name, entityDB, field.type, offsetValue);
                }

                PopTreeGroup();
            }
        } break;

        case SchemaType_Enum: {
            PushTextf("%s: %s", name, schema->_enum.stringValues[*(uint32 *)value]);
        } break;

        case SchemaType_Bitfield: {
            b32 pushTree = GetStyleb(UIStyle_BitfieldPushTree, manager);
            if (!pushTree) {
            }
            else if (PushTreeGroup(name)) {
                PopTreeGroup();
            }
        } break;

        case SchemaType_Bitarray: {

        } break;

        case SchemaType_TaggedUnion: {
            TaggedUnion_Untyped *union_ = (TaggedUnion_Untyped *)value;
            uint32 type = union_->type;
            SchemaField field = schema->_tagged_union.fields[type];
            PushTextf("%s: %s", name, schema->_tagged_union.typeTagStringValues[type]);
            PushSchemaDataValue_(field.name, entityDB, field.type, (void *)((u8 *)value + field.byteOffset));
        } break;

        case SchemaType_PersistentFunctionHandle: {
            FuncHandle *funcHandle = (FuncHandle *)value;
            PersistentFuncType funcType = schema->_func_handle.funcType;
        } break;

        case SchemaType_PersistentFunctionDataPair: {
            
        } break;

        case SchemaType_AssetHandle: {
            uint32 handleStringIndex = FindSubstring(schema->name, "Handle");
            char *sansHandle = String(schema->name, handleStringIndex);
            const char *typeName = StringConcat(Game.frameMem, 2, "AssetType_", sansHandle);

            // TODO: ...
        } break;
    }
}

